/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "line.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
GLine::GLine(const GVector& position, const GVector& direction)
	: u(direction),
	  v(direction.X(position))
{
}

// operations --------------------------------------------------------
bool
GLine::FindIntersection(GVector* result, const GLine& other) const
{
	//From <http://www.acm.org/tog/resources/RTNews/html/rtnv11n1.html#art3>,
	// by Ken Shoemake (ks@emanon.net):
	//U1.V2 + U2.V1 =? 0 tests if L1 = {U1:V1} and L2 = {U2:V2} are coplanar.
	//...
	//Let N be a unit vector along a coordinate axis, with (U1xU2).N non-zero.
	//((V1.N)U2-(V2.N)U1-(V1.U2)N:(U1xU2).N) is the point of intersection, if
	// any.
	
	if(!IsApproxZero(u * other.v + other.u * v))
		return false;        //not coplanar

	GVector U1xU2(u.X(other.u));
	
	//Find which coordinate axis N should lie on.
	int coord;
	if(!IsApproxZero(U1xU2[0]))
		coord = 0;
	else if(!IsApproxZero(U1xU2[1]))
		coord = 1;
	else if(!IsApproxZero(U1xU2[2]))
		coord = 2;
	else
		return false;        //parallel lines
	
	if(result == NULL)
		return true;         //no need to compute exact point
	
	*result = v[coord] * other.u - other.v[coord] * u;
	(*result)[coord] -= (v * other.u);
	*result /= U1xU2[coord];
	
	return true;
}

//FindDistance returns the shortest distance between the objects in
// question.
GCoordinate
GLine::FindDistance(const GVector& point) const
{
	//From <http://www.acm.org/tog/resources/RTNews/html/rtnv11n1.html#art3>,
	// by Ken Shoemake (ks@emanon.net):
	//(V.V)/(U.U) is the minimum squared distance of L from the origin.
	
	//So, we want to translate the line to be relative to the point (so the point
	// is at the line's origin)
	
	//v = u X p
	//So we want to do this:
	//newV = u X (p - point)
	//     = u X p - u X point
	//     = v - u X point
	//     = v + point X u
	//     = point X u + v
	
	GVector newV(point.X(u));
	newV += v;
	
	return sqrt((newV * newV) / (u * u));
}

GCoordinate
GLine::FindSquareDistance(const GVector& point) const
{
	//From <http://www.acm.org/tog/resources/RTNews/html/rtnv11n1.html#art3>,
	// by Ken Shoemake (ks@emanon.net):
	//(V.V)/(U.U) is the minimum squared distance of L from the origin.
	
	//So, we want to translate the line to be relative to the point (so the point
	// is at the line's origin)
	
	//v = u X p
	//So we want to do this:
	//newV = u X (p - point)
	//     = u X p - u X point
	//     = v - u X point
	//     = v + point X u
	//     = point X u + v
	
	GVector newV(point.X(u));
	newV += v;
	
	return (newV * newV) / (u * u);
}

GCoordinate
GLine::FindDistance(const GLine& other) const
{
	//If p1 is a point on line L1 = {u1:v1} and
	//   p2 is a point on line L2 = {u2:v2} then
	//   the shortest distance d between the two lines is as follows:
	//d = (p1 - p2) * (u1 X u2) / ||u1 X u2||
	//  = ((u1 X u2 * p1) - (u1 X u2 * p2)) / ||u1 X u2||
	//  =(-(u1 X p1 * u2) - (u2 X p2 * u1)) / ||u1 X u2||
	//  =-((v1 * u2) + (v2 * u1)) / ||u1 X u2||
	
	//I derived that myself!  :) -Temporal
	
	GVector U1xU2(u.X(other.u));
	GCoordinate squareMagnitude = U1xU2.SquareMagnitude();
	if(IsApproxZero(squareMagnitude))
	{
		//lines are parallel.  Find closest points on each line to the
		// origin and compare.
		
		//From <http://www.acm.org/tog/resources/RTNews/html/rtnv11n1.html#art3>,
		// by Ken Shoemake (ks@emanon.net):
		//(VxU:U.U) is the point of L closest to the origin.
		
		GVector p1(v.X(u));
		p1 /= u * u;
		
		GVector p2(other.v.X(other.u));
		p2 /= other.u * other.u;
		
		return (p1 - p2).Magnitude();
	}
	
	return (GCoordinate)fabs((other.v * u) + (v * other.u)) / sqrt(squareMagnitude);
}

GCoordinate
GLine::FindSquareDistance(const GLine& other) const
{
	//If p1 is a point on line L1 = {u1:v1} and
	//   p2 is a point on line L2 = {u2:v2} then
	//   the shortest distance d between the two lines is as follows:
	//d = (p1 - p2) * (u1 X u2) / ||u1 X u2||
	//  = ((u1 X u2 * p1) - (u1 X u2 * p2)) / ||u1 X u2||
	//  =(-(u1 X p1 * u2) - (u2 X p2 * u1)) / ||u1 X u2||
	//  =-((v1 * u2) + (v2 * u1)) / ||u1 X u2||
	
	//I derived that myself!  :) -Temporal
	
	GVector U1xU2(u.X(other.u));
	GCoordinate squareMagnitude = U1xU2.SquareMagnitude();
	if(IsApproxZero(squareMagnitude))
	{
		//lines are parallel.  Find closest points on each line to the
		// origin and compare.
		
		GVector p1(u.X(v));
		p1 /= u * u;
		
		GVector p2(other.u.X(other.v));
		p2 /= other.u * other.u;
		
		return (p1 - p2).SquareMagnitude();
	}
	
	GCoordinate d = ((other.v * u) + (v * other.u));
	return d * d / squareMagnitude;
}

// operators ---------------------------------------------------------
inline const GLine&
GLine::operator+=(const GVector& other)
{
	//v = u X point
	//So we want to do this:
	//newV = u X (point + other)
	//     = u X point + u X other
	//     = v + u X other
	v += u.X(other);
	return *this;
}

inline const GLine&
GLine::operator-=(const GVector& other)
{
	//v = u X point
	//So we want to do this:
	//newV = u X (point - other)
	//     = u X point - u X other
	//     = v - u X other
	v -= u.X(other);
	return *this;
}

} //namespace GAUGE3D
