/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "plane.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_LINE_H__INCLUDED__
#include "line.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
// operations --------------------------------------------------------
//FindIntersection returns the point of intersection between the plane
// and the line.
bool
GPlane::FindIntersection(GVector* result, const GLine& line) const
{
	//Intersection of line {U:V} and plane [N:d] is (VxN-Ud:U.N)
	
	GCoordinate w = line.u * n;
	
	if(IsApproxZero(w))
		return false;
	
	if(result == NULL)
		return true;
	
	*result = line.v.X(n);
	*result -= line.u * d;
	*result /= w;
	
	return true;
}

//Same sort of thing here, except find the intersection between two planes.
// This returns a line, of course.
bool
GPlane::FindIntersection(GLine* result, const GPlane& other) const
{
	//L = {ExF:fE-eF}, for [E:e] and [F:f] distinct planes containing L
	
	GVector U = n.X(other.n);
	
	if(IsApproxZero(U.x)
	&& IsApproxZero(U.y)
	&& IsApproxZero(U.z))
		return false;
	
	if(result == NULL)
		return true;
	
	result->u = U;
	result->v = other.d * n - d * other.n;
	
	return true;
}

//This one finds the point which is the intersection of three planes.
bool
GPlane::FindIntersection(GVector* result, const GPlane& other1, GPlane& other2) const
{
	//L = {ExF:fE-eF}, for [E:e] and [F:f] distinct planes containing L
	//(VxN-Un:U.N) is the point where L intersects plane [N:n] not parallel to L.
	//so...
	//((fE-eF)xG-(ExF)g:ExF*G) is the point where planes [E:e], [F:f], and [G:g]
	// intersect.
	
	GVector ExF;
	ExF = other1.n.X(other2.n);
	
	GCoordinate w = ExF * n;
	
	if(IsApproxZero(w))
		return false;
	
	if(result == NULL)
		return true;
	
	*result = other2.d * other1.n - other1.d * other2.n;  //fE - eF
	*result = result->X(n);
	*result -= ExF * d;
	*result /= w;
	
	return true;
}

//FindDistance returns the shortest distance between the objects in
// question.
GCoordinate
GPlane::FindDistance(const GVector& point) const
{
	return (GCoordinate)fabs(point * n + d) / n.Magnitude();
}

//FindSquareDistance returns the value of FindDistance squared.  This can
// be calculated more quickly than the actual distance.
GCoordinate
GPlane::FindSquareDistance(const GVector& point) const
{
	GCoordinate a = point * n + d;
	return a * a / n.SquareMagnitude();
}

} // namespace GAUGE3D
