/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "polygon.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_LINE_H__INCLUDED__
#include "line.h"
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_PLANE_H__INCLUDED__
#include "plane.h"
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_TRIANGLE_H__INCLUDED__
#include "triangle.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GPolygon::GPolygon()
	: mVertices (NULL),
	  mNumVertices(0)
{
}

GPolygon::GPolygon(const GTriangle& triangle)
	: mNumVertices(3)
{
	mVertices = new GVector[3];
	mVertices[0] = triangle[0];
	mVertices[1] = triangle[1];
	mVertices[2] = triangle[2];
}

GPolygon::GPolygon(const GPolygon& other)
	: mNumVertices(other.mNumVertices)
{
	mVertices = new GVector[mNumVertices];
	for(int i = 0; i < mNumVertices; i++)
		mVertices[i] = other.mVertices[i];
}

GPolygon::GPolygon(GVector vertices[], int num)
	: mNumVertices(num)
{
	mVertices = new GVector[num];
	for(int i = 0; i < num; i++)
		mVertices[i] = vertices[i];
}

GPolygon::~GPolygon()
{
	if(mNumVertices > 0)
		delete [] mVertices;
}

// operations --------------------------------------------------------
void
GPolygon::GetPlane(GPlane* plane) const
{
	GVector vec, normal;
	
	if(mNumVertices < 3)
		return;     //can't find plane without 3 vertices

	//Find normal vector.
	//first, find vector from vertex 0 to vertex 2
	normal = mVertices[2];
	normal -= mVertices[0];
	//then, find vector from vertex 0 to vertex 1
	vec = mVertices[1];
	vec -= mVertices[0];
	//and cross them
	normal = normal.X(vec);
	
	plane->n = normal;
	plane->d = -(mVertices[0] * normal);
}

//if you want to copy the old values to the resized list, pass true
// to copy.  To save time by deleting them, pass false.
void
GPolygon::Resize(int num, bool copy)
{
	GVector* oldVertices = NULL;

	if(mNumVertices <= 0)
		copy = false;         //don't copy (or delete) an empty list
	else if(copy)
		oldVertices = mVertices;
	else
		delete [] mVertices;

	mVertices = new GVector[num];
	
	if(copy)
	{
		if(mNumVertices > num)
			mNumVertices = num;

		for(int i = 0; i < mNumVertices; i++)
			mVertices[i] = oldVertices[i];
		delete [] oldVertices;
	}

	mNumVertices = num;
}

//vSplit splits the polygon into triangles.  mumVertices - 2 triangles
// will be created, so the target array should be of at least that size.
void
GPolygon::Split(GTriangle triangles[]) const
{
	GVector* vec1;
	GVector* vec2;
	
	vec1 = mVertices + 1;
	vec2 = mVertices + 2;
	
	for(int i = 0; i < mNumVertices - 2; i++)
	{
		triangles[i][0] = mVertices[0];
		triangles[i][1] = *vec1;
		triangles[i][2] = *vec2;
		vec1++;
		vec2++;
	}
}

void
GPolygon::Reverse()
{
	GVector* oldVertices;
	
	if(mNumVertices <= 0)
		return;    //no vertices to swap
	
	oldVertices = mVertices;
	mVertices = new GVector[mNumVertices];
	
	for(int i = 0; i < mNumVertices; i++)
		mVertices[i] = oldVertices[mNumVertices - i - 1];
	
	delete [] oldVertices;
}

//Contains returns true if the point is on the polygon.
bool
GPolygon::Contains(const GVector& point)
{
	if(mNumVertices < 3)
		return false; //polygon has no area, so screw it.

	GPlane plane;
	GetPlane(&plane);
	
	if(!plane.Contains(point))
		return false;    //point is not on plane.
	
	GLine l(point, plane.n);
	
	if(Intersects(l))
		return true;
	return false;
}

//When checking intersection with a line, 0 is returned if there is no
// intersection, 1 if the line hits the front of the polygon, or -1 if the
// line hits the back of the polygon.  (the front side is the side from which
// the vertices of the polygon appear clockwise)
//Note:  This function is faster than the one above.
//Note2: This function is very, *very* fast. >:)
int8
GPolygon::Intersects(const GLine& line)
{
	GVector u, v;  //u and v of edge line in Plucker coordinates.
	int i;
	GCoordinate dir;
	int8 result;
	
	if(mNumVertices < 3)
		return false;    //no area
	
	// L = {P-Q:PxQ}, for P and Q distinct points on L.
	i = mNumVertices - 1;
	u = mVertices[i] - mVertices[0];
	v = mVertices[i].X(mVertices[0]);
	
	dir = u * line.v + line.u * v;
	if(dir > 0.0f)
		result = -1;   //first edge is counter-clockwise
	else
		result = 1;    //first edge is clockwise
	
	//Now do the rest of the edges.
	for(i = 1; i < mNumVertices; i++)
	{
		// L = {P-Q:PxQ}, for P and Q distinct points on L.
		u = mVertices[i + 1] - mVertices[i];
		v = mVertices[i + 1].X(mVertices[i]);
		
		dir = u * line.v + line.u * v;
		if(dir > 0.0f)
			if(result > 0)
				return false;  //edge has different direction from first edge
		else
			if(result < 0)
				return false;  //edge has different direction from first edge
	}
	
	return result;
}

const GPolygon&
GPolygon::operator=(const GPolygon other)
{
	if(mNumVertices != other.mNumVertices)
	{
		if(mNumVertices != 0)
			delete [] mVertices;
		mNumVertices = other.mNumVertices;
		mVertices = new GVector[mNumVertices];
	}
	
	for(int i = 0; i < mNumVertices; i++)
		mVertices[i] = other.mVertices[i];

	return *this;
}

} // namespace GAUGE3D
