/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "gstring.h"

// system ------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
class GString::Buffer
{
	int   mLength;        //number of characters in string (not including terminator)
	int   mBufferSize;    //space allocated for buffer (may be larger than the string)
	                      // (mBufferSize is measured in sizeof(wchar_t) units, not bytes)
	int   mRefcount;      //number of GStrings using this buffer
	int   mUnicode;       //actually a bool.  Stored as int for alignment.  True if
	                      // the string is stored as Unicode.
	
public:
	Buffer();
	~Buffer();
	
	//The buffer object and the actual buffer space are allocated together.
	void* operator new(size_t size, int bufferSize, bool unicode);
	
	inline void Reference()
		{ if(this) ++mRefcount; }
	inline void UnReference()
		{ if(this) if(--mRefcount == 0) delete this; }
	inline bool Unique() const
		{ if(this) return mRefcount == 1; else return false; }
	
	Buffer* Copy() const;
	void Assign(const wchar_t* data, int length);
	void Assign(const char* data, int length);
	Buffer* Resize(int newSize);
	
	inline const wchar_t* UnicodeData() const
	{
		if(this)
			return (wchar_t*)(this + 1);
		else
			return (wchar_t*)"\0\0\0\0";
	}
	inline wchar_t* UnicodeData()
	{
		if(this)
			return (wchar_t*)(this + 1);
		else
			return (wchar_t*)"\0\0\0\0";
	}
	inline const char* AsciiData() const
	{
		if(this)
			return (char*)(this + 1);
		else
			return "";
	}
	inline char* AsciiData()
	{
		if(this)
			return (char*)(this + 1);
		else
			return "";
	}
	
	//Terminate() adds a null terminator to the string.  There is always space
	// to do this because the buffer is allocated with a size one character
	// larger than whan mBufferSize says.
	inline void Terminate() const
	{
		if(!this) return;
		if(mUnicode)
			((wchar_t*)(this + 1))[mLength] = 0;
		else
			((char*)(this + 1))[mLength] = 0;
	}
	
	//BufferSize() returns the size of the buffer in characters.  This does not
	// include any sort of null terminator.
	inline int BufferSize() const   {if(this) return mBufferSize; else return 0;}
	
	//Length() returns the number of characters in the string.
	inline int Length() const      {if(this) return mLength; else return 0;}
	inline void Length(int length) {if(this) mLength = length;}
	
	bool IsUnicode() {if(this) return mUnicode; return false;}
	
	inline wchar_t operator[](int index) const
	{
		if(mUnicode)
			return UnicodeData()[index];
		else
			return AsciiData()[index];
	}
};

GString::Buffer::Buffer()
	: mRefcount(0)
{
}

GString::Buffer::~Buffer()
{
}

//The buffer object and the actual buffer space are allocated together.
void*
GString::Buffer::operator new(size_t size, int bufferSize, bool unicode)
{
	Buffer* result;
	
	if(unicode)
		result = (Buffer*)GMalloc(size + (bufferSize + 1) * sizeof(wchar_t));
	else
		result = (Buffer*)GMalloc(size + (bufferSize + 1) * sizeof(char));
	
	result->mUnicode = unicode;
	result->mBufferSize = bufferSize;
	return result;
}

GString::Buffer*
GString::Buffer::Copy() const
{
	if(this)
	{
		Buffer* result = new(mLength, mUnicode) Buffer;
		result->mLength = mLength;
		if(mUnicode)
			memcpy(result->UnicodeData(), UnicodeData(), mLength * sizeof(wchar_t));
		else
			memcpy(result->AsciiData(), AsciiData(), mLength * sizeof(char));
		return result;
	}
	else
		return NULL;
}

void
GString::Buffer::Assign(const wchar_t* data, int length)
{
	if(this)
	{
		mLength = length;
		
		if(mLength >= mBufferSize)
			mLength = mBufferSize;
		
		memcpy(UnicodeData(), data, mLength * sizeof(wchar_t));
	}
}

void
GString::Buffer::Assign(const char* data, int length)
{
	if(this)
	{
		mLength = length;
		
		if(mLength >= mBufferSize)
			mLength = mBufferSize;
		
		memcpy(AsciiData(), data, mLength * sizeof(char));
	}
}

GString::Buffer*
GString::Buffer::Resize(int newSize)
{
	if(this)
	{
		if((mBufferSize >= newSize)
		&& (mBufferSize <= newSize * 2))
			return this;
		
		int size;
		
		//If the user is adding lots of small things to the string (a character
		// at a time, for instance), we don't want to re-allocate the buffer for
		// every bit added.  So, we double the size each time the buffer runs
		// out.
		if((mBufferSize < newSize)
		&& (mBufferSize * 2 >= newSize))
		{
			size = mBufferSize * 2;
		}
		else
			size = newSize;  //double size was not big enough.
		
		Buffer* result = new(size, mUnicode) Buffer;
		if(mUnicode)
			result->Assign(UnicodeData(), mLength);
		else
			result->Assign(AsciiData(), mLength);
		return result;
	}
	else
	{
		//this is an empty string
		Buffer* result = new(newSize, false) Buffer;
		result->Length(0);
		return result;
	}
}

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GString::GString()
{
	mBuffer = NULL;
	mBuffer->Reference();
}

GString::GString(const GString& other)
{
	mBuffer = other.mBuffer;
	mBuffer->Reference();
}

GString::GString(const char    cstr[])
{
	if(cstr)
	{
		int length;
		length = strlen(cstr);
		mBuffer = new(length, false) Buffer;
		mBuffer->Reference();
		mBuffer->Assign(cstr, length);
	}
	else
	{
		mBuffer = NULL;
	}
}

GString::GString(const wchar_t cstr[])
{
	if(cstr)
	{
		int length;
		length = wcslen(cstr);
		mBuffer = new(length, true) Buffer;
		mBuffer->Reference();
		mBuffer->Assign(cstr, length);
	}
	else
	{
		mBuffer = NULL;
	}
}

GString::GString(const char    cstr[], int length)
{
	if(cstr && length)
	{
		mBuffer = new(length, false) Buffer;
		mBuffer->Reference();
		mBuffer->Assign(cstr, length);
	}
	else
	{
		mBuffer = NULL;
	}
}

GString::GString(const wchar_t cstr[], int length)
{
	if(cstr)
	{
		mBuffer = new(length, true) Buffer;
		mBuffer->Reference();
		mBuffer->Assign(cstr, length);
	}
	else
	{
		mBuffer = NULL;
	}
}

GString::GString(int i)
{
	char number[128];
	//FIXME:  swprintf?  But it doesn't appear to be implemented in glibc.
	int length = sprintf(number, "%i", i);
	
	mBuffer = new(length, false) Buffer;
	mBuffer->Reference();
	mBuffer->Assign(number, length);
}

GString::GString(double f)
{
	char number[128];
	//FIXME:  swprintf?  But it doesn't appear to be implemented in glibc.
	int length = sprintf(number, "%f", f);
	
	mBuffer = new(length, false) Buffer;
	mBuffer->Reference();
	mBuffer->Assign(number, length);
}

GString::~GString()
{
	mBuffer->UnReference();
}

// operations --------------------------------------------------------
int
GString::Compare(const GString& other) const
{
	mBuffer->Terminate();
	other.mBuffer->Terminate();
	
	if(!mBuffer->IsUnicode() && !other.mBuffer->IsUnicode())
		return strcmp(mBuffer->AsciiData(), other.mBuffer->AsciiData());
	else if(mBuffer->IsUnicode() && other.mBuffer->IsUnicode())
		return wcscmp(mBuffer->UnicodeData(), other.mBuffer->UnicodeData());
	else
		return (this->ToUnicode()).Compare(other.ToUnicode());
}

int
GString::Compare(const wchar_t other[]) const
{
	GString str = ToUnicode();
	str.mBuffer->Terminate();
	return wcscmp(str.mBuffer->UnicodeData(), other);
}

int
GString::Compare(const char other[]) const
{
	if(mBuffer->IsUnicode())
	{
		mBuffer->Terminate();
		
		int i = 0;
		const wchar_t* data = mBuffer->UnicodeData();
		
		while(true)
		{
			if(data[i] != other[i])
			{
				if(data[i] < other[i])
					return -1;
				if(data[i] > other[i])
					return 1;
			}
			
			if(data[i] == 0)
				return 0;
			i++;
		}
	}
	else
	{
		mBuffer->Terminate();
		return strcmp(mBuffer->AsciiData(), other);
	}
}

int
GString::FindFirstOf(wchar_t c) const
{
	for(int i = 0; i < mBuffer->Length(); i++)
		if((*mBuffer)[i] == c)
			return i;
	return -1;
}

int
GString::FindLastOf(wchar_t c) const
{
	for(int i = mBuffer->Length() - 1; i >= 0; i--)
		if((*mBuffer)[i] == c)
			return i;
	return -1;
}

int
GString::FindNext(wchar_t c, int start) const
{
	for(int i = start; i < mBuffer->Length(); i++)
		if((*mBuffer)[i] == c)
			return i;
	return -1;
}

GString
GString::SubString(int start) const
{
	if(mBuffer->IsUnicode())
	{
		if((start < mBuffer->Length())
		&& (start >= 0))
			return GString(mBuffer->UnicodeData() + start, mBuffer->Length() - start);
		else
			return GString();
	}
	else
	{
		if((start < mBuffer->Length())
		&& (start >= 0))
			return GString(mBuffer->AsciiData() + start, mBuffer->Length() - start);
		else
			return GString();
	}
}

GString
GString::SubString(int start, int length) const
{
	if((start < mBuffer->Length())
	&& (start >= 0))
	{
		if(mBuffer->IsUnicode())
		{
			if((start + length) > mBuffer->Length())
				return GString(mBuffer->UnicodeData() + start, mBuffer->Length() - start);
			else
				return GString(mBuffer->UnicodeData() + start, length);
		}
		else
		{
			if((start + length) > mBuffer->Length())
				return GString(mBuffer->AsciiData() + start, mBuffer->Length() - start);
			else
				return GString(mBuffer->AsciiData() + start, length);
		}
	}
	else
		return GString();
}

// access ------------------------------------------------------------
int
GString::Length() const
{
	return mBuffer->Length();
}

void
GString::SetBuffer(Buffer* buffer)
{
	if(buffer != mBuffer)
	{
		if(mBuffer)
			mBuffer->UnReference();
		mBuffer = buffer;
		if(mBuffer)
			mBuffer->Reference();
	}
}

bool
GString::IsUnicode() const
{
	return mBuffer->IsUnicode();
}

GString
GString::ToUnicode() const
{
	if(mBuffer->IsUnicode())
		return *this;
	
	Buffer* newBuffer = new(mBuffer->Length(), true) Buffer;
	
	int j = mBuffer->Length();
	for(int i = 0; i < j; i++)
		newBuffer->UnicodeData()[i] = mBuffer->AsciiData()[i];
	newBuffer->Length(mBuffer->Length());
	
	GString ret;
	ret.SetBuffer(newBuffer);
	return ret;
}

GString
GString::ToAscii() const
{
	if(!mBuffer->IsUnicode())
		return *this;
	
	Buffer* newBuffer = new(mBuffer->Length(), false) Buffer;
	
	int j = mBuffer->Length();
	for(int i = 0; i < j; i++)
		newBuffer->AsciiData()[i] = mBuffer->UnicodeData()[i];
	newBuffer->Length(mBuffer->Length());
	
	GString ret;
	ret.SetBuffer(newBuffer);
	return ret;
}

const char*
GString::AsciiData() const
{
	mBuffer->Terminate();
	return mBuffer->AsciiData();
}

const wchar_t*
GString::UnicodeData() const
{
	mBuffer->Terminate();
	return mBuffer->UnicodeData();
}
	
void
GString::SetChar(int index, wchar_t value)
{
	if(mBuffer->IsUnicode())
		mBuffer->UnicodeData()[index] = value;
	else
		mBuffer->AsciiData()[index] = value;
}
	
// operators ---------------------------------------------------------
GString::operator int()
{
	//TODO:  This could probably be done better...
	GString ascii = ToAscii();
	
	return strtol(ascii.AsciiData(), NULL, 10);
}

GString::operator float()
{
	//TODO:  This could probably be done better...
	GString ascii = ToAscii();
	
	return strtod(ascii.AsciiData(), NULL);
}

wchar_t
GString::operator[](int index) const
{
	return (*mBuffer)[index];
}

const GString&
GString::operator=(const GString& other)
{
	SetBuffer(other.mBuffer);
	return *this;
}

GString
GString::operator+(const GString& other) const
{
	GString result(*this);
	result += other;
	return result;
}

const GString&
GString::operator+=(const GString& other)
{
	if(!mBuffer->IsUnicode() && !other.mBuffer->IsUnicode())
	{
		//sum will be ASCII
		int newSize = mBuffer->Length() + other.mBuffer->Length();
		
		if(!mBuffer->Unique())
			SetBuffer(mBuffer->Copy());
		SetBuffer(mBuffer->Resize(newSize));
		
		char* data = mBuffer->AsciiData();
		strncpy(data + mBuffer->Length(), other.mBuffer->AsciiData(), other.mBuffer->Length());
		mBuffer->Length(newSize);
		return *this;
	}
	else
	{
		if(!mBuffer->IsUnicode())
			*this = ToUnicode();
		GString str2 = other.ToUnicode();
		
		int newSize = mBuffer->Length() + str2.mBuffer->Length();
		
		if(!mBuffer->Unique())
			SetBuffer(mBuffer->Copy());
		SetBuffer(mBuffer->Resize(newSize));
		
		wchar_t* data = mBuffer->UnicodeData();
		wcsncpy(data + mBuffer->Length(), str2.mBuffer->UnicodeData(), str2.mBuffer->Length());
		mBuffer->Length(newSize);
		return *this;
	}
}

GString
GString::operator+(char other) const
{
	GString result(*this);
	result += other;
	return result;
}

const GString&
GString::operator+=(char other)
{
	if(!mBuffer->IsUnicode())
	{
		int newSize = mBuffer->Length() + 1;
		
		if(!mBuffer->Unique())
			SetBuffer(mBuffer->Copy());
		SetBuffer(mBuffer->Resize(newSize));
		
		char* data = mBuffer->AsciiData();
		data[mBuffer->Length()] = other;
		mBuffer->Length(mBuffer->Length() + 1);
		return *this;
	}
	else
	{
		int newSize = mBuffer->Length() + 1;
		
		if(!mBuffer->Unique())
			SetBuffer(mBuffer->Copy());
		SetBuffer(mBuffer->Resize(newSize));
		
		wchar_t* data = mBuffer->UnicodeData();
		data[mBuffer->Length()] = other;
		mBuffer->Length(mBuffer->Length() + 1);
		return *this;
	}
}

GString
GString::operator+(char other[]) const
{
	GString result(*this);
	result += other;
	return result;
}

const GString&
GString::operator+=(char other[])
{
	if(mBuffer->IsUnicode())
	{
		return (*this += GString(other).ToUnicode());
	}
	else
	{
		int otherlen = strlen(other);
		int newSize = mBuffer->Length() + otherlen;
		
		if(!mBuffer->Unique())
			SetBuffer(mBuffer->Copy());
		SetBuffer(mBuffer->Resize(newSize));
		
		char* data = mBuffer->AsciiData() + mBuffer->Length();
		strncpy(data, other, otherlen);
		mBuffer->Length(newSize);
		return *this;
	}
}

GString
GString::operator+(wchar_t other[]) const
{
	GString result(*this);
	result += other;
	return result;
}

const GString&
GString::operator+=(wchar_t other[])
{
	if(!mBuffer->IsUnicode())
		*this = ToUnicode();
	
	int otherlen = wcslen(other);
	int newSize = mBuffer->Length() + otherlen;
	
	if(!mBuffer->Unique())
		SetBuffer(mBuffer->Copy());
	SetBuffer(mBuffer->Resize(newSize));
	
	wcsncpy(mBuffer->UnicodeData() + mBuffer->Length(), other, otherlen);
	mBuffer->Length(newSize);
	return *this;
}

GString operator+(const char str1[], const GString& str2)
{
	GString result(str1);
	result += str2;
	return result;
}

GString operator+(const wchar_t str1[], const GString& str2)
{
	GString result(str1);
	result += str2;
	return result;
}

} //namespace GAUGE3D
