/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_FILES_FILE_H__INCLUDED__
#define __GAUGE3D_FILES_FILE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include "gauge3d/common/gstring.h"
#endif //redundant include guard

// local package -----------------------------------------------------
#ifndef __GAUGE3D_FILES_ISTREAM_H__INCLUDED__
#include "istream.h"
#endif //redundant include guard

#ifndef __GAUGE3D_FILES_OSTREAM_H__INCLUDED__
#include "ostream.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GFile file.h gauge3d/files/file.h
 *  \ingroup files
 *  \brief Represents a block of data.
 *
 *  This could be a file on the disk, a file in a compressed archive,
 *  a file on an ftp server, etc. */
class GAUGE3D_API GFile: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f35);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GFile();
	
	// operations ------------------------------------------------------
	//! Gets the size of the file in bytes.
	virtual int Size() = 0;
	
	//! Truncates the file to zero length.  Returns false if file was not writable.
	virtual bool Truncate() = 0;
	
	/*! \brief Get a GIStream for the file.  May be null if the file is not readable.
	 *  \remarks This just calls GetRawInputStream() and wraps a GIStream around it. */
	inline pGIStream GetInputStream()
	{
		pGRawIStream src = GetRawInputStream();
		if(src == NULL)
			return NULL;
		return new GIStream(src);
	}
	/*! \brief Get a GOStream for the file.  May be null if the file is not readable.
	 *  \remarks This just calls GetRawOutputStream() and wraps a GOStream around it. */
	inline pGOStream GetOutputStream()
	{
		pGRawOStream src = GetRawOutputStream();
		if(src == NULL)
			return NULL;
		return new GOStream(src);
	}
	
	//! Get a GRawIStream for the file.  May be null if the file is not readable.
	virtual pGRawIStream GetRawInputStream() = 0;
	//! Get a GRawOStream for the file.  May be null if the file is not writeable.
	virtual pGRawOStream GetRawOutputStream() = 0;
	
	//! Flags for mapping a file to memory.
	enum tMapFlags
	{
		READ    = 1 << 0,   //!< Make the map readable
		WRITE   = 1 << 1,   //!< Make it writable
	};
	/*! \class FileMap file.h gauge3d/files/file.h
	 *  \brief Represents a GFile mapped to memory. */
	class FileMap
	{
		void* mMap;
		
	public:
		//! Initialize the file map.
		inline FileMap(void* map)
			: mMap(map) {}
		
		virtual ~FileMap();
		
		//! Get the location of the map in merory.
		inline void* Map() {return mMap;}
	};
	typedef GSmartPointer<FileMap> pFileMap;
	
	/*! \brief Maps a portion of the file to memory.
	 *  \param offset Where to start the map relative to the beginning of the file.
	 *  \param size   How many bytes to map to memory.
	 *  \param flags  One or more of tMapFlags, bitwise or'd.
	 *  \returns      A pointer to a FileMap class describing the map, or NULL if
	 *                the file could not be mapped. */
	virtual pFileMap Map(int offset, int size, int flags) = 0;
	/*! \brief Maps the entire file to memory.
	 *  \param flags  One or more of tMapFlags, bitwise or'd.
	 *  \returns      A pointer to a FileMap class describing the map, or NULL if
	 *                the file could not be mapped. */
	inline pFileMap Map(int flags = READ) {return Map(0, Size(), flags);}
	
	/*! \brief Copies the file to a temporary file on disk, if it is not already on the disk.
	 *  \returns The path and name of the file on disk.
	 *
	 *  You should not write to the file on disk as other parts of the engine may
	 *  be reading from it as well.  The file on disk will be deleted when the GFile
	 *  that created it is deleted (unless the file was on the disk to begin with).
	 *
	 *  \remarks Do not use this if you can avoid it.  It is mainly meant for use
	 *           with GLibrary, which must load shared libraries from disk due to
	 *           the limitations of the library loading functions of every operating
	 *           system I know of. */
	virtual GString MapToDisk() = 0;
	
	/*! \brief Get the name of the file.
	 *  \note This is not always the same as the name of the directory entry that
	 *        points to the file, especially if the file is in a virtual filesystem. */
	virtual GString Name() = 0;
	
	/*! \brief Create a string containing the complete file.
	 *
	 *  \remarks This function pretty much just calls Map and makes a string out
	 *           of the results.  Do not use this on large files.  Note that
	 *           derived classes may define their own implementation of this, but
	 *           they are not required to. */
	virtual operator GString();
};
/*! \typedef pGFile
 *  \ingroup files
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GFile GFile\endlink. */
typedef GSmartPointer<GFile> pGFile;

} // namespace GAUGE3D

#endif //included
