/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "istream.h"

// system ------------------------------------------------------------
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

#ifdef PLATFORM_BIG_ENDIAN
const GIStream::tEndianness HOST_ENDIANNESS = GIStream::BIG;
#else
const GIStream::tEndianness HOST_ENDIANNESS = GIStream::LITTLE;
#endif

// class state definition ============================================
struct GIStream::Private
{
	enum
	{
		DEFAULT_BUFFER_SIZE = 1024,   //fixme:  what is a good size?
	};
	
	// state variables -------------------------------------------------
	GIStream* mThis;
	pGRawIStream mSource;
	
	char* mBuffer;
	int mBufferSize;
	int mBufferValid;
	int mBufferPos;
	
	tEndianness mEndianness;
	
	bool mBinary;
	bool mLastRead;  //whether or not last read operation succeeded.
	
	tNumberBase mNumberBase;
	
	// private functions -----------------------------------------------
	void Underflow();
	
	int BinaryRead(void* target, int num, int elementSize);
	
	void SkipWhiteSpace();
	bool IsDigit(char c);
	ulong ReadTextUInt(bool* error);
	long ReadTextInt(bool* error);
	double ReadTextFloat(bool* error);
	
	//My pretty member templates are not supported on MSVC!  >:(
//	template<class tType>
//	int ReadUInt(tType* target, int num);
//	template<class tType>
//	int ReadInt(tType* target, int num);
//	template<class tType>
//	int ReadFloat(tType* target, int num);
	
	//these two functions manage scancodes for string reading.
	//the bitfield contains eight 32-bit integers, which amounts
	// to 256 bits, one for each character.  Each bit specifies
	// whether or not the character was included in the scancode.
	void InitScancode(uint32 bitfield[], char* scancode);
	bool IsInScancode(uint32 bitfield[], char c);
};

void
GIStream::Private::Underflow()
{
	mBufferValid = mSource->Read(mBuffer, mBufferSize);
	mBufferPos = 0;
}

int
GIStream::Private::BinaryRead(void* target, int num, int elementSize)
{
	int elementsRead;

	if((mEndianness == HOST_ENDIANNESS)
	|| (elementSize == 1))
	{
		//no need to swap bytes, so we can read faster!
		//we no longer care about element size, so...
		num *= elementSize;
		int bytesRead;
		
		if(mBufferSize == 0)
		{
			//no buffer.  Just read it.
			bytesRead = mSource->Read((char*)target, num);
		}
		else if(num > (mBufferValid - mBufferPos))
		{
			//need to read more than just what is in the buffer.
			
			if(!mSource->InputAvailable())
			{
				//doh, the buffer is all we have left.  Read it.			
				bytesRead = mBufferValid - mBufferPos;
				memcpy(target, mBuffer + mBufferPos, bytesRead);
				mBufferPos = mBufferValid;
			}
			else if(num > (mBufferSize + mBufferValid - mBufferPos))
			{
				//really really long read needed.
				//First, copy the entire buffer in.
				bytesRead = mBufferValid - mBufferPos;
				memcpy(target, mBuffer + mBufferPos, bytesRead);
				mBufferValid = 0;
				mBufferPos = 0;
				
				//then, finish it off with a direct read.
				bytesRead += mSource->Read(((char*)target) + bytesRead, num - bytesRead);
			}
			else
			{
				//only one buffer's length extra needed.
				//First, copy the entire buffer in.
				bytesRead = mBufferValid - mBufferPos;
				memcpy(target, mBuffer + mBufferPos, bytesRead);
				
				//then, refill the buffer
				Underflow();
				
				//then, read in the rest from the new buffer.
				if(num - bytesRead > mBufferValid)
				{
					//not enough input to fill the request.  Oh well.
					memcpy((char*)target + bytesRead, mBuffer, mBufferValid);
					bytesRead += mBufferValid;
					mBufferPos = mBufferValid;
				}
				else
				{
					//all is good
					memcpy((char*)target + bytesRead, mBuffer, num - bytesRead);
					mBufferPos = num - bytesRead;
					bytesRead = num;
				}
			}
		}
		else
		{
			//what is in the buffer is sufficient.
			memcpy(target, mBuffer + mBufferPos, num);
			bytesRead = num;
			mBufferPos += num;
		}
		
		elementsRead = bytesRead / elementSize;
	}
	else
	{
		//damnit, we need to swap each element!
		char unswapped[20];  //MSVS does not allow variable-sized arrays on the stack!?
		                     //AAAAAAAAAAAAAAAAARRRRRRRRRRRRRRRRRRRGGGGGGGGGGGGGGGG!!!
		char* swapped = (char*)target;
		
		elementsRead = 0;
		while(elementsRead < num)
		{
			if(BinaryRead(unswapped, elementSize, 1) != elementSize)
				break;
			
			//swap the bytes
			for(int i = 0; i < elementSize; i++)
				swapped[i] = unswapped[elementSize - i - 1];
			
			//incerement swapped to point at the next element
			swapped += elementSize;
			
			++elementsRead;
		}
	}
	
	return elementsRead;
}

void
GIStream::Private::SkipWhiteSpace()
{
	char c;
	c = mThis->Peek();
	while(isspace(c))
	{
		mThis->Get();
		c = mThis->Peek();
	}
}

bool
GIStream::Private::IsDigit(char c)
{
	switch(mNumberBase)
	{
	case OCT:
		if('0' <= c && c <= '7')
			return true;
		return false;
	case DEC:
		if('0' <= c && c <= '9')
			return true;
		return false;
	case HEX:
		if(isxdigit(c))
			return true;
		return false;
	}
	
	return false;
}

ulong
GIStream::Private::ReadTextUInt(bool* error)
{
	char buffer[20];
	int bufferPos = 0;
	
	*error = true;      //true until a digit is read
	
	SkipWhiteSpace();
	
	buffer[0] = mThis->Peek();
	if(mNumberBase == HEX)
	{
		//hex can start with 0x
		if(buffer[0] == '0')
		{
			*error = false;
			mThis->Get();
			buffer[0] = mThis->Peek();
			if(buffer[0] == 'x')
			{
				mThis->Get();
				buffer[0] = mThis->Peek();
			}
		}
	}
	
	if(IsDigit(buffer[bufferPos]))
	{
		*error = false;
		
		//now read in the number
		do
		{
			mThis->Get();
			++bufferPos;
			buffer[bufferPos] = mThis->Peek();
		} while(IsDigit(buffer[bufferPos]));
	}
	
	//zero-terminate
	buffer[bufferPos] = 0;
	
	//and convert
	if(bufferPos == 0)
		return 0;
	else
		return strtoul(buffer, NULL, mNumberBase);
}

long
GIStream::Private::ReadTextInt(bool* error)
{
	char buffer[20];
	int bufferPos = 0;
	
	*error = true;      //true until a digit is read
	
	SkipWhiteSpace();
	
	buffer[0] = mThis->Peek();
	
	//check if first character is a + or -.  Accept it if so.
	if(buffer[0] == '-' || buffer[0] == '+')
	{
		mThis->Get();
		buffer[1] = mThis->Peek();
		bufferPos = 1;
	}
	
	if(mNumberBase == HEX)
	{
		//hex can start with 0x.  Ignore it.
		if(buffer[bufferPos] == '0')
		{
			*error = false;
			mThis->Get();
			buffer[bufferPos] = mThis->Peek();
			if(buffer[bufferPos] == 'x')
			{
				mThis->Get();
				buffer[bufferPos] = mThis->Peek();
			}
		}
	}
	
	if(IsDigit(buffer[bufferPos]))
	{
		*error = false;
		
		//now read in the number
		do
		{
			mThis->Get();
			++bufferPos;
			buffer[bufferPos] = mThis->Peek();
		} while(IsDigit(buffer[bufferPos]));
	}
	
	//zero-terminate
	buffer[bufferPos] = 0;
	
	//and convert
	if(bufferPos == 0)
		return 0;
	else
		return strtol(buffer, NULL, mNumberBase);
}

double
GIStream::Private::ReadTextFloat(bool* error)
{
	char buffer[20];
	int bufferPos = 0;
	bool foundPeriod = false;    //set true after the decimal point is read
	bool foundExponent = false;  //set true after 'e' or 'E' is read
	
	*error = true;      //true until a digit is read
	
	SkipWhiteSpace();
	
	buffer[0] = mThis->Peek();
	
	//check if first character is a + or -.  Accept it if so.
	if(buffer[0] == '-' || buffer[0] == '+')
	{
		mThis->Get();
		buffer[1] = mThis->Peek();
		bufferPos = 1;
	}
	
	if('0' <= buffer[bufferPos] && buffer[bufferPos] <= '9')
	{
		*error = false;
		
		//now read in the number
		while(('0' <= buffer[bufferPos] && buffer[bufferPos] <= '9')
		   || (buffer[bufferPos] == '.')
		   || (buffer[bufferPos] == 'e')
		   || (buffer[bufferPos] == 'E'))
		{
			if(buffer[bufferPos] == '.')
			{
				if(foundPeriod)
					break;        //only read one decimal point
				if(foundExponent)
					break;        //decimal point should come before exponent
				foundPeriod = true;
			}
			
			if((buffer[bufferPos] == 'e')
			|| (buffer[bufferPos] == 'E'))
			{
				if(foundExponent)
					break;        //only read one exponent
				foundExponent = true;
				
				mThis->Get();
				++bufferPos;
				buffer[bufferPos] = mThis->Peek();
				
				//Exponents can have a sign.
				if(buffer[bufferPos] == '-' || buffer[bufferPos] == '+')
				{
					mThis->Get();
					++bufferPos;
					buffer[bufferPos] = mThis->Peek();
				}
			}
			else
			{
				mThis->Get();
				++bufferPos;
				buffer[bufferPos] = mThis->Peek();
			}
		}
	}
	
	//zero-terminate
	buffer[bufferPos] = 0;
	
	//and convert
	if(bufferPos == 0)
		return 0;
	else
		return strtod(buffer, NULL);
}

//MSVC does not support member templates.
//How dumb is that!?
//template<class tType>
//int GIStream::Private::ReadUInt(tType* target, int num)
//{
#define READ_UINT_TEMPLATE(tType, target, num)\
	int numRead;\
	bool error;\
	\
	if(mPrivate.mBinary)\
	{\
		numRead = mPrivate.BinaryRead(target, num, sizeof(tType));\
	}\
	else\
	{\
		for(numRead = 0; numRead < num; numRead++)\
		{\
			target[numRead] = mPrivate.ReadTextUInt(&error);\
			if(error)\
				break;\
		}\
	}\
	\
	if(numRead == 0)\
		mPrivate.mLastRead = false;\
	else\
		mPrivate.mLastRead = true;\
	\
	return numRead;
//}

//template<class tType>
//int GIStream::Private::ReadInt(tType* target, int num)
//{
#define READ_INT_TEMPLATE(tType, target, num)\
	int numRead;\
	bool error;\
	\
	if(mPrivate.mBinary)\
	{\
		numRead = mPrivate.BinaryRead(target, num, sizeof(tType));\
	}\
	else\
	{\
		for(numRead = 0; numRead < num; numRead++)\
		{\
			target[numRead] = mPrivate.ReadTextInt(&error);\
			if(error)\
				break;\
		}\
	}\
	\
	if(numRead == 0)\
		mPrivate.mLastRead = false;\
	else\
		mPrivate.mLastRead = true;\
	\
	return numRead;
//}

//template<class tType>
//int GIStream::Private::ReadFloat(tType* target, int num)
//{
#define READ_FLOAT_TEMPLATE(tType, target, num)\
	int numRead;\
	bool error;\
	\
	if(mPrivate.mBinary)\
	{\
		numRead = mPrivate.BinaryRead(target, num, sizeof(tType));\
	}\
	else\
	{\
		for(numRead = 0; numRead < num; numRead++)\
		{\
			target[numRead] = mPrivate.ReadTextFloat(&error);\
			if(error)\
				break;\
		}\
	}\
	\
	if(numRead == 0)\
		mPrivate.mLastRead = false;\
	else\
		mPrivate.mLastRead = true;\
	\
	return numRead;
//}

//these two functions manage scancodes for string reading.
//the bitfield contains eight 32-bit integers, which amounts
// to 256 bits, one for each character.  Each bit specifies
// whether or not the character was included in the scancode.
void
GIStream::Private::InitScancode(uint32 bitfield[], char* scancode)
{
	bool invert = false;
	int pos = 0;
	
	if(scancode == NULL)
		scancode = "^\f\n\r\t\v ";  //everything except white space
	
	for(int i = 0; i < 8; i++)
		bitfield[i] = 0;
	
	if(scancode[0] == '^')
	{
		invert = true;
		pos = 1;
	}
	
	while(scancode[pos] != 0)
	{
		if((scancode[pos] == '-') && (scancode[pos + 1] != 0))
		{
			//hyphens indicate a range of characters.
			//so, add all characters in the range to the list.
			for(char c = scancode[pos - 1]; c <= scancode[pos + 1]; c++)
				bitfield[(uint8)c / 32] |= 1 << ((uint8)c % 32);
			pos++;    //incement one extra time.
		}
		else
		{
			//add the character
			bitfield[(uint8)scancode[pos] / 32] |= 1 << ((uint8)scancode[pos] % 32);
		}
		pos++;
	}
	
	if(invert)
	{
		for(int i = 0; i < 8; i++)
			bitfield[i] = ~bitfield[i];
	}
	
	bitfield[0] &= ~1; //never include `\0` in a scancode
}

bool
GIStream::Private::IsInScancode(uint32 bitfield[], char c)
{
	if(bitfield[(uint8)c / 32] &
	   (1 << ((uint8)c % 32)))
		return true;
	return false;
}


// class member functions ============================================
// lifecycle ---------------------------------------------------------
GIStream::GIStream(pGRawIStream streamSource)
	: mPrivate(*(new Private))
{
	mPrivate.mThis        = this;
	mPrivate.mSource      = streamSource;
	
	mPrivate.mBufferSize  = Private::DEFAULT_BUFFER_SIZE;
	mPrivate.mBuffer      = new char[mPrivate.mBufferSize];
	mPrivate.mBufferValid = 0;
	mPrivate.mBufferPos   = 0;
	
	mPrivate.mEndianness  = BIG;
	mPrivate.mBinary      = true;
	mPrivate.mNumberBase  = DEC;
	mPrivate.mLastRead    = true;
}

GIStream::GIStream(pGRawIStream streamSource, int bufferSize)
	: mPrivate(*(new Private))
{
	mPrivate.mThis = this;
	mPrivate.mSource = streamSource;
	
	mPrivate.mBufferSize = bufferSize;
	if(bufferSize != 0)
		mPrivate.mBuffer = new char[bufferSize];
	mPrivate.mBufferValid = 0;
	mPrivate.mBufferPos = 0;
	
	mPrivate.mEndianness  = BIG;
	mPrivate.mBinary = true;
	mPrivate.mNumberBase = DEC;
	mPrivate.mLastRead = true;
}

GIStream::~GIStream()
{
	if(mPrivate.mBufferSize != 0)
		delete [] mPrivate.mBuffer;
	delete &mPrivate;
}

// operations --------------------------------------------------------
void
GIStream::Seek(int pos, tSeekType seekType)
{
	if(seekType == CURRENT)
	{
		if((mPrivate.mBufferPos + pos >= 0)
		&& (mPrivate.mBufferPos + pos < mPrivate.mBufferValid))
		{
			//the seek position is within the current buffer.
			mPrivate.mBufferPos += pos;
			return;
		}
		
		//need to adjust pos for the stream source's position.
		pos -= mPrivate.mBufferValid - mPrivate.mBufferPos;
	}

	//invalidate the buffer and then seek the stream source
	mPrivate.mBufferValid = 0;
	mPrivate.mBufferPos = 0;
	mPrivate.mSource->Seek(pos, seekType);
}

int
GIStream::TellPos()
{
	return mPrivate.mSource->TellPos() - mPrivate.mBufferValid + mPrivate.mBufferPos;
}

int
GIStream::Read(uint64*  target, int num)
{
	//return mPrivate.ReadUInt(target, num);
	READ_UINT_TEMPLATE(uint64, target, num)
}
int
GIStream::Read(uint32*  target, int num)
{
//	return mPrivate.ReadUInt(target, num);
	READ_UINT_TEMPLATE(uint32, target, num)
}
int
GIStream::Read(uint16*  target, int num)
{
//	return mPrivate.ReadUInt(target, num);
	READ_UINT_TEMPLATE(uint16, target, num)
}

int
GIStream::Read(int64*   target, int num)
{
//	return mPrivate.ReadInt(target, num);
	READ_INT_TEMPLATE(int64, target, num)
}
int
GIStream::Read(int32*   target, int num)
{
//	return mPrivate.ReadInt(target, num);
	READ_INT_TEMPLATE(int32, target, num)
}
int
GIStream::Read(int16*   target, int num)
{
//	return mPrivate.ReadInt(target, num);
	READ_INT_TEMPLATE(int16, target, num)
}

int
GIStream::Read(float64* target, int num)
{
//	return mPrivate.ReadFloat(target, num);
	READ_FLOAT_TEMPLATE(float64, target, num)
}
int
GIStream::Read(float32* target, int num)
{
//	return mPrivate.ReadFloat(target, num);
	READ_FLOAT_TEMPLATE(float32, target, num)
}

int
GIStream::Read(char*    target, int num)
{
	int numRead;
	
	if(mPrivate.mBinary)
	{
		numRead = mPrivate.BinaryRead(target, num, sizeof(char));
	}
	else
	{
		for(numRead = 0; numRead < num; numRead++)
		{
			mPrivate.SkipWhiteSpace();
			target[numRead] = Get();
			if(!mPrivate.mLastRead)
				break;      //read failed
		}
	}
	
	if(numRead == 0)
		mPrivate.mLastRead = false;
	else
		mPrivate.mLastRead = true;
	
	return numRead;
}

//these two read strings.  scanCode specifies what characters to include
// in the string.  It has the same syntax as the scanf %[] conversion
// specifier, without the brackets.  So, "^\n" would say to stop when
// a newline is encountered.  By default, the set of all non-whitespace
// characters is used.
// Once a character not in the scancode is encountered, the string is
// complete.  If you are reading multiple strings in one call, the function
// will search for the next valid character and start again there.
//stringSize is the size in bytes of the (each) char array.  If the string
// read would be longer than this, the remainder is ignored.  The array will
// always be zero-terminated, however.
int
GIStream::Read(char* target[], int num, char* scanCode, int stringSize)
{
	uint32 scanBits[8];
	int numRead;
	int i;
	char c;
	
	mPrivate.InitScancode(scanBits, scanCode);
	
	for(numRead = 0; numRead < num; numRead++)
	{
		if(!InputAvailable())
			break;       //end of data stream, no more to read.
		
		//First, skip anything not in the scancode.
		while(true)
		{
			c = Peek();
			if(!mPrivate.mLastRead)
				break;       //peek failed
			if(mPrivate.IsInScancode(scanBits, c))
				break;
			Get();
		}
		
		for(i = 0; (i < stringSize - 1) || (stringSize == -1); i++)
		{
			c = Peek();
			if(!mPrivate.IsInScancode(scanBits, c))
				break;
			if(!mPrivate.mLastRead)
				break;       //peek failed
			Get();
			target[numRead][i] = c;
		}
		target[numRead][i] = 0;  //null-terminate
	}
	
	if(numRead == 0)
		mPrivate.mLastRead = false;
	else
		mPrivate.mLastRead = true;
	
	return numRead;
}

int
GIStream::Read(GString*  target, int num, char* scanCode)
{
	uint32 scanBits[8];
	int numRead;
	char c;
	
	mPrivate.InitScancode(scanBits, scanCode);
	
	for(numRead = 0; numRead < num; numRead++)
	{
		if(!InputAvailable())
			break;       //end of data stream, no more to read.
		
		//First, skip anything not in the scancode.
		while(true)
		{
			c = Peek();
			if(!mPrivate.mLastRead)
				break;       //peek failed
			if(mPrivate.IsInScancode(scanBits, c))
				break;
			Get();
		}
		
		//strings have variable size, so we will just keep reading
		// until we hit something not in the scancode or we run out
		// of data.
		target[numRead] = "";
		while(true)
		{
			c = Peek();
			if(!mPrivate.mLastRead)
				break;       //peek failed
			if(!mPrivate.IsInScancode(scanBits, c))
				break;
			Get();
			target[numRead] += c;
		}
	}
	
	if(numRead == 0)
		mPrivate.mLastRead = false;
	else
		mPrivate.mLastRead = true;
	
	return numRead;
}

//If you read void* (not recommended!), num is the number of bytes
// to read.
int
GIStream::Read(void* target, int num)
{
	return mPrivate.BinaryRead(target, num, 1);
}

//Peek returns the next character in the stream without extracting it,
// and Get returns the character and extracts it as well.
char
GIStream::Peek()
{
	char c;
	
#ifdef PLATFORM_WINDOWS
	//we need to ignore '\r' on Windows
	--mPrivate.mBufferPos;
	do {
	++mPrivate.mBufferPos;
#endif

	//check if buffer is empty
	if(mPrivate.mBufferPos == mPrivate.mBufferValid)
	{
		//empty.  check if we are at the end of the data stream.
		if(!mPrivate.mSource->InputAvailable())
		{
			//yep.  No more to read
			mPrivate.mLastRead = false;
			return 0;
		}
		
		//refill the buffer
		mPrivate.Underflow();
	}
	
	c = mPrivate.mBuffer[mPrivate.mBufferPos];
	
#ifdef PLATFORM_WINDOWS
	} while((c == '\r') && !mPrivate.mBinary);
#endif
	
	mPrivate.mLastRead = true;
	return c;
}

char
GIStream::Get()
{
	char c;
	
#ifdef PLATFORM_WINDOWS
	//we need to ignore '\r' on Windows
	do {
#endif

	//check if buffer is empty
	if(mPrivate.mBufferPos == mPrivate.mBufferValid)
	{
		//empty.  check if we are at the end of the data stream.
		if(!mPrivate.mSource->InputAvailable())
		{
			//yep.  No more to read
			mPrivate.mLastRead = false;
			return 0;
		}
		
		//refill the buffer
		mPrivate.Underflow();
	}
	
	c = mPrivate.mBuffer[mPrivate.mBufferPos];
	++mPrivate.mBufferPos;
	
#ifdef PLATFORM_WINDOWS
	} while((c == '\r') && !mPrivate.mBinary);
#endif
	
	mPrivate.mLastRead = true;
	return c;
}

// access ------------------------------------------------------------
void
GIStream::ReadMode(tReadMode mode)
{
	if(mode == BINARY)
		mPrivate.mBinary = true;
	else
		mPrivate.mBinary = false;
}

GIStream::tReadMode
GIStream::ReadMode()
{
	if(mPrivate.mBinary)
		return BINARY;
	else
		return ASCII;
}

//Endianness() tells the istream whether the input is big endian or
// little endian.  It will then convert the input as appropriate when
// you read it in.
void
GIStream::Endianness(tEndianness endianness)
{
	mPrivate.mEndianness = endianness;
}

GIStream::tEndianness
GIStream::Endianness()
{
	return mPrivate.mEndianness;
}

// inquiry -----------------------------------------------------------	
//InputAvailable returns true if there is input available now.
// This can be used to test for end-of-file.
bool
GIStream::InputAvailable()
{
	if(mPrivate.mBufferPos < mPrivate.mBufferValid)
		return true;
	if(mPrivate.mSource->InputAvailable())
		return true;
	return false;
}

//Using the stream as a truth value returns true if the last read
// operation succeeded.  This way, you can say something like
//   while(myStream >> data)
//     {...}
// to keep reading in and processing data until there is no longer data
// available.
GIStream::operator void*()
{
	if(mPrivate.mLastRead)
		return this;    //last read succeeded
	return NULL;       //last read failed
}


//GRawIStream member functions ====================================

GRawIStream::~GRawIStream() {}

void
GRawIStream::Seek(int pos, int seekType)
{
	return;
}
int
GRawIStream::TellPos()
{
	return 0;
}

//WaitForInupt waits for input to become available, and returns true if there is
// input.  If timeout is not 0, it specifies the maximum amount of time to wait.
bool
GRawIStream::WaitForInput(tTime timeout)
{
	return InputAvailable();
}

//InputAvailable returns true if there is input available now.
bool
GRawIStream::InputAvailable()
{
	return true;
}


GAUGE3D_DECLARE_INTERFACES(GIStream)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

GAUGE3D_DECLARE_INTERFACES(GRawIStream)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
