/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "virtualdirectory.h"

// system ------------------------------------------------------------
#include <map>
using namespace std;

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GVirtualDirectory::Private
{
	// state variables -------------------------------------------------
	typedef map<GString, pGFile > tFileMap;
	typedef map<GString, pGVirtualDirectory > tSubdirMap;
	tFileMap mFiles;
	tSubdirMap mSubdirs;
	
	// private functions -----------------------------------------------
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GVirtualDirectory::GVirtualDirectory()
	: mPrivate(*(new Private))
{
}

GVirtualDirectory::~GVirtualDirectory()
{
	delete &mPrivate;   //C++ will figure out the rest!  :)
}

// operations --------------------------------------------------------
GArray<GString>
GVirtualDirectory::ListFiles()
{
	GArray<GString> ret(mPrivate.mFiles.size());
	Private::tFileMap::iterator iter = mPrivate.mFiles.begin();
	
	for(int i = 0; i < ret.Size(); i++)
	{
		ret[i] = iter->first;
		++iter;
	}
	
	return ret;
}

GArray<GString>
GVirtualDirectory::ListSubdirs()
{
	GArray<GString> ret(mPrivate.mSubdirs.size());
	Private::tSubdirMap::iterator iter = mPrivate.mSubdirs.begin();
	
	for(int i = 0; i < ret.Size(); i++)
	{
		ret[i] = iter->first;
		++iter;
	}
	
	return ret;
}

//Use these to open the contents of a directory.  This way, if you have
// a pointer to the root directory of any filesystem, you can access
// everything in the system.
//Note that it is perfectly legal to say:
//   myDirectory.OpenFile("this/is/a/deeply/nested/file");
// Likewise with OpenSubdir().  You need not manually traverse the
// directory tree.
//(on the other hand,
//   myDirectory.OpenFile("this\\is\\a\\deeply\\nested\\file")
// will FAIL, miserably, so don't try it.)
pGFile
GVirtualDirectory::OpenFile(GString name)
{
	GString subdirName;
	Private::tSubdirMap::iterator subdirIter;
	
	//ignore leading slashes
	int i = 0;
	while(name[i] == '/')
		i++;
	
	if(i != 0)
		name = name.SubString(i);
	
	//if there is a slash in the name, the file is in a sub directory.
	int slashPos = name.FindFirstOf('/');
	if(slashPos != -1)
	{
		//found a slash.  Interpret everything before the slash as
		// a sub directory name, and pass the open request to it.
		subdirName = name.SubString(0, slashPos);
		subdirIter = mPrivate.mSubdirs.find(subdirName);
		if(subdirIter == mPrivate.mSubdirs.end())
			return NULL;          //no such sub directory.
		else
			return subdirIter->second->OpenFile(name.SubString(slashPos + 1));
	}
	
	//Not in a sub directory.  Check this directory.
	Private::tFileMap::iterator iter = mPrivate.mFiles.find(name);
	if(iter == mPrivate.mFiles.end())
		return NULL;      //not found
	else
		return iter->second;
}

pGDirectory
GVirtualDirectory::OpenSubdir(GString name)
{
	GString subdirName;
	Private::tSubdirMap::iterator subdirIter;
	
	//ignore leading slashes
	int i = 0;
	while(name[i] == '/')
		i++;
	
	if(i != 0)
		name = name.SubString(i);
	
	//if there is a slash in the name, the file is in a sub directory.
	int slashPos = name.FindFirstOf('/');
	if(slashPos != -1)
	{
		//found a slash.  Interpret everything before the slash as
		// a sub directory name, and pass the open request to it.
		subdirName = name.SubString(0, slashPos);
		subdirIter = mPrivate.mSubdirs.find(subdirName);
		if(subdirIter == mPrivate.mSubdirs.end())
			return NULL;          //no such sub directory.
		else
			return subdirIter->second->OpenSubdir(name.SubString(slashPos + 1));
	}
	
	//Not in a sub directory.  Check this directory.
	Private::tSubdirMap::iterator iter = mPrivate.mSubdirs.find(name);
	if(iter == mPrivate.mSubdirs.end())
		return NULL;
	else
		return iter->second;
}

//AddFile returns NULL if the file already exists, or could not be
// created.
pGFile
GVirtualDirectory::AddFile(GString name)
{
	//TODO:  Create a temporary file and add it?
	return NULL;
}
//AddSubdir returns NULL if the directory already exists, or could
// not be created.
pGDirectory
GVirtualDirectory::AddSubdir(GString name)
{
	return (mPrivate.mSubdirs[name] = new GVirtualDirectory);
}

//these two remove files and sub directories from the directory.
void
GVirtualDirectory::RemoveFile(GString name)
{
	mPrivate.mFiles.erase(name);
}
void
GVirtualDirectory::RemoveDirectory(GString name)
{
	mPrivate.mSubdirs.erase(name);
}

//Add stuff to the virtual directory.  The files/subdirs will be
// directly linked, so when people request them, the same pointer
// which you give to the virtual directory will be the one sent
// to the users.
void
GVirtualDirectory::AddFile(GString name, pGFile file)
{
	mPrivate.mFiles[name] = file;
}

void
GVirtualDirectory::AddSubdir(GString name, pGVirtualDirectory subdir)
{
	mPrivate.mSubdirs[name] = subdir;
}


GAUGE3D_DECLARE_INTERFACES(GVirtualDirectory)
	GAUGE3D_DECLARE_BASECLASS(GDirectory)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
