/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "messagequeue.h"

// system ------------------------------------------------------------
#include <queue>
#include <list>
using namespace std;

// external packages -------------------------------------------------
#include "gauge3d/scheduler.h"

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GMessageQueue::Private
{
	enum
	{
		//this could be 2 for all I care.  The important thing is that the
		// queue's DoFrame() is called after all input is enqueued.
		SCHED_PRIORITY = 0x1000
	};
	
	queue<pGMessage> mQueue;
	
	struct HandlerInfo
	{
		pHandler mHandler;
		float mPriority;
	};
	list<HandlerInfo> mHandlers;
	typedef list<HandlerInfo>::iterator tHandlerIterator;
};

// class member functions ============================================
// lifecycle -------------------------------------------------------
GMessageQueue::GMessageQueue()
	: mPrivate(*(new Private))
{
	gInputScheduler->CallEveryFrame(this, Private::SCHED_PRIORITY);
}
GMessageQueue::~GMessageQueue()
{
	gInputScheduler->StopCallingEveryFrame(this);
	delete &mPrivate;
}

// operations ------------------------------------------------------
void
GMessageQueue::EnqueueMessage(pGMessage message)
{
	mPrivate.mQueue.push(message);
}

void
GMessageQueue::DispatchMessage(pGMessage message)
{
	Private::tHandlerIterator i;
	
	//send to all handlers
	for(i = mPrivate.mHandlers.begin(); i != mPrivate.mHandlers.end(); i++)
	{
		if(i->mHandler->HandleMessage(message))
			break;   //message was handled, so stop loop
	}
}

void
GMessageQueue::DoFrame(tTime time, tTime frametime)
{
	//while the queue is not empty...
	while(!mPrivate.mQueue.empty())
	{
		//...handle the top message
		DispatchMessage(mPrivate.mQueue.front());
		//remove message from queue
		mPrivate.mQueue.pop();
	}
	
	//all messages handled
}

void
GMessageQueue::DiscardMessages()
{
//	mPrivate.mQueue.clear();    //empty the queue
	while(!mPrivate.mQueue.empty())
		mPrivate.mQueue.pop();
}

void
GMessageQueue::AddHandler(pHandler handler, float priority)
{
	Private::HandlerInfo info;
	
	info.mHandler = handler;
	info.mPriority = priority;
	
	Private::tHandlerIterator i;
	//insert the handler into the list, sorted by priority
	for(i = mPrivate.mHandlers.begin(); i != mPrivate.mHandlers.end(); i++)
	{
		if(priority < i->mPriority)
		{
			mPrivate.mHandlers.insert(i, info);
			return;
		}
	}
	//priority is last; insert at end.
	mPrivate.mHandlers.insert(i, info);
}

//! Remove a message handler from the list.
void
GMessageQueue::RemoveHandler(pHandler handler)
{
	Private::tHandlerIterator i;
	//insert the handler into the list, sorted by priority
	for(i = mPrivate.mHandlers.begin(); i != mPrivate.mHandlers.end(); i++)
	{
		if(i->mHandler == handler)
		{
			mPrivate.mHandlers.erase(i);
			return;
		}
	}
}

//! Remove all handlers from the message handler list.
void
GMessageQueue::ClearHandlers()
{
	mPrivate.mHandlers.clear();
}

pGMessageQueue gMainMessageQueue;

GAUGE3D_DECLARE_INTERFACES(GMessageQueue)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

GAUGE3D_DECLARE_INTERFACES(GMessageQueue::Handler)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
