/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_INPUT_MESSAGEQUEUE_H__INCLUDED__
#define __GAUGE3D_INPUT_MESSAGEQUEUE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_INPUT_MESSAGE_H__INCLUDED__
#include "message.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GMessageQueue messagequeue.h gauge3d/input/messagequeue.h
 *  \ingroup input
 *  \brief Facilitates the distribution of messages.
 *
 *  This class queues and distributes messages.  First, you give it pointers
 *  to one or more message handlers.  When a message is ready to be dispatched,
 *  it will be sent to each handler one by one until one of the handlers
 *  reports that it has handled the message. */
class GAUGE3D_API GMessageQueue: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b2c267);
public:
	/*! \class Handler messagequeue.h gauge3d/input/messagequeue.h
	 *  \brief Base class for message handlers. */
	class GAUGE3D_API Handler: public GObject
	{
		GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b2c272);
	public:
		/*! \brief Attempts to handle the message, and returns true if the
		 *         message was dealt with.
		 *
		 *  \returns True if the handler did something with the message, and
		 *           thus the message has served its purpose.  False if the
		 *           message was not dealt with and should be sent to other
		 *           handlers. */
		virtual bool HandleMessage(pGMessage message) = 0;
	};
	typedef GSmartPointer<Handler> pHandler;
	
	// lifecycle -------------------------------------------------------
	GMessageQueue();
	~GMessageQueue();
	
	// operations ------------------------------------------------------
	//! Enqueue a message.
	void EnqueueMessage(pGMessage message);
	//! Immediately dispatch a message, bypassing the queue.
	void DispatchMessage(pGMessage message);
	//! Discard all enqueued messages.
	void DiscardMessages();
	/*! \brief Add a handler to the list of handlers.
	 *  \param handler  A pointer to a Handler to add.
	 *  \param priority A priority for the handler, between 0.0 and 1.0.
	 *                  This determines the order in which the handlers
	 *                  will be called.  Lower numbers will be called
	 *                  first.  Do not set this to exactly 0.0 or 1.0
	 *                  unless you are doing something special (like
	 *                  creating an input filter).
	 *  \remarks You could have a handler that acts as an input filter by
	 *           giving it a priority of 0.0.  Have the handler return true
	 *           if the message should be ignored.  Since the priority is
	 *           0.0, it will always be called first, and if it returns
	 *           true, no other handlers will be called. */
	void AddHandler(pHandler handler, float priority = 0.5);
	//! Remove a message handler from the list.
	void RemoveHandler(pHandler handler);
	//! Remove all handlers from the message handler list.
	void ClearHandlers();
	
	/*! \brief Dispatch all enqueued messages that have not been sent.  Will
	 *         be called automatically.
	 *  \remarks You can pretend this function does not exist if you want.  You
	 *           should never call it yourself. */
	virtual void DoFrame(tTime time, tTime frametime);
};
/*! \typedef pGMessageQueue
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMessageQueue GMessageQueue\endlink. */
typedef GSmartPointer<GMessageQueue> pGMessageQueue;

/*! \brief The main GAUGE message queue.
 *  \ingroup input
 *
 *  In almost all cases, you'll want to use this message queue rather than
 *  create your own.  Plugins that supply input will enqueue messages to
 *  this queue. */
extern GAUGE3D_API pGMessageQueue gMainMessageQueue;

} //namespace GAUGE3D

#endif //included
