/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "filesystem.h"

// system ------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>

#ifdef PLATFORM_POSIX
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/mman.h>
#endif

#ifdef PLATFORM_WINDOWS
#include <windows.h>
//grrr...  damn lack of namespaces
#define Win32_CreateDirectory CreateDirectoryA
#define Win32_DeleteFile DeleteFileA
#undef CreateDirectory
#undef DeleteFile
#endif // PLATFORM_WINDOWS

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GFileSystem::Private
{
	// state variables -------------------------------------------------
	static GString mProgName;
	
	// private functions -----------------------------------------------
};

GString GFileSystem::Private::mProgName;

// osab_FileSystem member functions ==================================
// lifecycle ---------------------------------------------------------

void
GFileSystem::Init(GString progName)
{
	GString dirname, fileName;
	
	Private::mProgName = progName.ToAscii();
	
	//make sure the config directory exists.
	dirname = ConfigDir();
	if(!Exists(dirname))
		CreateDirectory(dirname);
	
#if 0   //temporary files currently disabled (because I would like to find a better
        //way to handle them.
	//make sure the temporary directory exists.
	GetTemporaryDir(&dirname);
	if(Exists(dirname))
		DeleteTree(dirname);    //delete old temporary files
	CreateDirectory(dirname);
	
	//Place a friendly message in the temp directory telling people that
	// they can delete it.
	fileName = dirname + "/README.TXT";
	file = fopen(fileName.AsciiData(), "w");
	fprintf(file, "This temporary directory belongs to: %s\n", Private::mProgName.AsciiData());
	fprintf(file, "You may delete this directory and its contents as long as the program\n");
	fprintf(file, "that owns it is not running.  If you do not delete it, it will be deleted\n");
	fprintf(file, "next time you run the program that created it.  Normally, the program\n");
	fprintf(file, "would have deleted it when it exited, but if the program crashed, it\n");
	fprintf(file, "might not have been able to delete this directory.\n");
	fprintf(file, "DO NOT PLACE ANYTHING OF ANY IMPORTANCE IN THIS DIRECTORY!\n");
	fclose(file);
#endif
}

void
GFileSystem::ShutDown()
{
	GString dirname;
	
#if 0 //temporarily disabled
	//we want to delete the temp directory now.
	GetTemporaryDir(&dirname);
	if(Exists(dirname))
		DeleteTree(dirname);
#endif
}

// operations --------------------------------------------------------

struct GFileSystem::File
{
#ifdef PLATFORM_POSIX
	int mFileDesc;
#else
	FILE* mStdioFile;
#endif
};

GFileSystem::File*
GFileSystem::Open(GString filename, int flags)
{
	filename = filename.ToAscii();
	
#ifdef PLATFORM_POSIX
	File* file = NULL;
	int filedesc;
	int openFlags;
	
	if(flags & READ)
	{
		if(flags & WRITE)
			openFlags = O_RDWR;
		else
			openFlags = O_RDONLY;
	}
	else if(flags & WRITE)
		openFlags = O_WRONLY;
	else
		return NULL;
	
	if(flags & CREATE)
	{
		openFlags |= O_CREAT;
		if(flags & NOREPLACE)
			openFlags |= O_EXCL;
	}
	if(flags & TRUNCATE)
		openFlags |= O_TRUNC;
	if(flags & APPEND)
		openFlags |= O_APPEND;
	
	filedesc = open(filename.AsciiData(), openFlags, S_IRWXU | S_IRGRP | S_IROTH);
	if(filedesc != -1)
	{
		file = new File;
		file->mFileDesc = filedesc;
	}
	
	return file;
	
#else //PLATFORM
	File* file = NULL;
	FILE* stdioFile = NULL;
	int flagPos = 0;
	
	if(!(flags & CREATE))
		if(!Exists(filename))
			return NULL;
	if(flags & NOREPLACE)
		if(Exists(filename))
			return NULL;
	
	if(flags & TRUNCATE)
	{
		if((flags & READ)
		&& (flags & WRITE))
			stdioFile = fopen(filename.AsciiData(), "wb+");
		else if(flags & WRITE)
			stdioFile = fopen(filename.AsciiData(), "wb");
	}
	else if(flags & APPEND)
	{
		if((flags & READ)
		&& (flags & WRITE))
			stdioFile = fopen(filename.AsciiData(), "ab+");
		else if(flags & WRITE)
			stdioFile = fopen(filename.AsciiData(), "ab");
	}
	else
	{
		if((flags & READ)
		&& (flags & WRITE))
		{
			stdioFile = fopen(filename.AsciiData(), "rb+");
			if(stdioFile == NULL)
				//file does not exist.  Open as "wb+" to create it.
				stdioFile = fopen(filename.AsciiData(), "wb+");
		}
		else if(flags & WRITE)
		{
			stdioFile = fopen(filename.AsciiData(), "rb+");  //attempt to avoid truncation
			if(stdioFile == NULL)
				stdioFile = fopen(filename.AsciiData(), "wb"); //could not avoid trunctation!
		}
		else if(flags & READ)
			stdioFile = fopen(filename.AsciiData(), "rb");
	}
	
	if(stdioFile != NULL)
	{
		file = new File;
		file->mStdioFile = stdioFile;
	}
	
	return file;
	
#endif //PLATFORM
}

int
GFileSystem::Read(File* file, void* target, int size)
{
#ifdef PLATFORM_POSIX
	return read(file->mFileDesc, target, size);
#else
	return fread(target, 1, size, file->mStdioFile);
#endif
}

int
GFileSystem::Write(File* file, const void* source, int size)
{
#ifdef PLATFORM_POSIX
	return write(file->mFileDesc, source, size);
#else
	return fwrite(source, 1, size, file->mStdioFile);
#endif
}

void
GFileSystem::Seek(File* file, int offset, tSeekType relativeTo)
{
#ifdef PLATFORM_POSIX
	int whence = 0;
	switch(relativeTo)
	{
	case BEGINNING:
		whence = SEEK_SET;
		break;
	case CURRENT:
		whence = SEEK_CUR;
		break;
	case END:
		whence = SEEK_END;
		break;
	};
	lseek(file->mFileDesc, offset, whence);

#else
	int whence = 0;
	switch(relativeTo)
	{
	case BEGINNING:
		whence = SEEK_SET;
		break;
	case CURRENT:
		whence = SEEK_CUR;
		break;
	case END:
		whence = SEEK_END;
		break;
	};
	fseek(file->mStdioFile, offset, whence);

#endif
}

int
GFileSystem::TellPos(File* file)
{
#ifdef PLATFORM_POSIX
	return lseek(file->mFileDesc, 0, SEEK_CUR);
#else
	return ftell(file->mStdioFile);
#endif
}

void
GFileSystem::Close(File* file)
{
#ifdef PLATFORM_POSIX
	close(file->mFileDesc);
	delete file;
#else
	fclose(file->mStdioFile);
	delete file;
#endif
}


struct GFileSystem::FileMapInternal
{
#ifdef HAVE_MMAP
	int mFileDesc;
	void* mRealMap;
	int mRealSize;
#elif defined(PLATFORM_WINDOWS)
	HANDLE mFile;
	HANDLE mFileMapping;
#endif
};

GFileSystem::FileMap*
GFileSystem::Map(GString filename, int offset, int size, int flags)
{
	filename = filename.ToAscii();
	
#ifdef HAVE_MMAP
	int fileDesc;
	int openFlags = O_RDONLY;
	int prot = 0;
	int mmapFlags = 0;
	int8* map;
	
	if(flags & WRITE)
		openFlags = O_RDWR;
	else
		openFlags = O_RDONLY;
	
	if(flags & CREATE)
	{
		openFlags |= O_CREAT;
		if(flags & NOREPLACE)
			openFlags |= O_EXCL;
	}
	
	fileDesc = open(filename.AsciiData(), openFlags, S_IRWXU | S_IRGRP | S_IROTH);
	if(fileDesc == -1)
		return NULL;
	
	if(flags & READ)
		prot |= PROT_READ;
	if(flags & WRITE)
		prot |= PROT_WRITE;
	if(prot == 0)
		prot = PROT_NONE;
	
	mmapFlags = MAP_SHARED;
	
	if(flags & WRITE)
	{
		struct stat st;
		fstat(fileDesc, &st);
		if(st.st_size < offset + size)
		{
			//grow the file.
			//FIXME: Is there a more efficient method than using write?
			lseek(fileDesc, 0, SEEK_END);
			char* junk = new char[offset + size - st.st_size];
			write(fileDesc, junk, offset + size - st.st_size);
			delete junk;
			lseek(fileDesc, 0, SEEK_SET);
		}
	}
	
	int realOffset, realSize;
	
	if(offset != 0)
	{
		//mmap wants an offset that is aligned to a page boundry.  Why they
		// can't simply do what I do here on their own I really don't know.
		realOffset = offset - offset % getpagesize();
		realSize = size + offset - realOffset;
	}
	else
	{
		realOffset = offset;
		realSize = size;
	}
	
	map = (int8*)mmap(NULL, realSize, prot, mmapFlags, fileDesc, realOffset);
	
	if(map == MAP_FAILED)
	{
		close(fileDesc);
		return NULL;
	}
	
	FileMap* fileMap = new FileMap;
	
	fileMap->mMap      = map + offset - realOffset;
	fileMap->mOffset   = offset;
	fileMap->mSize     = size;
	fileMap->mInternal = new FileMapInternal;
	fileMap->mInternal->mFileDesc = fileDesc;
	fileMap->mInternal->mRealMap  = map;
	fileMap->mInternal->mRealSize = realSize;
	
	return fileMap;

#elif defined(PLATFORM_WINDOWS)
	
	DWORD access = 0;
	DWORD shareMode = 0;
	DWORD protect;
	DWORD mapAccess;
	
	if(flags & READ)
	{
		access    |= GENERIC_READ;
		shareMode |= FILE_SHARE_READ;
		mapAccess  = FILE_MAP_READ;
	}
	if(flags & WRITE)
	{
		access    |= GENERIC_WRITE;
		shareMode |= FILE_SHARE_WRITE;
		mapAccess  = FILE_MAP_WRITE;
	}
	
	if((flags & READ) && !(flags & WRITE))
		protect = PAGE_READONLY;
	else
		protect = PAGE_READWRITE;
	
	//beware of large, ugly Win32 calls
	HANDLE file;
	file = CreateFile(filename.AsciiData(), access, shareMode,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(file == INVALID_HANDLE_VALUE)
		return NULL;  //ahhh, crap
	
	HANDLE fileMapping;
	fileMapping = CreateFileMapping(file, NULL, protect, 0, 0, NULL);
	if(fileMapping == INVALID_HANDLE_VALUE)
	{
		CloseHandle(file);
		return NULL;   //doh
	}

	void* map;
	map = MapViewOfFile(fileMapping, mapAccess, 0, offset, size);
	if(map == NULL)
	{
		CloseHandle(fileMapping);
		CloseHandle(file);
		return NULL;  //doubt this will ever happen...
	}

	//we made it!
	FileMap* fileMap = new FileMap;
	
	fileMap->mMap      = map;
	fileMap->mOffset   = offset;
	fileMap->mSize     = size;
	fileMap->mInternal = new FileMapInternal;
	fileMap->mInternal->mFile        = file;
	fileMap->mInternal->mFileMapping = fileMapping;
	
	return fileMap;
	
#else
# error "Don't know how to mmap"
#endif
}

void
GFileSystem::UnMap(FileMap* map)
{
#ifdef HAVE_MMAP
	munmap(map->mMap, map->mInternal->mRealSize);
	close(map->mInternal->mFileDesc);
	delete map->mInternal;
	delete map;
	return;

#elif defined(PLATFORM_WINDOWS)

	UnmapViewOfFile(map->mMap);
	CloseHandle(map->mInternal->mFileMapping);
	CloseHandle(map->mInternal->mFile);

	delete map->mInternal;
	delete map;
	
#else
# error "Don't know how to mmap"
#endif
}

int
GFileSystem::FileSize(GString filename)
{
	filename = filename.ToAscii();

#ifdef PLATFORM_POSIX
	struct stat filestat;
	
	if(stat(filename.AsciiData(), &filestat) != -1)
		return filestat.st_size;
	else
		return -1;
#else
	int size;
	FILE* file = fopen(filename.AsciiData(), "rb");
	if(file == NULL)
		return 0;
	fseek(file, 0, SEEK_END);
	size = ftell(file);
	fclose(file);
	return size;
#endif
}

int
GFileSystem::FileSize(File* file)
{
#ifdef PLATFORM_POSIX
	struct stat filestat;
	
	if(fstat(file->mFileDesc, &filestat) != -1)
		return filestat.st_size;
	else
		return -1;
#else
	int size;
	int pos;
	pos = TellPos(file);             //backup position
	Seek(file, 0, END);              //seek to end of file
	size = TellPos(file);            //get position
	Seek(file, pos, BEGINNING);      //restore old position
	return size;
#endif
}

//check if a file or directory exists.  To find out if it is a file or
// a directory, use IsDirecotry().
bool
GFileSystem::Exists(GString filename)
{
	filename = filename.ToAscii();

#ifdef PLATFORM_POSIX
	int fileDesc = open(filename.AsciiData(), O_RDONLY);
	if(fileDesc == -1)
		return false;
	close(fileDesc);
	return true;
#else
	FILE* file = fopen(filename.AsciiData(), "r");
	if(file == NULL)
		return false;
	fclose(file);
	return true;
#endif
}

void
GFileSystem::DeleteFile(GString filename)
{
	filename = filename.ToAscii();

#ifdef PLATFORM_POSIX
	unlink(filename.AsciiData());
#elif defined(PLATFORM_WINDOWS)
	Win32_DeleteFile(filename.AsciiData());
#else
# error "Don't know how to delete files on your OS"
#endif
}

//MANIPULATING DIRECTORIES--------------
struct GFileSystem::Directory
{
#ifdef PLATFORM_POSIX
	DIR* mPosixDir;
#elif defined(PLATFORM_WINDOWS)
	HANDLE mHandle;
	GString mName;
#endif
};

GFileSystem::Directory*
GFileSystem::OpenDirectory(GString dirname)
{
	dirname = dirname.ToAscii();
	
#ifdef PLATFORM_POSIX
	Directory* dir;
	DIR* posixDir;
	
	posixDir = opendir(dirname.AsciiData());
	if(posixDir == NULL)
		return NULL;
	dir = new Directory;
	dir->mPosixDir = posixDir;
	return dir;

#elif defined(PLATFORM_WINDOWS)
	
	if(!IsDirectory(dirname))
		return NULL;

	Directory* dir;

	dir = new Directory;
	dir->mName = dirname;
	dir->mHandle = NULL;

	return dir;
	
#else
# error "Don't know how to read directories on your OS"
#endif
}

bool
GFileSystem::ReadDirectory(GString* filename, Directory* directory)
{
#ifdef PLATFORM_POSIX
	dirent* entry;
	entry = readdir(directory->mPosixDir);
	if(entry == NULL)
		return false;
	*filename = entry->d_name;
	return true;

#elif defined(PLATFORM_WINDOWS)
	
	WIN32_FIND_DATA findData;
	
	if(directory->mHandle == NULL)
	{
		GString searchString = directory->mName + "/*";
		directory->mHandle = FindFirstFile(searchString.AsciiData(), &findData);
		if(directory->mHandle == INVALID_HANDLE_VALUE)
			return false;
	}
	else
	{
		if(!FindNextFile(directory->mHandle, &findData))
			return false;
	}
	
	*filename = findData.cFileName;
	return true;
	
#else
# error "Don't know how to read directories on you OS"
#endif
}

void
GFileSystem::RewindDirectory(Directory* directory)
{
#ifdef PLATFORM_POSIX
	rewinddir(directory->mPosixDir);
#elif defined(PLATFORM_WINDOWS)
	if(directory->mHandle != NULL)
	{
		FindClose(directory->mHandle);
		directory->mHandle = NULL;
	}
#else
# error "Don't know how to read directories on you OS"
#endif
}

void
GFileSystem::CloseDirectory(Directory* directory)
{
#ifdef PLATFORM_POSIX
	closedir(directory->mPosixDir);
	delete directory;
#elif defined(PLATFORM_WINDOWS)
	if(directory->mHandle != NULL)
		FindClose(directory->mHandle);
	delete directory;
#else
# error "Don't know how to read directories on you OS"
#endif
}

void
GFileSystem::CreateDirectory(GString dirname)
{
	dirname = dirname.ToAscii();
	
#ifdef PLATFORM_POSIX
	mkdir(dirname.AsciiData(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#elif defined(PLATFORM_WINDOWS)
	Win32_CreateDirectory(dirname.AsciiData(), NULL);
#else
#error "Don't know how to create directories on your OS"
#endif //PLATFORM_POSIX
}

void
GFileSystem::DeleteDirectory(GString dirname)
{
	dirname = dirname.ToAscii();
	
#ifdef PLATFORM_POSIX
	rmdir(dirname.AsciiData());
#elif defined(PLATFORM_WINDOWS)
	RemoveDirectory(dirname.AsciiData());
#else
# error "Don't know how to delete directories on your OS"
#endif
}

void GFileSystem::DeleteTree(GString dirname)
{
	dirname = dirname.ToAscii();
	
	Directory* dir;
	GString dirEntry;
	
	dir = OpenDirectory(dirname);
	if(dir == NULL)
		return;
	
	while(ReadDirectory(&dirEntry, dir))
	{
		dirEntry = dirname + '/' + dirEntry;
		if(IsDirectory(dirEntry))
			DeleteTree(dirEntry);
		else
			DeleteFile(dirEntry);
	}
	
	CloseDirectory(dir);
	
	DeleteDirectory(dirname);
}

bool
GFileSystem::IsDirectory(GString dirname)
{
	dirname = dirname.ToAscii();
	
#ifdef PLATFORM_POSIX
	DIR* test = opendir(dirname.AsciiData());
	if(test == NULL)
		return false;
	closedir(test);
	return true;
	
#elif defined(PLATFORM_WINDOWS)
	
	DWORD retval = GetFileAttributes(dirname.AsciiData());
	if(retval == 0xFFFFFFFF)
		return false;
	else if(retval & FILE_ATTRIBUTE_DIRECTORY)
		return true;
	else
		return false;
	
#else
# error "Don't know how to check if a directory exists on your OS"
#endif
}

//Get the current working directory.
GString
GFileSystem::WorkingDir()
{
#ifdef PLATFORM_POSIX
	GString ret;
	char* buffer;
	int bufSize = 100;
	
	buffer = new char[bufSize];
	while(getcwd(buffer, bufSize) == NULL)
	{
		delete [] buffer;            //the old buffer wasn't big enough, so
		bufSize *= 2;                    // expand it
		buffer = new char[bufSize];  // and try again
	}
	
	ret = buffer;
	delete [] buffer;
	
	return ret;

#elif defined(PLATFORM_WINDOWS)
	GString ret;
	char* buffer;
	char dummy[2];
	int size;

	//find out how much space we need.
	size = GetCurrentDirectory(2, dummy);

	//get directory name
	//(if MSVC supported dynamically-sized arrays on the stack like GCC,
	// we would not have to call new here and slow everything down!)
	buffer = new char[size];
	GetCurrentDirectory(size, buffer);

	ret = buffer;
	delete [] buffer;
	
	return ret;
#endif //PLATFORM_POSIX
}

GString
GFileSystem::ConfigDir()
{
#ifdef PLATFORM_WINDOWS
	return WorkingDir() + "/config";

#elif defined(PLATFORM_POSIX)
	char* homeDir = NULL;

	//Put config files in the user's home directory
	homeDir = getenv("HOME");
	
	GString ret = homeDir;
	ret += "/." + Private::mProgName;
	return ret;

#else
	#error "Port me, damnit!"
#endif
}

GString
GFileSystem::ResourceDir()
{
#ifdef PLATFORM_WINDOWS
	return WorkingDir();

#else

# ifdef PLATFORM_BEOS
#  error "where should the resource dir be on Be?"
# else
	//assume unix-like OS
	return PREFIX"/games/" + Private::mProgName;
# endif

#endif // PLATFORM_WINDOWS, else
}

GString
GFileSystem::PluginDir()
{
#ifdef PLATFORM_WINDOWS
	//TODO: Figure this one out.
	return WorkingDir();
	
#else
	
# ifdef PLATFORM_BEOS
#  error "where should plugins go on Be?"
# else
	//assume unix-like OS
	return PREFIX"/lib/gauge3d";
# endif

#endif
}

} // namespace GAUGE3D
