/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#if defined(PLATFORM_WINDOWS) && defined(HAVE_LIBDL)
#undef HAVE_LIBDL    //we don't want to use this on Windows
#endif

// includes ==========================================================
// system ------------------------------------------------------------
#ifndef __SYSTEM_IOSTREAM__INCLUDED__
#include <iostream>
#define __SYSTEM_IOSTREAM__INCLUDED__
#endif
using namespace std;

// component header --------------------------------------------------
#include "library.h"

// system ------------------------------------------------------------
#ifdef HAVE_LIBDL
#include <dlfcn.h>
#endif // HAVE_LIBDL

#ifdef PLATFORM_WINDOWS
#include <windows.h>
#endif // PLATFORM_WINDOWS

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GLibrary::Private
{
	// state variables -------------------------------------------------
	GString mFilename;
	bool mOpen;

#ifdef PLATFORM_WINDOWS
	HMODULE mLibrary;
#elif defined(HAVE_LIBDL)
	void* mLibrary;
#endif
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GLibrary::GLibrary(GString filename)
	: mPrivate(*(new Private))
{
	mPrivate.mFilename = filename.ToAscii();
	mPrivate.mOpen = false;
}

GLibrary::~GLibrary()
{
	if(mPrivate.mOpen)
		Close();

	delete &mPrivate;
}

// operations --------------------------------------------------------
bool
GLibrary::Open()
{
	if(!mPrivate.mOpen)
	{
#ifdef PLATFORM_WINDOWS
		char winPath[MAX_PATH];
		char* ptr;
		
		strncpy(winPath, mPrivate.mFilename.AsciiData(), MAX_PATH);
		
		for(int i = 0; i < mPrivate.mFilename.Length(); i++)
			if(winPath[i] == '/')
				winPath[i] = '\\';

		if(!strncmp(winPath, "\\\\", 2))
		{
			//convert universal path to windows path
			// ("//d/" -> "d:\")
			winPath[1] = winPath[2];
			winPath[2] = '\\';
			ptr = winPath + 1;
		}
		else
			ptr = winPath;

		mPrivate.mLibrary = LoadLibrary(ptr);
		if(mPrivate.mLibrary == NULL)
			return false;

#elif defined(HAVE_LIBDL)
		mPrivate.mLibrary = dlopen(mPrivate.mFilename.AsciiData(), RTLD_NOW);
		if(mPrivate.mLibrary == NULL)
		{
			cout << dlerror() << endl;
			return false;
		}

#else
#error "Don't know how to read dynamic libraries on your OS."
#endif

		mPrivate.mOpen = true;
	}
	
	return true;
}

void*
GLibrary::GetSymbol(GString symbolName)
{
	if(!mPrivate.mOpen)
		return NULL;
	
	symbolName = symbolName.ToAscii();

#ifdef PLATFORM_WINDOWS
	return GetProcAddress(mPrivate.mLibrary, symbolName.AsciiData());
#elif defined(HAVE_LIBDL)
	return dlsym(mPrivate.mLibrary, symbolName.AsciiData());
#else
#error "Don't know how to read dynamic libraries on your OS."
#endif // HAVE_LIBDL
}

void
GLibrary::Close()
{
	if(!mPrivate.mOpen)
		return;

#ifdef PLATFORM_WINDOWS
	FreeLibrary(mPrivate.mLibrary);
#elif defined(HAVE_LIBDL)
	dlclose(mPrivate.mLibrary);
#else
#error "Don't know how to read dynamic libraries on your OS."
#endif // HAVE_LIBDL

	mPrivate.mOpen = false;
}

// access ------------------------------------------------------------
GString
GLibrary::Filename()
{
	return mPrivate.mFilename;
}
	
// inquiry -----------------------------------------------------------
bool
GLibrary::IsOpen()
{
	return mPrivate.mOpen;
}


GAUGE3D_DECLARE_INTERFACES(GLibrary)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
