/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_OSABSTRACTION_LIBRARY_H__INCLUDED__
#define __GAUGE3D_OSABSTRACTION_LIBRARY_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include "gauge3d/common/gstring.h"
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GLibrary library.h gauge3d/osabstraction/library.h
 * \ingroup osabstraction
 * \brief Loads shared libraries (dll's and so's).
 *
 * This class allows you to load executable add-ons to your program.  You
 * could use it directly, but we recommend using GObjectLoader instead, as
 * it provides a much nicer system for doing this.  GObjectLoader uses GLibrary
 * to do what it does. */
class GAUGE3D_API GLibrary: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781ef8);
public:
	// lifecycle -------------------------------------------------------
	/*! \brief Create a GLibrary for a given library file.
	 *  \param filename The path and name of the library file on the disk.
	 *                  If no path is given (just a filename), the system's
	 *                  standard library directories will be searched.
	 *  \remarks  The library can be closed when not in use and re-openned later, but
	 *            the filename will never change.  This way, the library implementation
	 *            can opt to keep the libraries open when in debug mode in case the
	 *            debugger is easily confused (most are! even GDB!).*/
	GLibrary(GString filename);
	//! Close the library.
	~GLibrary();

	// operations ------------------------------------------------------
	/*! \brief Open the library.
	 * \returns True if the library was successfully opened, or false if an error
	 *          occured.
	 * \remarks Don't forget that if you close and re-open a library, you have to
	 *          get all your symbols again.  They will not be the same. */
	bool Open();
	/*! \brief Get a pointer to a function in the library.
	 * \param symbolName The name of the function to get a pointer to.
	 * \returns A pointer to the function, or NULL if an error occured or the function
	 *          was not found.
	 * \remarks Most operating systems allow you to get symbols other than functions
	 *          from a library.  Do not depend on this behavior. */
	void* GetSymbol(GString symbolName);
	/*! \brief Unload the library from memory.
	 *
	 * This will unload the library from RAM.  If you need to use it again, call Open(). */
	void Close();

	// access ----------------------------------------------------------
	//! Get the name of the library, as given to the constructor.
	GString Filename();
	
	// inquiry ---------------------------------------------------------
	//! Returns true if the library is open.
	bool IsOpen();
};
/*! \typedef pGLibrary
 *  \ingroup osabstraction
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GLibrary GLibrary\endlink. */
typedef GSmartPointer<GLibrary> pGLibrary;

} // namespace GAUGE3D

#endif //included
