/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "thread.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_IOSTREAM__INCLUDED__
#include <iostream>
#define __SYSTEM_IOSTREAM__INCLUDED__
#endif //redundant include guard
using namespace std;

#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif // HAVE_LIBPTHREAD

#ifdef PLATFORM_WINDOWS
#include <windows.h>
#endif // PLATFORM_WINDOWS

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GThread::Private
{
	// state variables -------------------------------------------------
	void* (*mStart)(void*);
#if defined(HAVE_LIBPTHREAD)
	pthread_t mThread;
#elif defined(PLATFORM_WINDOWS)
	HANDLE mThread;
	DWORD mThreadId;
#endif // PLATFORM_WINDOWS
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
//start is the thread's start function.
GThread::GThread(void* (*start)(void*))
	: mPrivate(*(new Private))
{
	mPrivate.mStart = start;
#ifdef PLATFORM_WINDOWS
	mPrivate.mThread = NULL;
#endif
}

GThread::~GThread()
{
#ifdef PLATFORM_WINDOWS
	if(mPrivate.mThread)
		CloseHandle(mPrivate.mThread);
#endif
	delete &mPrivate;
}

// operations --------------------------------------------------------
//vStart starts a new thread which immediately calls the thread's start
// function, passing rg as the argument.
void
GThread::Start(void* arg)
{
#if defined(HAVE_LIBPTHREAD)
	pthread_create(&mPrivate.mThread, NULL, mPrivate.mStart, arg);
#elif defined(PLATFORM_WINDOWS)
	if(mPrivate.mThread)
		CloseHandle(mPrivate.mThread);
	mPrivate.mThread = CreateThread(NULL, 0, reinterpret_cast<DWORD (__stdcall*)(void *)>(mPrivate.mStart), arg, 0, &mPrivate.mThreadId);
#endif // HAVE_LIBPTHREAD
}

//pvJoin returns the thread's return value.  If the thread has not yet
// finished, pvJoin waits for it to finish.
void*
GThread::Join()
{
#if defined(HAVE_LIBPTHREAD)
	void* pvRet;
	pthread_join(mPrivate.mThread, &pvRet);
	return pvRet;
#elif defined(PLATFORM_WINDOWS)
	DWORD ret;
	WaitForSingleObject(mPrivate.mThread, INFINITE);
	GetExitCodeThread(mPrivate.mThread, &ret);
	CloseHandle(mPrivate.mThread);
	mPrivate.mThread = NULL;
	return reinterpret_cast<void *>(ret);
#endif // HAVE_LIBPTHREAD
}

//vDetach says that the thread will not be joined ever.  You must call
// this if you don't plan on joining it.  Otherwise, a memory leak
// could occur.
void
GThread::Detach()
{
#ifdef HAVE_LIBPTHREAD
	pthread_detach(mPrivate.mThread);
#endif // HAVE_LIBPTHREAD
}
	
//vReturn quits from the current thread using retVal as the return
// value.
void
GThread::Return(void* retVal)
{
#if defined(HAVE_LIBPTHREAD)
	pthread_exit(retVal);
#elif defined(PLATFORM_WINDOWS)
	ExitThread(reinterpret_cast<DWORD>(retVal));
#endif // HAVE_LIBPTHREAD
}


GAUGE3D_DECLARE_INTERFACES(GThread)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES


// class state definition ============================================
struct GMutex::Private
{
	// state variables -------------------------------------------------
#if defined(HAVE_LIBPTHREAD)
	pthread_mutex_t mMutex;
#elif defined(PLATFORM_WINDOWS)
	HANDLE mMutex;
#endif // HAVE_LIBPTHREAD
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GMutex::GMutex()
	: mPrivate(*(new Private))
{
#if defined(HAVE_LIBPTHREAD)
	pthread_mutex_init(&mPrivate.mMutex, NULL);
#elif defined(PLATFORM_WINDOWS)
	mPrivate.mMutex = CreateMutex(NULL, FALSE, NULL);
#endif // HAVE_LIBPTHREAD
}

GMutex::~GMutex()
{
#if defined(HAVE_LIBPTHREAD)
	if(pthread_mutex_destroy(&mPrivate.mMutex) != 0)
	{
		cerr << "GMutex: ERROR: Tried to destroy locked mMutex.\n"
		     << "            This may lead to segfaults later on!" << endl;
	}
#elif defined(PLATFORM_WINDOWS)
	CloseHandle(mPrivate.mMutex);
#endif // HAVE_LIBPTHREAD

	delete &mPrivate;
}
	
// operations --------------------------------------------------------
//vLock waits until the mMutex is unlocked and then locks the
// mMutex to the calling thread.
void
GMutex::Lock()
{
#if defined(HAVE_LIBPTHREAD)
	pthread_mutex_lock(&mPrivate.mMutex);
#elif defined(PLATFORM_WINDOWS)
	WaitForSingleObject(mPrivate.mMutex, INFINITE);
#endif // HAVE_LIBPTHREAD
}

//bTryLock checks if the mMutex is unlocked before locking it.  If it
// is unlocked, bTryLock locks the mMutex to the calling thread and
// returns true.  Otherwise, the mMutex is not locked, and false is
// returned.  bTryLock will never wait for the mMutex to unlock.
bool
GMutex::TryLock()
{
#if defined(HAVE_LIBPTHREAD)
	if(pthread_mutex_trylock(&mPrivate.mMutex) == 0)
		return true;
	return false;
#elif defined(PLATFORM_WINDOWS)
	if(WaitForSingleObject(mPrivate.mMutex, 0) == WAIT_OBJECT_0)
		return true;
	return false;
#endif // HAVE_LIBPTHREAD
}

//vUnlock unlocks the mMutex.  Only the thread that the mMutex is locked
// to may call this.
void
GMutex::Unlock()
{
#if defined(HAVE_LIBPTHREAD)
	pthread_mutex_unlock(&mPrivate.mMutex);
#elif defined(PLATFORM_WINDOWS)
	ReleaseMutex(mPrivate.mMutex);
#endif // HAVE_LIBPTHREAD
}


GAUGE3D_DECLARE_INTERFACES(GMutex)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
