/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_OSABSTRACTION_THREAD_H__INCLUDED__
#define __GAUGE3D_OSABSTRACTION_THREAD_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GThread thread.h gauge3d/osabstraction/thread.h
 *  \ingroup osabstraction
 *  \brief Multithreading support.
 *
 *  This class allows for multitasking support. */
class GAUGE3D_API GThread: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781ee5);
public:
	// lifecycle -------------------------------------------------------
	/*! \brief Create the thread, but don't start it yet.
	 * \param start A pointer to the function to run from the new thread.
	 *              The function takes one parameter of type void* and returns
	 *              void*. */
	GThread(void* (*start)(void*));
	//! Kill and destroy the thread.
	~GThread();

	// operations ------------------------------------------------------
	/*! \brief Start the thread with the given argument.
	 * \param arg The argument to pass to the thread's start function. */
	void Start(void* arg);
	/*! \brief Wait for the thread to finish and get the return value.
	 * \returns The return value from the threads start function.
	 * \remarks This function will not return until the thread has finished
	 *          executing.  Be careful when you call this. */
	void* Join();
	/*! \brief Call if you will not be joining the thread.
	 *
	 * If you do not call this and do not call Join(), some resources associated
	 * with the thread may never be freed. */
	void Detach();
	
	/*! \brief Quit the current thread.
	 * \param retVal Value to use as the return value from the start function.
	 *
	 * Execution of the thread will not continue after calling this. */
	static void Return(void* retVal);
};
/*! \typedef pGThread
 *  \ingroup osabstraction
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GThread GThread\endlink. */
typedef GSmartPointer<GThread> pGThread;


/*! \class GMutex thread.h gauge3d/osabstraction/thread.h
 *  \ingroup osabstraction
 *  \brief Use to prevent data corruption caused by multiple threads accessing
 *         the same data.
 *
 *  If two or more threads can access the same data, and at least one is writing to it,
 *  you should use a mutex to prevent them from accessing it at the same time. */
class GAUGE3D_API GMutex: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781eed);
public:
	// lifecycle -------------------------------------------------------
	//! Create the mutex.
	GMutex();
	//! Destroy the mutex.  It must be unlocked when this happens.
	~GMutex();
	
	// operations ------------------------------------------------------
	//! Wait until mutex is unlocked, and then lock the mutex to this thread.
	void Lock();
	/*! \brief Check if mutex is locked, and lock it to this thread if not.
	 *  \returns True if the mutex was not locked (and is now locked to the calling
	 *           thread), or false if the mutex was locked by another thread at the time. */
	bool TryLock();
	/*! \brief Release lock on a thread.
	 *
	 *  Only the thread that the mutex is locked to may call this. */
	void Unlock();
};
/*! \typedef pGMutex
 *  \ingroup osabstraction
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMutex GMutex\endlink. */
typedef GSmartPointer<GMutex> pGMutex;

} // namespace GAUGE3D

#endif //included
