/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "objectloader.h"

// system ------------------------------------------------------------
#include <vector>
#include <map>
using namespace std;

#include <ctype.h>
#include <assert.h>
#include <string.h>

// external packages -------------------------------------------------
#ifndef __GAUGE3D_OSABSTRACTION_LIBRARY_H__INCLUDED__
#include "gauge3d/osabstraction/library.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_PLUGINS_PLUGIN_H__INCLUDED__
#include "plugin.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GObjectLoader::Private
{
	typedef void* (*tCreateFunc)();
	
	typedef int tObjectIndex;  //index to an ObjectEntry in mObjects
	typedef int tPluginIndex;  //index to a  PluginEntry in mPlugins
	
	struct ObjectEntry
	{
		ObjectInfo mInfo;
		tPluginIndex mPlugin;
	};
	
	struct PluginEntry
	{
		GString mName;
		GString mDescription;
		GPlugin* mPlugin;
		pGLibrary mLibrary;
		pGFile mLibraryFile;
		bool mOpen;     //should always be true for now...
		tObjectIndex mFirstObject;
		int mNumObjects;
		
		//TODO:  implement stuff for unloading plugins that are not in use.
	};
	
	vector<ObjectEntry> mObjects;
	vector<PluginEntry> mPlugins;
	
	typedef vector<tObjectIndex>       tObjectList;
	typedef map<GString, tObjectList > tObjectMap;
	typedef map<GString, tObjectIndex> tClassnameMap;
	
	struct BaseClassGroup
	{
		tObjectList mObjectList;   //all objects with this base class
		tObjectMap  mKeywordMap;   //maps keywords to objects
		tObjectMap  mFileTypeMap;  //maps file type extensions to objects
	};
	typedef map<GString, BaseClassGroup> tBaseClassMap;
	
	// maps for accellerated plugin searches
	tClassnameMap mClassnameMap;
	tBaseClassMap mBaseClassMap;
	tObjectMap    mFileTypeMap;
	tObjectMap    mKeywordMap;
};

// class member functions ============================================
// lifecycle -------------------------------------------------------
GObjectLoader::GObjectLoader()
	: mPrivate(*(new Private))
{
}
GObjectLoader::~GObjectLoader()
{
	delete &mPrivate;
}

// operations ------------------------------------------------------

// loading objects -------------------
pGObject
GObjectLoader::LoadObject(const GString& className)
{
	Private::tClassnameMap::iterator i;
	i = mPrivate.mClassnameMap.find(className);
	if(i == mPrivate.mClassnameMap.end())
		return NULL;
	
	Private::ObjectEntry* object = &mPrivate.mObjects[i->second];
	Private::PluginEntry* plugin = &mPrivate.mPlugins[object->mPlugin];
	assert(plugin->mOpen);
	return plugin->mPlugin->CreateObject(className);
}

pGObject
GObjectLoader::LoadObjectOfType(const GString& baseClass)
{
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return NULL;
	
	//take the first object in the list, since we don't know any better.
	// (if there are multible objects with the same baseclass,
	// the user will have to use the search functions to choose one, or specify
	// a keyword or a file.)
	Private::ObjectEntry* object = &mPrivate.mObjects[i->second.mObjectList[0]];
	Private::PluginEntry* plugin = &mPrivate.mPlugins[object->mPlugin];
	assert(plugin->mOpen);
	return plugin->mPlugin->CreateObject(object->mInfo.mClassname);
}

pGObject
GObjectLoader::LoadObjectWithKeyword(const GString& keyWord, const GString& baseClass)
{
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return NULL;
	
	Private::tObjectMap::iterator j;
	j = i->second.mKeywordMap.find(keyWord);
	if(j == i->second.mKeywordMap.end())
		return NULL;
	
	//take the first object in the list, since we don't know any better.
	// (if there are multible objects with the same keyword and baseclass,
	// the user will have to use the search functions to choose one.)
	Private::ObjectEntry* object = &mPrivate.mObjects[j->second[0]];
	Private::PluginEntry* plugin = &mPrivate.mPlugins[object->mPlugin];
	assert(plugin->mOpen);
	return plugin->mPlugin->CreateObject(object->mInfo.mClassname);
}

pGObject
GObjectLoader::LoadObjectForFile(pGFile file, const GString& baseClass)
{
	GString filename = file->Name();
	int extpos = filename.FindLastOf('.');
	if(extpos < 0)
		return NULL;  //no extension
	
	GString extension = filename.SubString(extpos + 1);
	
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return NULL;
	
	Private::tObjectMap::iterator j;
	j = i->second.mFileTypeMap.find(extension);
	if(j == i->second.mFileTypeMap.end())
		return NULL;
	
	//take the first object in the list, since we don't know any better.
	// (if there are multible objects with the same file type and baseclass,
	// the user will have to use the search functions to choose one.)
	Private::ObjectEntry* object = &mPrivate.mObjects[j->second[0]];
	Private::PluginEntry* plugin = &mPrivate.mPlugins[object->mPlugin];
	assert(plugin->mOpen);
	return plugin->mPlugin->CreateObject(object->mInfo.mClassname, file);
}

// searching for objects ------------
//these functions allow you to search for particular objects to fulfill
// your needs.
GSmartPointer<GObjectLoader::ObjectInfo>
GObjectLoader::FindObject(const GString& classname)
{
	Private::tClassnameMap::iterator i;
	i = mPrivate.mClassnameMap.find(classname);
	if(i == mPrivate.mClassnameMap.end())
		return NULL;
	
	GSmartPointer<ObjectInfo> ret = new ObjectInfo;
	*ret = mPrivate.mObjects[i->second].mInfo;
	
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindObjectsOfType(const GString& baseClass)
{
	if(baseClass == "GObject")
		return FindAllObjects();
	
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectList& objectList = i->second.mObjectList;
	
	GArray<ObjectInfo> ret(objectList.size());
	for(int j = 0; j < ret.Size(); j++)
		ret[j] = mPrivate.mObjects[objectList[j]].mInfo;
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindObjectsWithKeyword(const GString& keyword)
{
	Private::tObjectMap::iterator i;
	i = mPrivate.mKeywordMap.find(keyword);
	if(i == mPrivate.mKeywordMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectList& objectList = i->second;
	
	GArray<ObjectInfo> ret(objectList.size());
	for(int j = 0; j < ret.Size(); j++)
		ret[j] = mPrivate.mObjects[objectList[j]].mInfo;
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindObjectsWithKeyword(const GString& keyword, const GString& baseClass)
{
	if(baseClass == "GObject")
		return FindObjectsWithKeyword(keyword);
	
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectMap::iterator j;
	j = i->second.mKeywordMap.find(keyword);
	if(j == i->second.mKeywordMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectList& objectList = j->second;
	
	GArray<ObjectInfo> ret(objectList.size());
	for(int k = 0; k < ret.Size(); k++)
		ret[k] = mPrivate.mObjects[objectList[k]].mInfo;
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindObjectsForFile(pGFile file)
{
	GString filename = file->Name();
	int extpos = filename.FindLastOf('.');
	if(extpos < 0)
		return GArray<ObjectInfo>();  //no extension
	
	GString extension = filename.SubString(extpos + 1);
	
	Private::tObjectMap::iterator i;
	i = mPrivate.mFileTypeMap.find(extension);
	if(i == mPrivate.mFileTypeMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectList& objectList = i->second;
	
	GArray<ObjectInfo> ret(objectList.size());
	for(int j = 0; j < ret.Size(); j++)
		ret[j] = mPrivate.mObjects[objectList[j]].mInfo;
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindObjectsForFile(pGFile file, const GString& baseClass)
{
	if(baseClass == "GObject")
		return FindObjectsForFile(file);
	
	GString filename = file->Name();
	int extpos = filename.FindLastOf('.');
	if(extpos < 0)
		return GArray<ObjectInfo>();  //no extension
	
	GString extension = filename.SubString(extpos + 1);
	
	Private::tBaseClassMap::iterator i;
	i = mPrivate.mBaseClassMap.find(baseClass);
	if(i == mPrivate.mBaseClassMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectMap::iterator j;
	j = i->second.mFileTypeMap.find(extension);
	if(j == i->second.mFileTypeMap.end())
		return GArray<ObjectInfo>();
	
	Private::tObjectList& objectList = j->second;
	
	GArray<ObjectInfo> ret(objectList.size());
	for(int k = 0; k < ret.Size(); k++)
		ret[k] = mPrivate.mObjects[objectList[k]].mInfo;
	return ret;
}

GArray<GObjectLoader::ObjectInfo>
GObjectLoader::FindAllObjects()
{
	GArray<ObjectInfo> ret(mPrivate.mObjects.size());
	for(int i = 0; i < ret.Size(); i++)
		ret[i] = mPrivate.mObjects[i].mInfo;
	return ret;
}

// searching for plugins ------------
GArray<GObjectLoader::PluginInfo>
GObjectLoader::FindAllPlugins()
{
	GArray<PluginInfo> ret(mPrivate.mPlugins.size());
	Private::PluginEntry* entry;
	
	for(int i = 0; i < ret.Size(); i++)
	{
		entry = &mPrivate.mPlugins[i];
		ret[i].mName        = entry->mName;
		ret[i].mDescription = entry->mDescription;
		ret[i].mLibName     = entry->mLibraryFile->Name();
		
		ret[i].mObjects = GArray<ObjectInfo>(entry->mNumObjects);
		
		for(int j = 0; j < ret[i].mObjects.Size(); j++)
			ret[i].mObjects[j] = mPrivate.mObjects[entry->mFirstObject + j].mInfo;
	}
	
	return ret;
}

GSmartPointer<GObjectLoader::PluginInfo>
GObjectLoader::FindPluginForObject(GString classname)
{
	Private::tClassnameMap::iterator i;
	i = mPrivate.mClassnameMap.find(classname);
	if(i == mPrivate.mClassnameMap.end())
		return NULL;
	
	Private::ObjectEntry* object = &mPrivate.mObjects[i->second];
	Private::PluginEntry* plugin = &mPrivate.mPlugins[object->mPlugin];
	
	GSmartPointer<PluginInfo> ret = new PluginInfo;
	
	ret->mName        = plugin->mName;
	ret->mDescription = plugin->mDescription;
	ret->mLibName     = plugin->mLibraryFile->Name();
	
	ret->mObjects = GArray<ObjectInfo>(plugin->mNumObjects);
	
	for(int j = 0; j < ret->mObjects.Size(); j++)
		ret->mObjects[j] = mPrivate.mObjects[plugin->mFirstObject + j].mInfo;
	
	return ret;
}

// adding plugins -------------------
//add a plugin to the list of plugins to search for objects.
void
GObjectLoader::AddPlugin(pGFile libraryFile)
{
	int i, j, k;
	
	//libraries need to be real files on the disk under all OS's I know of.
	GString diskFile = libraryFile->MapToDisk();
	pGLibrary library(new GLibrary(diskFile));
	
	if(!library->Open())
		return;        //error loading library
	
	Private::tCreateFunc createFunc;
	GPlugin* plugin;
	
	//get the library function which creates the plugin.
	createFunc = (Private::tCreateFunc)library->GetSymbol("__gauge3d_plugin_create");
	if(createFunc == NULL)
		return;   //could not find entry symbol
	plugin = (GPlugin*)(createFunc());  //create the plugin
	
	//interrogate the poor little plugin
	Private::tObjectIndex objectIndex = mPrivate.mObjects.size();
	GArray<ObjectInfo> objects = plugin->ObjectList();
	
	Private::PluginEntry pluginEntry;
	pluginEntry.mName        = plugin->Name();
	pluginEntry.mDescription = plugin->Description();
	pluginEntry.mPlugin      = plugin;
	pluginEntry.mLibrary     = library;
	pluginEntry.mLibraryFile = libraryFile;
	pluginEntry.mOpen        = true;
	pluginEntry.mFirstObject = objectIndex;
	pluginEntry.mNumObjects  = objects.Size();
	
	Private::tPluginIndex pluginIndex = mPrivate.mPlugins.size();
	mPrivate.mPlugins.push_back(pluginEntry);
	
	//process all the new objects available.
	Private::ObjectEntry curObject;
	Private::BaseClassGroup* baseClass;
	for(i = 0; i < objects.Size(); i++)
	{
		//need to add each object to all the appropriat maps
		curObject.mInfo = objects[i];
		curObject.mPlugin = pluginIndex;
		
		//add to object list and get index
		objectIndex = mPrivate.mObjects.size();
		mPrivate.mObjects.push_back(curObject);
		
		//add to primary classname map
		mPrivate.mClassnameMap[curObject.mInfo.mClassname] = objectIndex;
		
		//add to base class map
		for(j = 0; j < curObject.mInfo.mBaseClasses.Size(); j++)
		{
			baseClass = &mPrivate.mBaseClassMap[curObject.mInfo.mBaseClasses[j]];
			
			//add to main list
			baseClass->mObjectList.push_back(objectIndex);
			
			//add to base class group's keyword map
			for(k = 0; k < curObject.mInfo.mKeywords.Size(); k++)
				baseClass->mKeywordMap[curObject.mInfo.mKeywords[k]].push_back(objectIndex);
			
			//add to base class group's file type map
			for(k = 0; k < curObject.mInfo.mFileTypes.Size(); k++)
				baseClass->mFileTypeMap[curObject.mInfo.mFileTypes[k]].push_back(objectIndex);
		}
		
		//add to primary keyword map
		for(k = 0; k < curObject.mInfo.mKeywords.Size(); k++)
			mPrivate.mKeywordMap[curObject.mInfo.mKeywords[k]].push_back(objectIndex);
		
		//add to primary file type map
		for(k = 0; k < curObject.mInfo.mFileTypes.Size(); k++)
			mPrivate.mFileTypeMap[curObject.mInfo.mFileTypes[k]].push_back(objectIndex);
	}
}

pGObjectLoader gMainObjectLoader;

GAUGE3D_DECLARE_INTERFACES(GObjectLoader)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
