/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_RENDERER_SPRITE_H__INCLUDED__
#define __GAUGE3D_RENDERER_SPRITE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include "gauge3d/3dmath/vector.h"
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_QUATERNION_H__INCLUDED__
#include "gauge3d/3dmath/quaternion.h"
#endif //redundant include guard

#ifndef __GAUGE3D_GRAPHICS_MODEL_H__INCLUDED__
#include "gauge3d/graphics/model.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_RENDERER_TEXTURE_H__INCLUDED__
#include "texture.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GSprite sprite.h gauge3d/renderer/sprite.h
 *  \ingroup renderer
 *  \brief Represents an object in the 3D scene.
 *
 * A sprite is any independently moving object in the scene.  In GAUGE, a
 * "sprite" is not necessarily a "billboard object", although it could be.
 * A sprite can have any geometry which can be described by a GModel. */
class GAUGE3D_API GSprite: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781fa6);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GSprite();
	
	// operations ------------------------------------------------------
	//! Tells the renderer to start drawing this object whenever it is visible.
	virtual void Show() = 0;
	//! Tells the renderer not to draw this object at all.
	virtual void Hide() = 0;
	//! Permenantly removes the sprite from the scene.
	virtual void Remove() = 0;
	
	// access ----------------------------------------------------------
	//! Sets the sprite's GModel.
	virtual void Model(pGModel model) = 0;
	//! Sets the animation state for the sprite's model.
	virtual void AnimState(GModel::pAnimState anim) = 0;
	//! Sets the texture to apply to the model.
	virtual void Texture(pGTexture texture) = 0;
	
	//! Sets the position of the sprite.
	virtual void Position(const GVector& position) = 0;
	/*! \brief Sets the direction that the sprite is facing.
	 *  \remarks Don't forget that you can convert a GAngles into a GQuaternion. */
	virtual void Angle(const GQuaternion& angle) = 0;
	
	/*! \brief Set the color of the sprite.  (alpha unchanged)
	 *
	 * \remarks
	 *  The color is multiplied component-wise with the texture color.  If you don't
	 *  want any color other than the texture color to be applied, just set the color
	 *  to 1.0, 1.0, 1.0, 1.0 (full white). */
	virtual void Color(float red, float green, float blue) = 0;
	//! Set the color and alpha of the sprite.
	virtual void Color(float red, float green, float blue, float alpha) = 0;
	//! Set the color of the sprite from a three-element array.  (alpha unchanged)
	virtual void Color3v(const float color[]) = 0;
	//! Set the color and alpha of the sprite from a four-element array.
	virtual void Color4v(const float color[]) = 0;
	//! Set the alpha of the sprite.  (color unchanged)
	virtual void Alpha(float alpha) = 0;
	
	/*! \brief Set the reflectivity properties of the sprite.
	 *  \param specular  Amount of specular light to reflect.
	 *  \param shininess How focused teh specular highlight is.
	 *  \param emission  The amount of light emitted from the sprite. */
	virtual void LightProps(float specular, float shininess, float emission) = 0;
	//! Set true to ignore lighting and render the sprite at maximum brightness.
	virtual void FullBright(bool fullBright) = 0;
	
	//! Sets a uniform scale factor for the sprite.
	virtual void Scale(GCoordinate magnification) = 0;
};
/*! \typedef pGSprite
 *  \ingroup renderer
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GSprite GSprite\endlink. */
typedef GSmartPointer<GSprite> pGSprite;

} // namespace GAUGE3D

#endif //included
