/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_SCHEDULER_H__INCLUDED__
#define __GAUGE3D_SCHEDULER_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GScheduler scheduler.h gauge3d/scheduler.h
 *  \ingroup control
 *  \breif Schedules events.
 *
 *  GScheduler is used to synchronize events throughout the engine.  Each
 *  subsystem has a scheduler for events which relate to that subsystem.
 *  Many classes which are part of that subsystem will register themselves
 *  with the scheduler.  Then, the scheduler will call those objects'
 *  DoFrame() functions every frame.
 *
 *  For example, when you create a GDisplay and give it a GCamera to
 *  render from, you do not have to tell it when to render each frame.
 *  This is because the GDisplay asks gVideoScheduler to call it every
 *  frame.  When it gets called back by gVideoScheduler, the GDisplay
 *  renders the scene. */
class GAUGE3D_API GScheduler: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x3a04a7b8);
public:
	// lifecycle -------------------------------------------------------
	//! Constructor.
	GScheduler();
	//! Destructor.
	~GScheduler();
	
	// operations ------------------------------------------------------
	/*! \brief Tell the scheduler to call object->DoFrame() every frame.
	 *  \param priority This value can optionally be used to tell the scheduler
	 *                  in what order to call objects.  Objects with lower
	 *                  priorities will be called earlier in the frame.
	 *  \remarks Typically, this function will be called by the constructor
	 *           of an object, and StopCallingEveryFrame() will be called by
	 *           the destructor. */
	void CallEveryFrame(GObject* object, int priority = 0);
	//! Remove the object from the list of objects to call every frame.
	void StopCallingEveryFrame(GObject* object);
	
	//! Get the priority of an object on the call every frame list.
	int GetObjectPriority(GObject* object);
	//! Change the priority of an object on the call every frame list.
	void SetObjectPriority(GObject* object, int priority);
	
	/*! \brief Call the DoFrame() functions for all objects which requested it.
	 *  \param time      The current time.
	 *  \param frametime The amount of time spent processing the previous frame.
	 *
	 * This function is normally called by GEngine::MainLoop(). */
	virtual void DoFrame(tTime time, tTime frametime);
	
	// access ----------------------------------------------------------
	//! Get the current time.
	tTime Time();
	//! Get the amount of time spent processing the previous frame.
	tTime FrameTime();
};
/*! \typedef pGScheduler
 *  \ingroup control
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GScheduler GScheduler\endlink. */
typedef GSmartPointer<GScheduler> pGScheduler;

/*! \brief Scheduler for miscellaneous tasks.
 *  \ingroup control
 *
 *  GEngine::MainLoop() will call gMiscScheduler->DoFrame() every frame. */
extern GAUGE3D_API pGScheduler gMiscScheduler;
/*! \brief Scheduler for input and message handling tasks.
 *  \ingroup control
 *
 *  GEngine::MainLoop() will call gInputScheduler->DoFrame() every frame
 *  if the input subsystem has been activated. */
extern GAUGE3D_API pGScheduler gInputScheduler;
/*! \brief Scheduler for video and rendering tasks.
 *  \ingroup control
 *
 *  GEngine::MainLoop() will call gVideoScheduler->DoFrame() every frame
 *  if the video subsystem has been activated. */
extern GAUGE3D_API pGScheduler gVideoScheduler;

} //namespace GAUGE3D

#endif //included
