/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_LINE_H__INCLUDED__
#define __GAUGE3D_3DMATH_LINE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include"vector.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GLine line.h gauge3d/3dmath/line.h
 *  \ingroup d3math
 *  \brief Represents a line in 3D space.
 *
 *  The line is represented using Plcker coordinates.  A line is represented
 *  by two vectors U and V.  U is the direction of the line, and the cross
 *  product of U and any point on the line is equal to V. */
class GAUGE3D_API GLine
{
public:
	/*! \name Coordinates
	 *  \brief Plcker coordinates for the line. */
	//@{
	GVector u;
	GVector v;
	//@}

	// lifecycle -------------------------------------------------------
	GLine();
	GLine(const GVector& position, const GVector& direction);
	
	// operations ------------------------------------------------------
	
	/*! \brief Find intersection between two lines.
	 * \param result Pointer to a vector which should be overwritten with
	 *               the point of intersection.  If NULL, it will not be
	 *               overwritten.
	 * \param other  The line with which to test for intersection.
	 * \returns      True if a point of intersection was found, or false if
	 *               the lines did not intersect or if they were equal.
	 */
	bool FindIntersection(GVector* result, const GLine& other) const;
	
	/*! \brief test if a point is on the line.
	 * \returns True if the given point was on the line or false otherwise. */
	bool Contains(const GVector& point) const;
	
	//! Finds the shortest distance between the line and a point.
	GCoordinate FindDistance(const GVector& point) const;
	/*! \brief Finds the square of the shortest distance between the line and a point (faster).
	 *
	 * This is faster because no square root needs to be performed. */
	GCoordinate FindSquareDistance(const GVector& point) const;
	//! Finds the shortest distance between two lines.
	GCoordinate FindDistance(const GLine& other) const;
	/*! \brief Finds the square of the shortest distance between two lines (faster).
	 *
	 * This is faster because no square root needs to be performed. */
	GCoordinate FindSquareDistance(const GLine& point) const;
	
	// operators -------------------------------------------------------
	//! Moves the line.
	const GLine& operator+=(const GVector& other);
	//! Moves the line.
	const GLine& operator-=(const GVector& other);
};

// inline member functions ===========================================
// lifecycle ---------------------------------------------------------
inline
GLine::GLine()
{
}

// operations --------------------------------------------------------
inline bool
GLine::Contains(const GVector& point) const
{
	return u.X(point) == v;
}

} // namespace GAUGE3D

#endif //included
