/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_QUATERNION_H__INCLUDED__
#define __GAUGE3D_3DMATH_QUATERNION_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include <gauge3d/3dmath/vector.h>
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GQuaternion quaternion.h gauge3d/3dmath/quaternion.h
 * \ingroup d3math
 * \brief Represents a quaternion, or an axis-angle rotation.
 *
 * A quaternion is a complex number with one real component and three imaginary
 * components.  But you don't need to worry about that.  All you care about
 * is the fact that quaternions are an efficient way to represent axis-angle
 * rotations. */
class GAUGE3D_API GQuaternion
{
public:
	/*! \name Coordinates */
	//@{
	GCoordinate x;  //!< imaginary component 1
	GCoordinate y;  //!< imaginary component 2
	GCoordinate z;  //!< imaginary component 3
	GCoordinate w;  //!< real component
	//@}
	
	// lifecycle -------------------------------------------------------
	//! Creates an uninitialized quaternion.
	GQuaternion();
	/*! \brief Creates a quaternion that represents the given axis-angle rotation.
	 *         The axis must be normalized! */
	GQuaternion(const GVector& axis, GCoordinate angle);
	
	// operations ------------------------------------------------------
	/*! \brief Applies the quaternion rotation to a set of vectors.
	 *  \param in  An array of vectors to rotate.
	 *  \param out Location to which to write the rotated vectors.  May be
	 *             the same as in.
	 *  \param num Number of vectors to rotate.
	 *  \remarks Rotating many vectors at once is faster than rotating vectors individually.
	 */
	void Rotate(const GVector in[], GVector out[], int num) const;
	
	//! Rotate a single vector in place.
	inline void Rotate(GVector* vec) const
	{
		Rotate(vec, vec, 1);
	}
	
	/*! \brief Spherical linear interpolation between quaternions.
	 *
	 * Interpolate() interpolates along a sphere between q1 and q2 using
	 * time value t, which is between 0.0 (q1), and 1.0 (q2).  The
	 * results are stored in this.  This is useful if you want to find
	 * the orientation between two quaternions, or want to animate an object
	 * moving between these two orientations. */
	void Interpolate(const GQuaternion& q1, const GQuaternion& q2, GCoordinate t);
	
	// access ----------------------------------------------------------
	//note that these functions have to extract the return values from the
	// quaternion, so they are not extremely fast.
	GCoordinate Angle() const; //!< Extracts the angle of the quaternion rotation.
	GVector Axis() const;      //!< Extracts the axis of the quaternion rotation.
	GVector RawAxis() const;   //!< Returns the axis without normalizing it.  Faster.
	
	/*! \brief Sets the quaternion to represent the given axis-angle rotation.
	 *         The axis must be a unit vector! */
	void Set(const GVector& axis, GCoordinate angle);
	
	// operators -------------------------------------------------------
	/*! \brief Combines two quaternion rotations into one.
	 *
	 * Applying the resulting quaternion to a vector will have the same effect
	 * as if you had first applied the second operand of the multiplication to
	 * the vector, and then the first operand.  This is similar to matrix
	 * multiplication. */
	GQuaternion operator*(const GQuaternion& other) const;
	/*! \brief Combines two quaternion rotations into one.
	 *
	 * Applying the resulting quaternion to a vector will have the same effect
	 * as if you had first applied the second operand of the multiplication to
	 * the vector, and then the first operand.  This is similar to matrix
	 * multiplication. */
	const GQuaternion& operator*=(const GQuaternion& other);
};

//inline member functions ============================================
inline GQuaternion::GQuaternion() {}

inline GVector GQuaternion::RawAxis() const
{
	return GVector(x, y, z);
}

} //namespace GAUGE3D

#endif //included
