/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_STRING_H__INCLUDED__
#define __GAUGE3D_STRING_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
#ifdef PLATFORM_MSVC
//MSVC needs you to actually include stdlib.h to get wchar_t to work.
// Funny, since Microsoft seems to like unicode so much.
#ifndef __SYSTEM_STDLIB_H__INCLUDED__
#include <stdlib.h>
#define __SYSTEM_STDLIB_H__INCLUDED__
#endif
#endif

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GString gstring.h gauge3d/gstring.h
 *  \ingroup common
 *  \brief A unicode-capable string class.
 *
 *  GString can represent both ASCII and Unicode strings and it can convert
 *  between them.  The internal buffer is reference-counted, so it is
 *  perfectly performance-safe to pass strings by value.  When a write
 *  operation on the string occurs, the internal buffer will be copied if
 *  and only if it does not have a reference count of one. */
class GAUGE3D_API GString
{
	class Buffer;
	Buffer* mBuffer;
	
	void SetBuffer(Buffer* buffer);

public:
	// lifecycle -------------------------------------------------------
	//! Create an empty (zero-length) string.
	GString();
	//! Copy constructor.
	GString(const GString& other);
	//! Construct from ASCII character array.
	GString(const char    cstr[]);
	//! Construct from Unicode character array.
	GString(const wchar_t cstr[]);
	//! Construct from ASCII character array of given length (in characters).
	GString(const char    cstr[], int length);
	//! Construct from Unicode character array of given length (in wide characters).
	GString(const wchar_t cstr[], int length);
	//! Create a string representation of an int (base-10).
	explicit GString(int i);
	//! Create a string representation of a float (base-10).
	explicit GString(double f);
	~GString();
	
	// operations ------------------------------------------------------
	/*! \brief Compare one string to another.
	 *  \returns Negative if the string is less than <i>other</i>, positive if it
	 *           is greater than <i>other</i>, or zero if the strings are equal.
	 *  \remarks This function basically just calls strcmp (or wcscmp for Unicode)
	 *           and returns the result. */
	int Compare(const GString& other) const;
	//! Compare to a null-terminated char string.
	int Compare(const char other[]) const;
	//! Compare to a null-terminated wide char string.
	int Compare(const wchar_t other[]) const;
	
	//! Returns the index of the first instance of the given character in the string.
	int FindFirstOf(wchar_t c) const;
	//! Returns the index of the last instance of the given character in the string.
	int FindLastOf(wchar_t c) const;
	//! Returns the index of the first instance of the character occurring after the start position.
	int FindNext(wchar_t c, int start) const;
	
	/*! \brief Returns a sub string of the string.
	 *  \param start The index of the character at which to start the substring.
	 *  \returns The substring, or an empty string if <i>start</i> was out-of-range. */
	GString SubString(int start) const;
	/*! \brief Returns a sub string of the string.
	 *  \param start The index of the character at which to start the substring.
	 *  \param length The number of characters to place in the substring.
	 *  \returns The substring, or an empty string if <i>start</i> was out-of-range. */
	GString SubString(int start, int length) const;
	
	// access ----------------------------------------------------------
	//! Gets the length of the string in characters (or wide characters).
	int Length() const;
	
	/*! \brief Returns true if the string is represented as Unicode internally.
	 *  \remarks Internally, the string could be represented in either ASCII or
	 *           Unicode.  Do not assume to know what a string is without testing
	 *           it. */
	bool IsUnicode() const;
	
	/*! \brief Returns the same string, except with a Unicode internal representation.
	 *  \remarks You only need to use this if you are planning to call UnicodeData()
	 *           and the string is not already in Unicode format.  Note that if the
	 *           string <i>is</i> already in Unicode format, then this function will
	 *           return the same string, so calling IsUnicode() first is not faster. */
	GString ToUnicode() const;
	/*! \brief Returns the same string, except with an ASCII internal representation.
	 *  \remarks You only need to use this if you are planning to call AsciiData()
	 *           and the string is not already in ASCII format.  Note that if the
	 *           string <i>is</i> already in ASCII format, then this function will
	 *           return the same string, so calling IsUnicode() first is not faster.
	 *  \note If any non-ASCII characters are part of the string, they will be corrupted
	 *        by this function. */
	GString ToAscii() const;
	
	/*! \brief Returns a pointer to the internal buffer of the string.
	 *  \returns A pointer to a null-terminated ASCII character array, or NULL if
	 *           the string is not represented in ASCII internally.
	 *  \remarks Avoid doing operations on raw character arrays.  You should try to
	 *           use GString's member functions to manipulate the string.  If you must
	 *           operate on char arrays directly, remember to write versions of the
	 *           operation for both ASCII <b>and</b> Unicode! */
	const char*    AsciiData() const;
	/*! \brief Returns a pointer to the internal buffer of the string.
	 *  \returns A pointer to a null-terminated Unicode character array, or NULL if
	 *           the string is not represented in Unicode internally.
	 *  \remarks Avoid doing operations on raw character arrays.  You should try to
	 *           use GString's member functions to manipulate the string.  If you must
	 *           operate on char arrays directly, remember to write versions of the
	 *           operation for both ASCII <b>and</b> Unicode! */
	const wchar_t* UnicodeData() const;
	
	/*! \brief Set the given character in the string to the given value.
	 *  \note  This function is relatively slow (compared to real array access). */
	void SetChar(int index, wchar_t value);
	
	// operators -------------------------------------------------------
	//! Interpret string as an integer and convert.
	operator int();
	//! Interpret string as a floating point value and convert.
	operator float();
	
	/*! \brief Array-like access to the characters in the string.
	 *  \remarks Remember, this function does not return a reference to the
	 *           character, so you cannot assign to it.  Use SetChar() to
	 *           change a character if you must.
	 *  \note    This function is relatively slow (compared to real array access). */
	wchar_t operator[](int index) const;
	
	//! Assignment operator.
	const GString& operator=(const GString& other);
	
	/*! \name String concatination operators. */
	//@{
	GString operator+(const GString& other) const;
	const GString& operator+=(const GString& other);
	GString operator+(char other) const;
	const GString& operator+=(char other);
	
	GString operator+(char other[]) const;
	const GString& operator+=(char other[]);
	GString operator+(wchar_t other[]) const;
	const GString& operator+=(wchar_t other[]);
	//@}
	
	/*! \name String comparison operators. */
	//@{
	inline bool operator==(const GString& other) const
		{return Compare(other) == 0;}
	inline bool operator!=(const GString& other) const
		{return Compare(other) != 0;}
	inline bool operator<(const GString& other) const
		{return Compare(other) < 0;}
	inline bool operator>(const GString& other) const
		{return Compare(other) > 0;}
	inline bool operator<=(const GString& other) const
		{return Compare(other) <= 0;}
	inline bool operator>=(const GString& other) const
		{return Compare(other) >= 0;}
	
	inline bool operator==(const char other[]) const
		{return Compare(other) == 0;}
	inline bool operator!=(const char other[]) const
		{return Compare(other) != 0;}
	inline bool operator<(const char other[]) const
		{return Compare(other) < 0;}
	inline bool operator>(const char other[]) const
		{return Compare(other) > 0;}
	inline bool operator<=(const char other[]) const
		{return Compare(other) <= 0;}
	inline bool operator>=(const char other[]) const
		{return Compare(other) >= 0;}
	
	inline bool operator==(const wchar_t other[]) const
		{return Compare(other) == 0;}
	inline bool operator!=(const wchar_t other[]) const
		{return Compare(other) != 0;}
	inline bool operator<(const wchar_t other[]) const
		{return Compare(other) < 0;}
	inline bool operator>(const wchar_t other[]) const
		{return Compare(other) > 0;}
	inline bool operator<=(const wchar_t other[]) const
		{return Compare(other) <= 0;}
	inline bool operator>=(const wchar_t other[]) const
		{return Compare(other) >= 0;}
	//@}
};

/*! \relates GString */
GString operator+(const char* str1, const GString& str2);
/*! \relates GString */
GString operator+(const wchar_t str1[], const GString& str2);

} //namespace GAUGE3D

#endif //included
