/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_FILES_ISTREAM_H__INCLUDED__
#define __GAUGE3D_FILES_ISTREAM_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GRawIStream istream.h gauge3d/files/istream.h
 *  \ingoup files
 *  \brief A source of data.
 *
 *  GRawIStream is the actual source of data read using a GIStream.
 *  GIStream itself provides buffering and other services for reading
 *  the data more easily.  Often, however, it is desireable to read
 *  the raw data directory, especially when layering streams (you don't
 *  want double buffering).
 *
 *  GRawIStream is the class that should be inherited when creating
 *  new sources for stream data.  When inheriting from GRawIStream,
 *  you do not have to write implementations of all of the functions.
 *  For example, if your stream does not support seeking, you do not
 *  have to write Seek() or TellPos().  If input is always available
 *  for your class, you need not define WaitForInput() or InputAvailable().*/
class GAUGE3D_API GRawIStream: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f04);
public:
	virtual ~GRawIStream();
	
	//A derived class which does not support seeking need not define these functions.
	//! Position to seek relative to.
	enum tSeekType
	{
		BEGINNING, //!< Seek relative to the start of the stream.
		CURRENT,   //!< Seek relative to the current position.
		END        //!< Seek relative to the end of the stream.
	};
	/*! \brief Seek the stream position.
	 *  \param pos      Position to seek to, in bytes.
	 *  \param seekType Point in the stream that pos is relative to. */
	virtual void Seek(int pos, int seekType);
	//! Get the distance in bytes between the beginning of the stream and the current position.
	virtual int TellPos();
	
	//! Attempts to fill the buffer and returns the amount actually read.
	virtual int Read(void* buffer, int size) = 0;
	
	//A derived class for which input is always available need not define these functions.
	
	/*! \brief Wait for input to become available.
	 * \param timout Maximum amount of time to wait for input, or 0 to wait indefinately.
	 * \returns      True if input became available, false otherwise.
	 * \note This function may return false immediately if it knows that input will
	 *       not be available in the future.  For example, if the stream is a file and
	 *       it is at the end of the file. */
	virtual bool WaitForInput(tTime timeout = 0.0);
	
	//! Returns true if input is currently available.
	virtual bool InputAvailable();
};

typedef GSmartPointer<GRawIStream> pGRawIStream;

/*! \class GIStream istream.h gauge3d/files/istream.h
 *  \ingroup files
 *  \brief Provides many functions for reading data from a stream.
 *
 *  GIStream does not provide the stream data itself.  It wraps around a
 *  GRawIStream.  GIStream provides many different ways to read the
 *  data, however, whereas GRawIStream only lets you read the data
 *  raw.
 *
 *  You can use a GIStream in either binary or text mode.  Text mode works
 *  like regular C++ istreams, where numbers are read as ASCII, human-readable
 *  numbers, etc.  In binary mode, on the other hand, the numbers are read in
 *  actual binary code, where a 32-bit number should be exactly four bytes in
 *  size in the file.  Also, when in binary mode you can set whether to expect
 *  the data in big-endian or little-endian format.  The data will automatically
 *  be converted to the host's endianness when read. */
class GAUGE3D_API GIStream: public GObject
{
	struct Private;
	Private& mPrivate;
	
   GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f10);
public:
	// lifecycle -------------------------------------------------------
	/*! \brief Create a buffered input stream.
	 *  \param streamSource A pointer to a GRawIStream from which to read data. */
	GIStream(pGRawIStream streamSource);
	/*! \brief Create a possibly buffered input stream with a given buffer size.
	 *  \param streamSource A pointer to a GRawIStream from which to read data.
	 *  \param bufferSize   The size of the internal buffer in bytes.  You can set
	 *                      this to zero to create an unbuffered stream, but this
	 *                      will usually be slower unless you are reading data in
	 *                      large blocks.  Also, only buffered streams can be used
	 *                      in text (non-binary) mode. */
	GIStream(pGRawIStream streamSource, int bufferSize);
	//! Destructor.
	~GIStream();
	
	// operations ------------------------------------------------------
	//! Position to seek relative to.  Used by Seek().
	enum tSeekType
	{
		BEGINNING, //!< Seek relative to the start of the stream.
		CURRENT,   //!< Seek relative to the current position.
		END        //!< Seek relative to the end of the stream.
	};
	/*! \brief Seek the stream position.
	 *  \param pos      Position to seek to, in bytes.
	 *  \param seekType Point in the stream that pos is relative to. */
	void Seek(int pos, tSeekType seekType);
	//! Get the distance in bytes between the beginning of the stream and the current position.
	int TellPos();
	
	//! Read one or more uint64's into the target buffer.  Returns the number successfully read.
	int Read(uint64*  target, int num = 1);
	//! Read one or more uint32's into the target buffer.  Returns the number successfully read.
	int Read(uint32*  target, int num = 1);
	//! Read one or more uint16's into the target buffer.  Returns the number successfully read.
	int Read(uint16*  target, int num = 1);
	
	//! Read one or more int64's into the target buffer.  Returns the number successfully read.
	int Read(int64*   target, int num = 1);
	//! Read one or more int32's into the target buffer.  Returns the number successfully read.
	int Read(int32*   target, int num = 1);
	//! Read one or more int16's into the target buffer.  Returns the number successfully read.
	int Read(int16*   target, int num = 1);
	
	//! Read one or more float64's into the target buffer.  Returns the number successfully read.
	int Read(float64* target, int num = 1);
	//! Read one or more float32's into the target buffer.  Returns the number successfully read.
	int Read(float32* target, int num = 1);
	
	//! Read one or more char's into the target buffer.  Returns the number successfully read.
	int Read(char*    target, int num = 1);
	
	/*! \brief Read one or more strings into the target buffer.  Returns the number successfully read.
	 * \param target     A pointer to an array of pointers to arrays of char's to which to
	 *                   write the strings.
	 * \param num        The number of strings to read.
	 * \param scanCode   Specifies what characters to include in the string.  It has the
	 *                   same syntax as the scanf %[] conversion specifier, without the
	 *                   brackets.  So, "^\n" would say to stop when a newline is
	 *                   encountered.  By default, the set of all non-whitespace
	 *                   characters is used.
	 * \param stringSize The number of characters in each character array.  If the string
	 *                   being read into one particular character array is longer than this,
	 *                   the remainder will be ignored.  Each array will always be
	 *                   zero-terminated.  A value of -1 indicates that no limit should be
	 *                   placed on the size of the strings read (not recommended!). */
	int Read(char*    target[], int num = 1, char* scanCode = NULL, int stringSize = -1);
	/*! \brief Read one or more strings into the target buffer.  Returns the number successfully read.
	 * \param target     A pointer to an array of strings to which to write the data read.
	 * \param num        The number of strings to read.
	 * \param scanCode   Specifies what characters to include in the string.  It has the
	 *                   same syntax as the scanf %[] conversion specifier, without the
	 *                   brackets.  So, "^\n" would say to stop when a newline is
	 *                   encountered.  By default, the set of all non-whitespace
	 *                   characters is used. */
	int Read(GString*  target, int num = 1, char* scanCode = NULL);
	
	//! Read <i>num</i> bytes of raw data.
	int Read(void* target, int num);
	
	/*! \brief Check what the next byte in the stream is without extracting it.
	 *  \note This can only be done with buffered streams. */
	char Peek();
	/*! \brief Get the next byte in the stream.
	 *  \note This can only be done with buffered streams. */
	char Get();
	
	/*! \brief Wait for input to become available.
	 * \param timout Maximum amount of time to wait for input, or 0 to wait indefinately.
	 * \returns      True if input became available, false otherwise.
	 * \note This function may return false immediately if it knows that input will
	 *       not be available in the future.  For example, if the stream is a file and
	 *       it is at the end of the file. */
	bool WaitForInput(tTime timeout = 0.0);
	
	// access ----------------------------------------------------------
	//! Read mode specifiers for ReadMode().
	enum tReadMode
	{
		BINARY,       //!< Binary mode reading.
		ASCII         //!< ASCII text mode reading.
	};
	
	/*! \breif Set how to interpret the data in the stream (binary or text).
	 *  \param mode One of \ref tReadMode. */
	void ReadMode(tReadMode mode);
	//! Get the current read mode.
	tReadMode ReadMode();
	
	//Endianness() tells the istream whether the input is big endian or
	// little endian.  It will then convert the input as appropriate when
	// you read it in.  This only applies to binary streams.
	
	//! Endianness specifiers for Endianness().
	enum tEndianness
	{
		BIG,   //!< Big endian.  Most significant byte comes first.  PPC processors are big endian.
		LITTLE //!< Little endian.  Least significant byte comes first.  Intel processors are little endian.
	};
	//! Set the endianness of the input data.  Only applies to binary streams.
	void Endianness(tEndianness endianness);
	//! Get the endianness of the input data.
	tEndianness Endianness();
	
	//NumberBase() sets the number base to read text numbers in.  This
	// only applies to text mode, not binary.
	
	//! Number base specifiers for NumberBase().
	enum tNumberBase
	{
		OCT = 8,  //!< Octal (base 8).
		DEC = 10, //!< Decimal (base 10).
		HEX = 16  //!< Hexidecimal (base 16).
	};
	//! Set the number base of the input data.  Only applies to text streams.
	void NumberBase(tNumberBase base);
	//! Get the number base of the input data.
	tNumberBase NumberBase();
	
	// inquiry ---------------------------------------------------------	
	//InputAvailable returns true if there is input available now.
	// This can be used to test for end-of-file.
	
	/*! \brief Returns true if input is currently available.
	 *  \remarks You can use this to test if you are at the end of a file. */
	bool InputAvailable();
	
	/*! \brief Use to check if the last read operation succeeded.
	 *
	 *  \remarks
	 * Using the stream as a truth value returns true if the last read
	 * operation succeeded.  This way, you can say something like:<br>
	 * <ul><li>while(myStream >> data) {...}</ul>
	 * to keep reading in and processing data until there is no longer data
	 * available. */
	operator void*();
};

/*! \typedef pGIStream
 *  \ingroup files
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GIStream GIStream\endlink. */
typedef GSmartPointer<GIStream> pGIStream;

/*! \brief The >> operator works the same as it does for C++ istreams.
 *  \ingroup files
 *  \relates GIStream */
template <class tType>
inline GIStream& operator>>(GIStream& is, tType& target)
	{ is.Read(&target); return is; }
inline GIStream& operator>>(GIStream& is, char* target)
	{ is.Read(&target); return is; }

} // namespace GAUGE3D

#endif //included
