/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_FILES_OSTREAM_H__INCLUDED__
#define __GAUGE3D_FILES_OSTREAM_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GRawOStream ostream.h gauge3d/files/ostream.h
 *  \ingoup files
 *  \brief A destination for data.
 *
 *  GRawOStream is the actual destination of data written using a GOStream.
 *  GOStream itself provides buffering and other services for writing
 *  the data more easily.  Often, however, it is desireable to write
 *  the raw data directory, especially when layering streams (you don't
 *  want double buffering).
 *
 *  GRawOStream is the class that should be inherited when creating
 *  new destinations for stream data.  When inheriting from GRawOStream,
 *  you do not have to write implementations of all of the functions.
 *  If your stream does not support seeking, you do not
 *  have to write Seek() or TellPos(). */
class GAUGE3D_API GRawOStream: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f1c);
public:
	virtual ~GRawOStream();
	
	//! Position to seek relative to.
	enum tSeekType
	{
		BEGINNING, //!< Seek relative to the start of the stream.
		CURRENT,   //!< Seek relative to the current position.
		END        //!< Seek relative to the end of the stream.
	};
	/*! \brief Seek the stream position.
	 *  \param pos      Position to seek to, in bytes.
	 *  \param seekType Point in the stream that pos is relative to. */
	virtual void Seek(int pos, int seekType);
	//! Get the distance in bytes between the beginning of the stream and the current position.
	virtual int TellPos();
	
	//! Attempts to write the contents of buffer and returns the amount actually written.
	virtual int Write(const void* buffer, int size) = 0;
};

typedef GSmartPointer<GRawOStream> pGRawOStream;

/*! \class GOStream ostream.h gauge3d/files/ostream.h
 *  \ingroup files
 *  \brief Provides many functions for writing data to a stream.
 *
 *  GOStream does not provide the stream destination itself.  It wraps around a
 *  GRawOStream.  GOStream provides many different ways to write the
 *  data, however, whereas GRawOStream only lets you write the data
 *  raw.
 *
 *  You can use a GOStream in either binary or text mode.  Text mode works
 *  like regular C++ ostreams, where numbers are written as ASCII, human-readable
 *  numbers, etc.  In binary mode, on the other hand, the numbers are written in
 *  actual binary code, where a 32-bit number will be exactly four bytes in
 *  size in the file.  Also, when in binary mode you can set whether to write
 *  the data in big-endian or little-endian format.  The data will automatically
 *  be converted from the host's endianness when written. */
class GAUGE3D_API GOStream: public GObject
{
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f25);
public:
	// lifecycle -------------------------------------------------------
	/*! \brief Create a buffered output stream.
	 *  \param streamDest A pointer to a GRawOStream to which to write data. */
	GOStream(pGRawOStream streamDest);
	/*! \brief Create a buffered output stream.
	 *  \param streamDest A pointer to a GRawOStream to which to write data.
	 *  \param bufferSize   The size of the internal buffer in bytes.  You can set
	 *                      this to zero to create an unbuffered stream, but this
	 *                      will usually be slower unless you are writing data in
	 *                      large blocks.  Also, only buffered streams can be used
	 *                      in text (non-binary) mode. */
	GOStream(pGRawOStream streamDest, int bufferSize);
	//! Destructor.
	virtual ~GOStream();
	
	// operations ------------------------------------------------------
	//! Position to seek relative to.  Used by Seek().
	enum tSeekType
	{
		BEGINNING, //!< Seek relative to the start of the stream.
		CURRENT,   //!< Seek relative to the current position.
		END        //!< Seek relative to the end of the stream.
	};
	/*! \brief Seek the stream position.
	 *  \param pos      Position to seek to, in bytes.
	 *  \param seekType Point in the stream that pos is relative to. */
	void Seek(int pos, tSeekType seekType);
	//! Get the distance in bytes between the beginning of the stream and the current position.
	int TellPos();
	
	//! Write one or more uint64's from the source buffer.  Returns the number successfully written.
	int Write(const uint64*  source, int num = 1);
	//! Write one or more uint32's from the source buffer.  Returns the number successfully written.
	int Write(const uint32*  source, int num = 1);
	//! Write one or more uint16's from the source buffer.  Returns the number successfully written.
	int Write(const uint16*  source, int num = 1);
	
	//! Write one or more int64's from the source buffer.  Returns the number successfully written.
	int Write(const int64*   source, int num = 1);
	//! Write one or more int32's from the source buffer.  Returns the number successfully written.
	int Write(const int32*   source, int num = 1);
	//! Write one or more int16's from the source buffer.  Returns the number successfully written.
	int Write(const int16*   source, int num = 1);
	
	//! Write one or more float64's from the source buffer.  Returns the number successfully written.
	int Write(const float64* source, int num = 1);
	//! Write one or more float32's from the source buffer.  Returns the number successfully written.
	int Write(const float32* source, int num = 1);
	
	//! Write one or more char's from the source buffer.  Returns the number successfully written.
	int Write(const char*    source, int num = 1);
	
	/*! \brief Write one or more strings from the source buffer.
	 *         Returns the number successfully written.
	 *  \param source An array of pointers to null-terminated character strings.
	 *  \param num    The number of strings to write.
	 *  \returns      The number of strings successfully written. */
	int Write(const char*    source[], int num = 1);
	/*! \brief Write one or more strings from the source buffer.
	 *         Returns the number successfully written.
	 *  \param source An array of strings.
	 *  \param num    The number of strings to write.
	 *  \returns      The number of strings successfully written. */
	int Write(const GString*  source, int num = 1);
	
	//! Write <u>num</u> bytes of raw data.
	int Write(const void* source, int num);
	
	/*! \brief Write a character to the stream.
	 *  \note This can only be done with buffered streams. */
	void Put(char c);
	
	/*! \brief Write any data that is in the buffer to the output stream.
	 *  \returns The number of bytes still in the buffer (should always be
	 *           zero). */
	int Flush();
	
	// access ----------------------------------------------------------
	//! Write mode specifiers for WriteMode().
	enum tWriteMode
	{
		BINARY,       //!< Binary mode writing.
		ASCII         //!< ASCII text mode writing.
	};
	
	/*! \breif Set how to output the data in the stream (binary or text).
	 *  \param mode One of \ref tWriteMode. */
	void WriteMode(tWriteMode mode);
	//! Get the current write mode.
	tWriteMode WriteMode();
	
	/*! \brief Specify a character to place after each item.
	 *  \param separator A character that should be placed after every item
	 *                   that is written.
	 *  \param enable    Whether or not to enable this functionality.
	 *  \remarks This allows you to output an array of numbers without
	 *           getting them all mushed together.
	 *  \note    This only applies to test i/o, not binary.  The separator
	 *           will <i>not</i> be used after Put() calls (as the other
	 *           write calls use Put() for text output). */
	void Separator(char separator, bool enable);
	//! Get the current separator character.
	char Separator();
	
	//! Endianness specifiers for Endianness().
	enum tEndianness
	{
		BIG,   //!< Big endian.  Most significant byte comes first.  PPC processors are big endian.
		LITTLE //!< Little endian.  Least significant byte comes first.  Intel processors are little endian.
	};
	//! Set the endianness in which to write the output data.  Only applies to binary streams.
	void Endianness(tEndianness endianness);
	//! Get the endianness of the output data.
	tEndianness Endianness();
	
	//! Number base specifiers for NumberBase().
	enum tNumberBase
	{
		OCT = 8,  //!< Octal (base 8).
		DEC = 10, //!< Decimal (base 10).
		HEX = 16  //!< Hexidecimal (base 16).
	};
	//! Set the number base in which to write the output data.  Only applies to text streams.
	void NumberBase(tNumberBase base);
	//! Get the number base of the output data.
	tNumberBase NumberBase();
	
	// inquiry ---------------------------------------------------------
	//! Use to check if the last write operation succeeded.
	operator void*();
};
/*! \typedef pGOStream
 *  \ingroup files
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GOStream GOStream\endlink. */
typedef GSmartPointer<GOStream> pGOStream;

/*! \brief The << operator works the same as it does for C++ ostreams.
 *  \ingroup files
 *  \relates GOStream */
template <class tType>
inline GOStream& operator<<(GOStream& os, const tType& source)
	{ os.Write(&source); return os; }
inline GOStream& operator<<(GOStream& os, const char* source)
	{ os.Write(&source); return os; }

} // namespace GAUGE3D

#endif //included
