/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_GRAPHICS_MODEL_H__INCLUDED__
#define __GAUGE3D_GRAPHICS_MODEL_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_QUATERNION_H__INCLUDED__
#include <gauge3d/3dmath/quaternion.h>
#endif

#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include <gauge3d/common/array.h>
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_PRIMITIVES_H__INCLUDED__
#include <gauge3d/3dmath/primitives.h>
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_GRAPHICS_IMAGE_H__INCLUDED__
#include "image.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GModel model.h gauge3d/graphics/model.h
 *  \ingroup graphics
 *  \brief Base class for 3D model plugins.
 *
 *  Plugins that allow GAUGE to load various 3D model formats (like MD2
 *  from Quake 2) are derived from this class. */
class GAUGE3D_API GModel: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781e94);
public:
	virtual ~GModel();
	
	/*! \class AnimState model.h gauge3d/graphics/model.h
	 *  \brief The animation state for a GModel.
	 *
	 *  AnimState stores the current position of a model's animation.  This
	 *  allows several objects to use the same model but be in different
	 *  positions (i.e. one player is walking while another is jumping).
	 *  Each object would have a different AnimState, but the same GModel. */
	class GAUGE3D_API AnimState: public GObject
	{
		GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781ea5);
	public:
		virtual ~AnimState();

#ifdef DOXYGEN
		//! This interface is not done yet.
		Under Construction;
#endif
		// control info ------------------------------------------------
		//TODO: animation name, velocity, timer?
		// i.e.:
		//void UseTimer(pGTimer timer);
		//void SetAnimation(tUID name, GVector velocity);
	};
	/*! \typedef pAnimState
	 *  \brief \link GSmartPointer Smart pointer\endlink to a
	 *         \link AnimState AnimState\endlink. */
	typedef GSmartPointer<AnimState> pAnimState;
	
	//! Generates an AnimState object associated with this model.
	virtual pAnimState CreateAnimState() = 0;
	
	// rendering info ----------------------------------------------
	/*! \class RenderInfo model.h gauge3d/graphics/model.h
	 *  \brief Provides information required to render a model.
	 *
	 *  A RenderInfo object is bound to a particular GModel object and is
	 *  used to request the information necessary to render the model.
	 *  The renderer will need one RenderInfo for each sprite in the
	 *  scene which uses a particular model.  The idea here is to allow
	 *  safe multithreaded access to this information.  (So, the renderer
	 *  could be drawing two different sprites that have the same model at
	 *  the same time.) */
	class GAUGE3D_API RenderInfo: public GObject
	{
		GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x3a2094d5);
	public:
		virtual ~RenderInfo();
		
		/*! \struct RenderSet model.h gauge3d/graphics/model.h
		 *  \brief Contains all of the data needed to render a model.
		 *
		 *  A RenderSet contains a list of all of the vertices in a GModel.  For
		 *  each vertex, it gives the position, surface normal, and texture coordinates.
		 *  These lists are all strided arrays and can be interleaved.  Interleaving your
		 *  arrays can increase performance.
		 *
		 *  The RenderSet also contains a list of indices forming triangles, lines, or points. */
		struct RenderSet
		{
			/*! \brief Specifies which properties of the model are dynamic.
			 *  \remarks Typically, vertex positions and normals are dynamic (if anything). */
			enum tDynamicFlags
			{
				STATIC          = 0,       //!< Model is not dynamic.
				VERTEX_POSITION = 1 << 0,  //!< Vertex positions (and normals) are dynamic.
				VERTEX_TEXCOORD = 1 << 1,  //!< Vertex texture coordinates are dynamic.
				INDEX_LIST      = 1 << 16  //!< Index list is dynamic.
			};
			//! One or more of tDynamicFlags.
			int mDynamicFlags;
			
			//! Number of elements in each of the arrays.
			int mNumVertices;
			//! Array of vertex positions.
			GVector* mVertices;
			//! Array of normal vectors.  Must be unit vectors!
			GVector* mNormals;
			//! Array of texture coordinates.  Can be NULL if texturing is not supported.
			GVector* mTexCoords;
			
			//! Specifies how to make geometry out of the indices.
			enum tIndexType
			{
				POINTS,      //!< Every indice is a point / particle
				LINES,       //!< Every two indices make one line segment
				TRIANGLES    //!< Every three indices make one triangle
			};
			
			//! One of the values in tIndexType.
			int mIndexType;
			//! The number of elements in the index array.
			int mNumIndices;
			//! Pointer to an array of indices.
			uint16* mIndices;
			
			/*! \brief Called by the renderer to reset the RenderSet.
			 *  \remarks The idea is that older model plugins may not support newer
			 *           RenderSet data fields.  In order to make sure the renderer
			 *           does not read uninitialized data from those fields, they are
			 *           set to some default value.  This is all for binary and source
			 *           code backwards compatibility, but also makes writing model
			 *           plugins easier. */
			inline void SetDefaults()
			{
				mDynamicFlags = STATIC;
				mNumVertices = 0;
				mVertices = mNormals = mTexCoords = NULL;
				mIndexType = TRIANGLES;
				mNumIndices = 0;
				mIndices = NULL;
			}
		};
		
		/*! \brief Sets everything up to render a frame with the given parameters.
		 *  \param animState The animation state of the model.
		 *  \param viewPoint The position of the camera.
		 *  \param geoLod    The geometric level of detail in which to render the model.
		 *                   This is simply the area that an average triangle in the model
		 *                   should have, if it were one meter (one unit) from the camera.
		 *                   This value is *not* adjusted for distance from the viewpoint,
		 *                   so the model will have to do that internally.  (hint:  multiply
		 *                   by viewPoint.SquareMagnitude().)  Note that a lower geoLOD value
		 *                   means a higher detail image (smaller triangles).
		 *  \returns The number of RenderSet structures needed to complete the model.  The
		 *           renderer will then call GetStaticInfo and GetDynamicInfo once for
		 *           each RenderSet. */
		virtual int StartFrame(pAnimState animState, const GVector& viewPoint,
		                       const GQuaternion& viewAngle, const GFrustum& viewFrustum,
		                       float geoLod) = 0;
		
		/*! \brief Fills in the static parts of the render set.
		 *  \param target The RenderSet to which the info will be written.
		 *  \param setNum The number of the RenderSet which is to be filled in.
		 *
		 *  The renderer will use the values of the mDynamicFlags, mNumVertices, and
		 *  mNumIndices fields of <i>target</i> to allocate buffers to store the dynamically
		 *  generated parts of the model.  Then, GetDynamicInfo will be called to fill
		 *  in those buffers with the model data. */
		virtual void GetStaticInfo(RenderSet* target, int setNum) = 0;
		
		/*! \brief Fills in the dynamic parts of the render set.
		 *  \param target The RenderSet to which the info will be written.
		 *  \param setNum The number of the RenderSet which is to be filled in.
		 *
		 *  Before calling this, the renderer will fill in <i>target</i> with pointers
		 *  to buffers where the dynamic data for the model should be written.  For example,
		 *  if target->mDynamicFlags was set to VERTEX_POSITION by GetStaticInfo, then when
		 *  GetDynamicInfo is called, target->mVertices and target->mNormals will point to
		 *  arrays of vertices allocated by the renderer, ready to be filled in with vertex
		 *  data.  This way, the renderer can efficiently handle allocating memory for dynamic
		 *  model data.
		 *
		 *  \remarks When writing to the buffers which the renderer has allocated, a model
		 *           plugin should write linearly and should not attempt to read from the
		 *           buffers.  This is because the buffers may be allocated in AGP RAM or
		 *           even VRAM, and are therefore uncached.  This means that the CPU will
		 *           be able to write much more efficiently if it can combine several writes
		 *           into one. */
		virtual void GetDynamicInfo(RenderSet* target, int setNum) = 0;
	};
	/*! \typedef pRenderInfo
	 *  \brief \link GSmartPointer Smart pointer\endlink to a
	 *         \link RenderInfo RenderInfo\endlink. */
	typedef GSmartPointer<RenderInfo> pRenderInfo;
	
	/*! \brief Create a RenderInfo bound to this model.
	 *
	 *  \remarks Obviously, this function uses dynamic memory and should not be
	 *           called every frame.  A renderer should call this once when the model
	 *           is assigned to a sprite and then use the same RenderInfo object every
	 *           time the sprite is drawn. */
	virtual pRenderInfo CreateRenderInfo() = 0;
	
	// physics info ------------------------------------------------
	//TODO: OBB-trees?
};
/*! \typedef pGModel
 *  \ingroup graphics
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GModel GModel\endlink. */
typedef GSmartPointer<GModel> pGModel;

} // namespace GAUGE3D

#endif //included
