/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_RENDERER_DISPLAY_H__INCLUDED__
#define __GAUGE3D_RENDERER_DISPLAY_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include <gauge3d/3dmath/vector.h>
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_ANGLES_H__INCLUDED__
#include <gauge3d/3dmath/angles.h>
#endif //redundant include guard

#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include <gauge3d/common/array.h>
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_RENDERER_CAMERA_H__INCLUDED__
#include "camera.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GDisplay camera.h gauge3d/renderer/camera.h
 *  \ingroup renderer
 *  \brief Represents a two dimentional surface to which scenes can be rendered.
 *
 * This is basically the frame buffer of the graphics card, though in some
 * more interesting cases it might represent only a portion of the framebuffer
 * (e.g. for a split-screen effect) or even a texture (render-to-texture). */
class GAUGE3D_API GDisplay: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781fcd);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GDisplay();
	
	// operations ------------------------------------------------------
	/*! \brief Resets the display resolution and color depth.
	 *  \param width      The width, in pixels, of the desired display.
	 *  \param height     The height, in pixels, of the desired display.
	 *  \param depth      The number of bits per pixel of the desired display.
	 *  \param fullScreen True if the display should be fullscreen.  False otherwise.
	 *
	 * This function must be called once before any rendering can be done.
	 * If the display could not be openned, false is returned.
	 * Passing zero for colorDepth will cause the current display depth to be used.
	 *
	 * \note This function will be removed before GAUGE v1.0.  By that time,
	 *       the user will be able to configure the display type themselves
	 *       through the rendering plugin's configuration dialog. */
	virtual bool Open(int width, int height, int depth, bool fullScreen) = 0;
	
#ifndef DOXYGEN
	//I think this is self-explanitory...
	//This function will be removed soon, and is not documented because it
	// will probably never be implemented anyway.
	struct DisplayProps
	{
		int  mWidth;
		int  mHeight;
		int  mColorDepth;
		bool mFullScreen;
	};
	virtual GArray<DisplayProps> ListAvailableDisplays() = 0;
#endif
	
	/*! \brief Sets the viewpoint from which to render the display.
	 * \remarks Obviously, the camera must have come from a GScene object which came
	 *          from the same renderer as this GDisplay object.  If you don't follow
	 *          this rule, the renderer is free to crash your program. :) */
	virtual void Camera(pGCamera camera) = 0;
	
	/*! \brief Set whether the display should capture mouse input.
	 *
	 *  This controls whether the mouse is treated as controling a pointer on the
	 *  screen or if the mouse input is captured raw.  When mouse capturing is
	 *  enabled, mouse input messages will no longer contain absolute display
	 *  coordinates, but the mouse can move infinately in any direction without
	 *  hitting the side of the display or leaving the display window.  When off,
	 *  the mouse just controls a pointer that moves about the screen.
	 *
	 *  By default, the mouse is not captured. */
	virtual void CaptureMouse(bool enable) = 0;
};
/*! \typedef pGDisplay
 *  \ingroup renderer
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GDisplay GDisplay\endlink. */
typedef GSmartPointer<GDisplay> pGDisplay;

} // namespace GAUGE3D

#endif //included
