/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_RENDERER_LIGHT_H__INCLUDED__
#define __GAUGE3D_RENDERER_LIGHT_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include <gauge3d/3dmath/vector.h>
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GLight light.h gauge3d/renderer/light.h
 *  \ingroup renderer
 *  \brief Represents a light source in a 3D scene. */
class GAUGE3D_API GLight: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781fb6);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GLight();
	
	// operations ------------------------------------------------------
	/*! \brief Turn on the light.
	 *
	 * All lights start out hidden.  After you have cached the light, call Show() to
	 * enable it.  Show and Hide are a good way to turn lights on and off. */
	virtual void Show() = 0;
	//! Turn off the light.
	virtual void Hide() = 0;
	//! Permenantly remove the light from the scene.
	virtual void Remove() = 0;
	
	// access ----------------------------------------------------------
	//! Set the position of the light, and make the light local.
	virtual void Position(const GVector& position) = 0;
	
	//! Set the direction of the light, and make the light infinite.
	virtual void Direction(const GVector& direction) = 0;
	
	//! Set the color of the light.
	virtual void Color(float red, float green, float blue) = 0;
	//! Set the color of the light. (3-element array: RGB)
	virtual void Color(const float color[]) = 0;
	
	/*! \brief Set the properties of the light.
	 *  \param ambient  The ambient light level.  This represets the light that is not coming
	 *                  directly from the source, but has bounced off a few walls first.
	 *                  It is simulated through simple distance attenuation.
	 *  \param diffuse  The diffuse light level.  This represets the light that is coming
	 *                  directly from the source.  The orientation of the surface relative to
	 *                  the light source affects the level of the light, but the viewpoint
	 *                  does not matter.
	 *  \param specular The specular light level.  This represents light that has been
	 *                  reflected off the surface of the object.  Both the orientation of the
	 *                  object's surface as well as the viewpoint are taken into account. */
	virtual void LightProps(float ambient, float diffuse, float specular = 0.0) = 0;
	//! Set the properties of the light.  (ambient and diffuse)
	virtual void LightProps2v(const float lightProps[]) = 0;
	//! Set the properties of the light.  (ambient, diffuse, and specular)
	virtual void LightProps3v(const float lightProps[]) = 0;
	
	//! Sets linear and quadratic attenuation factors.  (makes light dimmer with distance)
	virtual void Attenuation(float linear, float quadratic = 0.0) = 0;
	
	/*! \brief Make the light act like a spotlight.
	 *  \param direction The direction of the spot light.
	 *  \param exponent  How focused the light is.  Setting this higher will make
	 *                   the center of the spot brighter and the edges dimmer.
	 *  \param width     The width of the spotlight in degrees.  Set to -1.0 to disable
	 *                   the spotlight effect. */
	virtual void Spotlight(const GVector& direction, float exponent, float width) = 0;
};
/*! \typedef pGLight
 *  \ingroup renderer
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GLight GLight\endlink. */
typedef GSmartPointer<GLight> pGLight;

} // namespace GAUGE3D

#endif //included
