/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: bmpimage.cpp
 *      Package: BMP image file plugin
 *    Component: ImgPak
 *  Description: Loads BMP Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "bmpimage.h"

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

struct BmpFileHeader {
	uint16 mType;
	uint32 mSize;
	uint32 mReserved;
	uint32 mOffsetBits;
};

GIStream& operator >> (GIStream& is, BmpFileHeader& hdr)
{
	is.Read(&hdr.mType, 1);
	is.Read(&hdr.mSize, 3);
	return is;
}

struct BmpInfoHeader {
	uint32 mSize;
	uint32 mWidth;
	uint32 mHeight;
	uint16 mPlanes;
	uint16 mBitCount;
	uint32 mCompression;
	uint32 mSizeImage;
	uint32 mXPelsPerMeter;
	uint32 mYPelsPerMeter;
	uint32 mClrUsed;
	uint32 mClrImportant;
};

GIStream& operator >> (GIStream& is, BmpInfoHeader& hdr)
{
	is.Read(&hdr.mSize, 3);
	is.Read(&hdr.mPlanes, 2);
	is.Read(&hdr.mCompression, 6);
	return is;
}

struct BmpCoreHeader {
	uint32 mSize;
	uint16 mWidth;
	uint16 mHeight;
	uint16 mPlanes;
	uint16 mBitCount;
};

GIStream& operator >> (GIStream& is, BmpCoreHeader& hdr)
{
	is.Read(&hdr.mSize, 1);
	is.Read(&hdr.mWidth, 4);
	return is;
}

struct RgbQuad {
	uint8 mBlue;
	uint8 mGreen;
	uint8 mRed;
	uint8 mReserved;
};

struct RgbTriple {
	uint8 mBlue;
	uint8 mGreen;
	uint8 mRed;
};

// Magic value constants
const uint16 BmpTypeValue = 0x4D42;		// BM
const uint32 BmpInfoHeaderSize = 40;
const uint32 BmpCoreHeaderSize = 12;

// Compression constants
const uint32 BI_RGB = 0;
const uint32 BI_RLE8 = 1;
const uint32 BI_RLE4 = 2;

enum {
	DATA_FORMAT_UNKNOWN,
	DATA_FORMAT_1BIT,
	DATA_FORMAT_4BIT,
	DATA_FORMAT_8BIT,
	DATA_FORMAT_24BIT
};
// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpBmpImage::GpBmpImage(pGFile bmpfile)
	: mReadOk(false), mDataFormat(DATA_FORMAT_UNKNOWN)
{
	BmpFileHeader bmpFileHeader;
	uint32 headerSize;
	
	pGRawIStream inSrc = bmpfile->GetRawInputStream();
	if(inSrc == NULL)
		return;
	GIStream input(inSrc);
	input.Endianness(GIStream::LITTLE);
	input >> bmpFileHeader;
	if(bmpFileHeader.mType != BmpTypeValue)
		return;
	input.Read(&headerSize,1);
	input.Seek(-sizeof(uint32),GIStream::CURRENT);
	if(headerSize == BmpInfoHeaderSize) {
		BmpInfoHeader bmpInfoHeader;
		int LineLength,BmpLineLengthDiff,h;

		input >> bmpInfoHeader;
		if(bmpInfoHeader.mBitCount <= 8) {
			int PaletteColors;
			if(bmpInfoHeader.mClrUsed > 0)
				PaletteColors = bmpInfoHeader.mClrUsed;
			else
				PaletteColors = 1 << bmpInfoHeader.mBitCount;
			RgbQuad *bmpPalette = new RgbQuad[PaletteColors];
			
			input.Read(bmpPalette,sizeof(RgbQuad) * PaletteColors);
			mPalette = GArray<uint8>(PaletteColors * 3);
			for(int i=0;i<PaletteColors;i++) {
				mPalette[i * 3] = bmpPalette[i].mRed;
				mPalette[i * 3 + 1] = bmpPalette[i].mGreen;
				mPalette[i * 3 + 2] = bmpPalette[i].mBlue;
			}
		}
		mFormat.mWidth = bmpInfoHeader.mWidth;
		mFormat.mHeight = bmpInfoHeader.mHeight;
		switch(bmpInfoHeader.mBitCount) {
		case 1:
			LineLength = (mFormat.mWidth + 7) / 8;
			mDataFormat = DATA_FORMAT_1BIT;
			break;
		case 4:
			LineLength = (mFormat.mWidth + 1) / 2;
			mDataFormat = DATA_FORMAT_4BIT;
			break;
		case 8:
			LineLength = mFormat.mWidth;
			mDataFormat = DATA_FORMAT_8BIT;
			break;
		case 24:
			LineLength = mFormat.mWidth * 3;
			mDataFormat = DATA_FORMAT_24BIT;
			break;
		}
		BmpLineLengthDiff = (((LineLength + 3) / 4) * 4) - LineLength;
		mData = GArray<uint8>(LineLength * mFormat.mHeight);
		for(h=mFormat.mHeight-1;h>=0;h--) {
			uint8 *Line = &mData[(h * LineLength)];
			switch(bmpInfoHeader.mCompression) {
			case BI_RGB:
				input.Read(Line,LineLength);
				if(BmpLineLengthDiff)
					input.Seek(BmpLineLengthDiff,GIStream::CURRENT);
				break;
			case BI_RLE8:
				do {
					uint8 byte;
					input.Read(&byte,1);
					if(byte) {
						// read run-length data
						uint8 color;
						input.Read(&color,1);
						memset(Line,color,byte);
						Line += byte;
					}
					else {
						// read escape character
						input.Read(&byte,1);
						if(byte == 0 || byte == 1) {
							break;	// end of line, or end of bitmap
						}
						else if(byte == 2) {
							uint8 delta[2];		// don't use deltas
							input.Read(delta,2);
						}
						else {
							// read absolute data
							Line += input.Read(Line,byte);
							if(byte % 2)	// ensure word alignment for next pair
								input.Read(&byte,1);
						}
					}
				} while(1);
				break;
			case BI_RLE4:
				{
					bool out_high = true,in_high;
					do {
						uint8 byte;

						input.Read(&byte,1);
						if(byte) {
							// read run-length data
							uint8 color;
							input.Read(&color,1);
							in_high = true;
							for(int c=0;c<byte;c++) {
								if(in_high) {
									if(out_high)
										*Line = color;
									else
										*(Line++) |= color >> 4;
								}
								else {
									if(out_high)
										*Line = color << 4;
									else
										*(Line++) |= color & 0x0F;
								}
								out_high = !out_high;
								in_high = !in_high;
							}
						}
						else {
							// read escape character
							input.Read(&byte,1);
							if(byte == 0 || byte == 1) {
								break;	// end of line, or end of bitmap
							}
							else if(byte == 2) {
								uint8 delta[2];		// don't use deltas
								input.Read(delta,2);
							}
							else {
								uint8 color;
								in_high = true;
								for(int c=0;c<byte;c++) {
									if(in_high) {
										input.Read(&color,1);
										if(out_high)
											*Line = color;
										else
											*(Line++) |= color >> 4;									
									}
									else {
										if(out_high)
											*Line = color << 4;
										else
											(*Line++) |= color & 0xF0;
									}
									in_high = !in_high;
									out_high = !out_high;
								}
								if(((byte + 1) / 2) % 2)
									input.Read(&byte,1);
							}
						}
					} while(1);
				}
				break;
			}
		}
	}
	else if(headerSize == BmpCoreHeaderSize) {
		BmpCoreHeader bmpCoreHeader;
		int LineLength,BmpLineLengthDiff;

		input >> bmpCoreHeader;
		if(bmpCoreHeader.mBitCount <= 8) {
			int PaletteColors = 1 << bmpCoreHeader.mBitCount;
			RgbTriple *bmpPalette = new RgbTriple[PaletteColors];
			
			input.Read(bmpPalette,sizeof(RgbTriple) * PaletteColors);
			mPalette = GArray<uint8>(PaletteColors * 3);
			for(int i=0;i<PaletteColors;i++) {
				mPalette[i * 3] = bmpPalette[i].mRed;
				mPalette[i * 3 + 1] = bmpPalette[i].mGreen;
				mPalette[i * 3 + 2] = bmpPalette[i].mBlue;
			}
		}
		mFormat.mWidth = bmpCoreHeader.mWidth;
		mFormat.mHeight = bmpCoreHeader.mHeight;
		switch(bmpCoreHeader.mBitCount) {
		case 1:
			LineLength = (mFormat.mWidth + 7) / 8;
			mDataFormat = DATA_FORMAT_1BIT;
			break;
		case 4:
			LineLength = (mFormat.mWidth + 1) / 2;
			mDataFormat = DATA_FORMAT_4BIT;
			break;
		case 8:
			LineLength = mFormat.mWidth;
			mDataFormat = DATA_FORMAT_8BIT;
			break;
		case 24:
			LineLength = mFormat.mWidth * 3;
			mDataFormat = DATA_FORMAT_24BIT;
			break;
		}
		BmpLineLengthDiff = (((LineLength + 3) / 4) * 4) - LineLength;
		mData = GArray<uint8>(LineLength * mFormat.mHeight);
		for(int h=mFormat.mHeight - 1;h>=0;h--) {
			uint8 *Line = &mData[h * LineLength];
			input.Read(Line,LineLength);
			if(BmpLineLengthDiff)
				input.Seek(BmpLineLengthDiff,GIStream::CURRENT);
		}
	}
	else {
		return;
	}
	mFormat.mNumChannels = 3;
	mFormat.mBitsPerChannel = 8;
	mFormat.mHasMipmaps = false;
	mFormat.mEncoding = "";   //TODO
	mReadOk = true;
}

GpBmpImage::~GpBmpImage()
{
}

// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8>
GpBmpImage::Data(int mipmapLevel)
{
	int ImageSize = mFormat.mWidth * mFormat.mHeight;
	GArray<uint8> ImageData(ImageSize * 3);
	uint8 *FormattedData = &ImageData[0];
	int i;

	switch(mDataFormat) {
	case DATA_FORMAT_1BIT:
		{
			int LinePtr = 0;
			int DataSize = mData.Size();

			for(i=0;i<DataSize;i++) {
				uint8 Data = mData[i];
				uint8 Mask = 0x80;
				for(int b=0;b<8;b++) {
					int ind = (Data & Mask) ? 3 : 0;
					*(FormattedData++) = mPalette[ind++];
					*(FormattedData++) = mPalette[ind++];
					*(FormattedData++) = mPalette[ind++];
					Mask >>= 1;
					if(++LinePtr >= mFormat.mWidth) {
						LinePtr = 0;
						break;
					}
				}
			}
		}
		break;
	case DATA_FORMAT_4BIT:
		{
			int DataSize = mData.Size();
			int LinePtr = 0;

			for(i=0;i<DataSize;i++) {
				int ind = ((mData[i] & 0xF0) >> 4) * 3;
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				if(++LinePtr >= mFormat.mWidth) {
					LinePtr = 0;
					continue;
				}	
				ind = (mData[i] & 0x0F) * 3;
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				if(++LinePtr >= mFormat.mWidth)
					LinePtr = 0;
			}
		}
		break;
	case DATA_FORMAT_8BIT:
		for(i=0;i<ImageSize;i++) {
			int ind = mData[i] * 3;
			*(FormattedData++) = mPalette[ind++];
			*(FormattedData++) = mPalette[ind++];
			*(FormattedData++) = mPalette[ind++];
		}
		break;
	case DATA_FORMAT_24BIT:
		for(i=0;i<ImageSize;i++) {
			*(FormattedData++) = mData[i * 3 + 2];
			*(FormattedData++) = mData[i * 3 + 1];
			*(FormattedData++) = mData[i * 3];
		}
		break;
	}
	return ImageData;
}

GAUGE3D_DECLARE_INTERFACES(GpBmpImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
