/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: pcximage.cpp
 *      Package: PCX image file plugin
 *    Component: ImgPak
 *  Description: Loads PCX Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "pcximage.h"

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

typedef struct {
	int8  mManufacturer;
	int8  mVersion;
	int8  mEncoding;
	int8  mBPP;
	int16 mXMin;
	int16 mYMin;
	int16 mXMax;
	int16 mYMax;
	int16 mHRes;
	int16 mVRes;
	int8  mColorMap[48];
	int8  mReserved;
	int8  mPlanes;
	int16 mBytesLine;
	int16 mPaletteInfo;
	int8  mFiller[58];
} PcxHeader;

GIStream& operator >> (GIStream& is, PcxHeader& hdr)
{
	is.Read(&hdr.mManufacturer, 4);
	is.Read(&hdr.mXMin, 6);
	is.Read(&hdr.mColorMap, 50);
	is.Read(&hdr.mBytesLine, 2);
	is.Read(hdr.mFiller, 58);
	return is;
}

static uint8 pcxMaskTable[] = {
	0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static uint8 pcxBitTable[] = {
	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
};

enum {
	DATA_FORMAT_UNKNOWN,
	DATA_FORMAT_1BIT,
	DATA_FORMAT_4BIT,
	DATA_FORMAT_8BIT,
	DATA_FORMAT_24BIT
};
// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpPcxImage::GpPcxImage(pGFile pcxfile)
	: mReadOk(false), mDataFormat(DATA_FORMAT_UNKNOWN), mExtraBytes(0)
{
	PcxHeader pcxHeader;
	int bitCount;
	int bytesPerLine;
	uint8 byte;

	pGRawIStream inSrc = pcxfile->GetRawInputStream();
	if(inSrc == NULL)
		return;
	GIStream input(inSrc);
	input.Endianness(GIStream::LITTLE);
	input >> pcxHeader;
	// Calculate dimensions
	mWidth = pcxHeader.mXMax - pcxHeader.mXMin + 1;
	mHeight = pcxHeader.mYMax - pcxHeader.mYMin + 1;
	bytesPerLine = pcxHeader.mBytesLine * pcxHeader.mPlanes;
	bitCount = pcxHeader.mBPP * pcxHeader.mPlanes;
	// Calculate extra bytes that each scan line has over the minimum storage requirements.
	mExtraBytes = bytesPerLine - ((bitCount * mWidth + 7) / 8);
	if(bitCount == 1) {	// mono palette
		mPalette = GArray<uint8>(6);
		mPalette[0] = mPalette[1] = mPalette[2] = 0;
		mPalette[3] = mPalette[4] = mPalette[5] = 255;
		mDataFormat = DATA_FORMAT_1BIT;
	}
	else if(bitCount == 4) {	// standard 16 color palette stored in header
		mPalette = GArray<uint8>(48);
		memcpy(&mPalette[0],pcxHeader.mColorMap,48);
		mDataFormat = DATA_FORMAT_4BIT;
	}
	else if(bitCount == 24) {
		mDataFormat = DATA_FORMAT_24BIT;
	}
	mData = GArray<uint8>(bytesPerLine * mHeight);
	GArray<uint8> Line(bytesPerLine);
	for(int h=0;h<mHeight;h++) {
		int index = 0;
		do {
			input.Read(&byte,1);
			if((byte & 0xC0) == 0xC0) {		// run-length encoded
				int count = byte & 0x3F;
				input.Read(&byte,1);
				memset(&Line[index],byte,count);
				index += count;
			}
			else {							// raw data
				Line[index++] = byte;
			}
		} while(index < bytesPerLine);
		if(pcxHeader.mPlanes == 1) {
			memcpy(&mData[bytesPerLine * h],&Line[0],bytesPerLine);
		}
		else {
			uint8 *DestPtr = &mData[bytesPerLine * h];
			if(bitCount == 24) {			// 24-bit data read as all REDS, then all BLUES, then all GREENS
				for(int i=0;i<pcxHeader.mBytesLine;i++) {
					for(int j=0;j<3;j++)
						*(DestPtr++) = Line[i + pcxHeader.mBytesLine * j];
				}
			}
			else {			// get data from planes
				memset(DestPtr,0,bytesPerLine);
				for(int i=0;i<pcxHeader.mBytesLine;i++) {
					for(int j=0;j<8;j+=2) {
						for(int k=0;k<pcxHeader.mPlanes;k++) {
							if(Line[i + k * pcxHeader.mBytesLine] & pcxMaskTable[j])
								*DestPtr |= pcxBitTable[k + 4];
							if(Line[i + k * pcxHeader.mBytesLine] & pcxMaskTable[j + 1])
								*DestPtr |= pcxBitTable[k];
						}
						DestPtr++;
					}
				}
			}
		}
	}
	// Read palette for 8-bit (indexed) image
	if((pcxHeader.mVersion >= 5) && (bitCount == 8)) {
		input.Read(&byte,1);
		if(byte != 12)
			return;
		mPalette = GArray<uint8>(768);
		input.Read(&mPalette[0],768);
		mDataFormat = DATA_FORMAT_8BIT;
	}
	// Standard format for image
	mFormat.mWidth = mWidth;
	mFormat.mHeight = mHeight;
	mFormat.mNumChannels = 3;
	mFormat.mBitsPerChannel = 8;
	mFormat.mHasMipmaps = false;
	mFormat.mEncoding = "";   //TODO
	mReadOk = true;
}

GpPcxImage::~GpPcxImage()
{
}

// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8>
GpPcxImage::Data(int mipmapLevel)
{
	GArray<uint8> ImageData(mWidth * mHeight * 3);
	switch(mDataFormat) {
	case DATA_FORMAT_1BIT:
		{
			uint8 *FormattedData = &ImageData[0];
			int DataSize = mData.Size();
			int LinePtr = 0;
			for(int c=0;c<DataSize;c++) {
				uint8 data = mData[c];
				uint8 mask = 0x80;
				for(int b=0;b<8;b++) {
					int ind = (data & mask) ? 3 : 0;
					*(FormattedData++) = mPalette[ind++];
					*(FormattedData++) = mPalette[ind++];
					*(FormattedData++) = mPalette[ind++];
					mask >>= 1;
					// Keep track of where we are in the line so that if it finishes within a
					//  byte, then we can move off the byte correctly. Also advance by the number
					//  of extra bytes used in the PCX image.
					if(++LinePtr >= mFormat.mWidth) {
						LinePtr = 0;
						c += mExtraBytes;
						break;
					}
				}
			}
		}
		break;
	case DATA_FORMAT_4BIT:
		{
			uint8 *FormattedData = &ImageData[0];
			int DataSize = mData.Size();
			int LinePtr = 0;
			for(int c=0;c<DataSize;c++) {
				int ind = ((mData[c] & 0xF0) >> 4) * 3;
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				// Keep track of where we are in the line so that if it finishes within a
				//  byte, then we can move off the byte correctly. Also advance by the number
				//  of extra bytes used in the PCX image.
				if(++LinePtr >= mFormat.mWidth) {
					LinePtr = 0;
					c += mExtraBytes;
					continue;
				}
				ind = (mData[c] & 0x0F) * 3;
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				if(++LinePtr >= mFormat.mWidth) {
					LinePtr = 0;
					c += mExtraBytes;
				}
			}
		}
		break;
	case DATA_FORMAT_8BIT:
		{
			uint8 *FormattedData = &ImageData[0];
			int DataSize = mData.Size();
			int LinePtr = 0;
			for(int c=0;c<DataSize;c++) {
				int ind = mData[c] * 3;
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				*(FormattedData++) = mPalette[ind++];
				if(++LinePtr >= mFormat.mWidth) {
					LinePtr = 0;
					c += mExtraBytes;
				}
			}
		}
		break;
	case DATA_FORMAT_24BIT:
		{
			uint8 *FormattedData = &ImageData[0];
			uint8 *SourceData = &mData[0],*EndSourceData = &mData[mData.Size()];
			int LinePtr = 0;
			do {
				*(FormattedData++) = *(SourceData++);
				*(FormattedData++) = *(SourceData++);
				*(FormattedData++) = *(SourceData++);
				if(++LinePtr >= mFormat.mWidth) {
					LinePtr = 0;
					SourceData += mExtraBytes;
				}
			} while(SourceData < EndSourceData);
		}
		break;
	}
	return ImageData;
}

GAUGE3D_DECLARE_INTERFACES(GpPcxImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
