/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2001 Will Newton
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: pngimage.cpp
 *      Package: PNG image file plugin
 *    Component: PngImg
 *  Description: Loads PNG Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "pngimage.h"
#include <stdio.h>

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {


// class member functions ============================================
// lifecycle ---------------------------------------------------------


void user_error_fn(png_structp png_ptr, png_const_charp error_msg)
{
	// Ignore
}

void user_warning_fn(png_structp png_ptr, png_const_charp error_msg)
{
	        // Ignore
}


void user_read_fn(png_structp png_ptr,png_bytep data, png_size_t length)
{
	GIStream *istr = (GIStream *)png_get_io_ptr(png_ptr);

	istr->Read(data,length);
}	
	

GpPngImage::GpPngImage(pGFile pngfile)
	: mReadOk(false)
{
	png_structp png_ptr;
	png_infop info_ptr;
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type,row_stride;   
	
	pGRawIStream inSrc = pngfile->GetRawInputStream();

	if(inSrc == NULL)
		return;

	GIStream input(inSrc);
        input.Endianness(GIStream::LITTLE);

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
		NULL, user_error_fn, user_warning_fn);

	info_ptr = png_create_info_struct(png_ptr);
       	
	png_set_read_fn(png_ptr, (void *)&input, user_read_fn);
	
	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth,&color_type,
		&interlace_type, NULL, NULL);
	
       	
	png_set_expand(png_ptr);

	// TODO: Gamma correction?
	//png_set_gamma(png_ptr, screen_gamma, 0.45455);
	
	png_read_update_info(png_ptr, info_ptr);

	//png_bytep row_pointers[height];

	row_stride = png_get_rowbytes(png_ptr, info_ptr);
	mData = GArray<uint8>(row_stride * height);

	if (mData)
	{
		for (png_uint_32 y = 0; y < height; y++)
		{
			uint8 *line = &mData[y*row_stride];
			png_read_rows(png_ptr, &line, NULL, 1);
		}
	}
	
	//png_read_image(png_ptr, &mData);
	
	png_read_end(png_ptr, info_ptr);
	
	mReadOk = true;

	mFormat.mWidth = width;
	mFormat.mHeight = height;
	mFormat.mNumChannels = png_get_channels(png_ptr, info_ptr);
	mFormat.mBitsPerChannel = bit_depth;
	mFormat.mHasMipmaps = false;
	//mFormat.mEncoding = "raw24";
	
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL); 
}

GpPngImage::~GpPngImage()
{
}


// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8> GpPngImage::Data(int mipmapLevel)
{
	return GArray<uint8>(mData);
}

GAUGE3D_DECLARE_INTERFACES(GpPngImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
