/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: 3dstudio.cpp
 *      Package:
 *    Component:
 *  Description:
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "3dstudio.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
Gp3DSModel::Gp3DSModel(pGFile file)
	: mOpen(false)
{
	pGRawIStream rawInput = file->GetRawInputStream();
	
	if(rawInput == NULL)
		return;
	
	GIStream input(rawInput);
	input.Endianness(GIStream::LITTLE);
	
	uint16 chunkType;
	int32 chunkSize;
	int maxSize;
	int pos;
	char junk;
	
	input >> chunkType;
	if(chunkType != 0x4D4D)
		return;     //not a 3ds file
	
	input >> chunkSize;
	maxSize = chunkSize;
	
	while(input.TellPos() < maxSize)
	{
		input >> chunkType >> chunkSize;
		
		switch(chunkType)
		{
		case EDIT3DS:
			//enter this chunk
			break;
		case EDIT_OBJECT:
			do
				input >> junk;
			while(junk != 0);
			//enter this chunk
			break;
		case OBJ_TRIMESH:
			//read the trimesh
			pos = input.TellPos();
			ReadTriMesh(input, chunkSize);
			input.Seek(pos + chunkSize - 6, GIStream::BEGINNING);
			break;
		default:
			//ignore chunk
			input.Seek(chunkSize - 6, GIStream::CURRENT);
			break;
		}
	}
	
	mModel.mMeshes      = GStridedArray<Mesh>(&mMesh);
	mModel.mNumMeshes   = 1;
	mMesh.mFrames       = GStridedArray<MeshFrame>(&mFrame);
	mMesh.mNumFrames    = 1;
	mMesh.mDynamicFlags = RenderInfo::RenderSet::STATIC;
	mMesh.mNumVertices  = mVertices.size();
	mMesh.mNumIndices   = mIndices.size();
	mMesh.mIndices      = &mIndices[0];
	mMesh.mIndexType    = RenderInfo::RenderSet::TRIANGLES;
	mFrame.mVertices    = &mVertices[0];
	mFrame.mNormals     = &mNormals[0];
	
	if(mTexCoords.size() == mVertices.size())
		mFrame.mTexCoords = &mTexCoords[0];
	else
		mFrame.mTexCoords = NULL;
	
	ModelData(&mModel);
	
	mOpen = true;
}

Gp3DSModel::~Gp3DSModel()
{
	//Everything will take care of itself.  Yay C++!
}

struct Face
{
	uint32 mGroups;
	uint32 mVertices[3];
	GVector mNormal;
};

struct FaceRef
{
	int mFace;
	GCoordinate mWeight;
	
	FaceRef(int face, GCoordinate weight)
		: mFace(face), mWeight(weight) {}
};

struct VertexStuff
{
	GVector mPosition;
	GVector mNormal;
	GVector mTexCoords;
	vector<FaceRef> mFaces;
};

void
Gp3DSModel::ReadTriMesh(GIStream& input, int size)
{
	//Help!  I'm being overrun by variables!
	uint16 chunkType;
	int32 chunkSize;
	int endPos = input.TellPos() + size - 6;
	uint i, j;
	uint16 numVertices = 0, numFaces = 0;
	uint16 inputIndices[4];
	float32 coords[3];
	GVector matrix[4];
	bool useMatrix = false;
	uint32* groups;
	GVector vec;
	GVector norm;
	GVector a, b, c, ab, bc, ac;
	int startIndex = mVertices.size();
	GCoordinate weight;
	int pos;
	Face* faces;
	vector<VertexStuff> vertices;
	VertexStuff curVertex;
	bool foundTexCoords = false;
	
	while(input.TellPos() < endPos)
	{
		input >> chunkType >> chunkSize;
		pos = input.TellPos();
		
		switch(chunkType)
		{
		case TRI_VERTEXL:
			input >> numVertices;
			for(i = 0; i < numVertices; i++)
			{
				input.Read(coords, 3);
				vec.x = coords[0] * 0.01;
				vec.y = coords[1] * 0.01;
				vec.z = coords[2] * 0.01;
				if(useMatrix)
					vec = vec.x * matrix[0] + vec.y * matrix[1] + vec.z * matrix[2] + matrix[3];
				curVertex.mPosition = vec;
				vertices.push_back(curVertex);
			}
			break;
		case TRI_MAPPINGCOORDS:
			foundTexCoords = true;
			input >> numVertices;
			for(i = 0; i < numVertices; i++)
			{
				input.Read(coords, 2);
				vec.x = coords[0];
				vec.y = -coords[1];
				vec.z = 0;
				
				vertices[i].mTexCoords = vec;
			}
			break;
		case TRI_FACEL1:
			input >> numFaces;
			faces = new Face[numFaces];
			for(i = 0; i < numFaces; i++)
			{
				input.Read(inputIndices, 4);
				
				//reverse order to make clockwise
				faces[i].mVertices[0] = inputIndices[2];
				faces[i].mVertices[1] = inputIndices[1];
				faces[i].mVertices[2] = inputIndices[0];
				
				//calculate face normal
				a = vertices[inputIndices[0]].mPosition;
				b = vertices[inputIndices[1]].mPosition;
				c = vertices[inputIndices[2]].mPosition;
				
				ab = b - a;
				ac = c - a;
				bc = c - b;
				
				GCoordinate area;
				area = (ab.X(ac)).Magnitude();  // actually, this is double the area
				
				ab.Normalize();
				ac.Normalize();
				bc.Normalize();
				
				norm = ab.X(ac);
				norm.Normalize();
				
				faces[i].mNormal = norm;
				
				//map vertices to faces, and store weights which are the angle
				// that the face makes at that vertex multiplied by the area
				// of the face.  (the weights are used to calculate normals later on)
				weight = acos(ab * ac) * area;
				vertices[inputIndices[0]].mFaces.push_back(FaceRef(i, weight));
				
				weight = acos(bc * (-1 * ab)) * area;
				vertices[inputIndices[1]].mFaces.push_back(FaceRef(i, weight));
				
				weight = acos((-1 * ac) * (-1 * bc)) * area;
				vertices[inputIndices[2]].mFaces.push_back(FaceRef(i, weight));
			}
			
			//note: TRI_FACEL1 has sub-chunks (noteably TRI_SMOOTH).
			break;
		case TRI_SMOOTH:
			//smoothing groups indicate if vertex normals should be smoothed between
			// two faces.
			groups = new uint32[numFaces];
			input.Read(groups, numFaces);
			
			for(i = 0; i < numFaces; i++)
				faces[i].mGroups = groups[i];
			delete [] groups;
			
			break;
		case TRI_LOCAL:
#if 0
			//FIXME: this code doesn't seem to work...
			//a matrix by which to multiply all vertices
			useMatrix = true;
			input.Read(&matrix[0].x, 12);
			matrix[3] *= 0.01;
			
			cout << "Local axis: " << endl;
			cout << matrix[0] << endl;
			cout << matrix[1] << endl;
			cout << matrix[2] << endl;
			cout << matrix[3] << endl;
			
			//now we multiply all vertices by the local axis matrix
			for(i = 0; i < vertices.size(); i++)
			{
				vec = vertices[i].mPosition;
				
				vertices[i].mPosition = vec.x * matrix[0] + vec.y * matrix[1] +
				                        vec.z * matrix[2] + matrix[3];
			}
			break;
#endif
		default:
			//ignore chunk
			input.Seek(pos + chunkSize - 6, GIStream::BEGINNING);
			break;
		}
	}
	
	//OK, time to deal with 3D Studio's wacked system of finding vertex
	// normals.  Here's the deal:  We now have a set of vertices and a set
	// of faces.  We also have a map which allows us to determine which faces
	// contain a given vertex.  The problem is that we can only render models
	// which have one and only one normal vector per vertex, but since some
	// of the corners in our model are sharp, those corners have multiple normal
	// vectors.  We have to split those vertices such that we hawe one vertex
	// per normal vector.
	
	for(i = 0; i < vertices.size(); i++)
	{
		//if we need to split the vertex, curVertex will be the new vertex.
		curVertex.mFaces.clear();
		for(j = 1; j < vertices[i].mFaces.size(); j++)
		{
			//check if face 0 and face j share any common smoothing groups.
			if(!(faces[vertices[i].mFaces[0].mFace].mGroups
			   & faces[vertices[i].mFaces[j].mFace].mGroups))
			{
				//no common groups.  Add the face to the split vertex.
				curVertex.mFaces.push_back(vertices[i].mFaces[j]);
				//remove the face from this vertex.
				vertices[i].mFaces.erase(vertices[i].mFaces.begin() + j);
				--j;
				
				//update the face's index to match the new vertex.
				Face* curFace = faces + curVertex.mFaces.back().mFace;
				if(curFace->mVertices[0] == i)
					curFace->mVertices[0] = vertices.size();
				if(curFace->mVertices[1] == i)
					curFace->mVertices[1] = vertices.size();
				if(curFace->mVertices[2] == i)
					curFace->mVertices[2] = vertices.size();
			}
		}
		
		if(!(curVertex.mFaces.empty()))
		{
			//a new vertex needs to be created.
			curVertex.mPosition  = vertices[i].mPosition;
			curVertex.mTexCoords = vertices[i].mTexCoords;
			
			vertices.push_back(curVertex);
		}
	}
	
	if(numFaces <= 0 || vertices.size() <= 0)
		return;  //nothing to see here.  Move along.
	
	//OK, now the vertices should be split nicely, and we can go and find
	// a texture coordinate for each one.
	
	for(i = 0; i < vertices.size(); i++)
	{
		norm = GVECTOR_0;
		
		for(j = 0; j < vertices[i].mFaces.size(); j++)
		{
			norm += faces[vertices[i].mFaces[j].mFace].mNormal
			        * vertices[i].mFaces[j].mWeight;
		}
		
		norm.Normalize();
		vertices[i].mNormal = norm;
	}
	
	//We have all the data.  Now we can put it into the main list.
	
	for(i = 0; i < vertices.size(); i++)
	{
		mVertices.push_back(vertices[i].mPosition);
		mNormals.push_back(vertices[i].mNormal);
		if(foundTexCoords)
			mTexCoords.push_back(vertices[i].mTexCoords);
	}
	
	for(i = 0; i < numFaces; i++)
	{
		mIndices.push_back(faces[i].mVertices[0] + startIndex);
		mIndices.push_back(faces[i].mVertices[1] + startIndex);
		mIndices.push_back(faces[i].mVertices[2] + startIndex);
	}
	
	delete [] faces;
}

GAUGE3D_DECLARE_INTERFACES(Gp3DSModel)
	GAUGE3D_DECLARE_BASECLASS(GSuperModel)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
