/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7camera.cpp
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: Display
 *  Description: Represents the rendering window.
 * Dependencies: GpD3D7Scene, GpD3D7Sprite, GpD3D7Light
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "d3d7camera.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_D3DX_H__INCLUDED__
#include <d3dx.h>
#define __SYSTEM_D3DX_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
// lifecycle -------------------------------------------------------
GpD3D7Camera::GpD3D7Camera(pGpD3D7Scene scene,LPDIRECT3DDEVICE7 device)
	: mScene(scene),
	  mPosition(GVECTOR_0),
	  mDirection(GVECTOR_I, 0.0f),
	  mZoom(1.0f),
	  mDevice(device)
{
}

GpD3D7Camera::~GpD3D7Camera()
{
}

// operations ------------------------------------------------------
//Position sets the position of the camera.
void
GpD3D7Camera::Position(const GVector& position)
{
	mPosition = position;
}

//Direction sets the direction in which the camera is facing.
void
GpD3D7Camera::Direction(const GQuaternion& direction)
{
	mDirection = direction;
}

//Zoom!  1.0 is normal, higher is zoomed in.  Setting the zoom below
// 1.0 is probably a bad idea, as this will increase the FOV beyond
// 90 degrees, which could cause distortion.
void
GpD3D7Camera::Zoom(GCoordinate zoom)
{
	mZoom = zoom;
}

//Render the scene from the point of view of the camera.
void
GpD3D7Camera::Draw(float geoLod)
{
	D3DMATRIX mat;

	//Apply camera direction to view matrix
	D3DXMatrixIdentity(reinterpret_cast<D3DXMATRIX *>(&mat));
	mDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,&mat);
	GQuaternion dir = GQuaternion(GVECTOR_I, PI_2);  //forward
	dir *= mDirection;
	GVector axis = dir.Axis();
	GCoordinate angle = dir.Angle();
	if(!IsApproxZero(angle)) {
		D3DXVECTOR3 d3daxis;

		d3daxis.x = axis.x;
		d3daxis.y = axis.y;
		d3daxis.z = axis.z;
		D3DXMatrixRotationAxis(reinterpret_cast<D3DXMATRIX *>(&mat),&d3daxis,angle);
		mDevice->MultiplyTransform(D3DTRANSFORMSTATE_VIEW,&mat);
	}
	//Apply camera position to modelview matrix
	D3DXMatrixTranslation(reinterpret_cast<D3DXMATRIX *>(&mat),-mPosition.x,-mPosition.y,-mPosition.z);
	mDevice->MultiplyTransform(D3DTRANSFORMSTATE_VIEW,&mat);



	GFrustum viewFrustum;
	viewFrustum.mNear = 0.1;
	viewFrustum.mFar  = 100.0;
	viewFrustum.mHorizontalSlope = 1.0f  / mZoom; //90 degree FOV
	viewFrustum.mVerticalSlope   = 0.75f / mZoom; //4/3 aspect ratio
	viewFrustum.mWidth  = viewFrustum.mHorizontalSlope * viewFrustum.mNear * 2.0f;
	viewFrustum.mHeight = viewFrustum.mVerticalSlope   * viewFrustum.mNear * 2.0f;
	
	ZeroMemory(&mat, sizeof(D3DMATRIX));
	mat(0,0) = 1.0f / viewFrustum.mHorizontalSlope;
	mat(1,1) = 1.0f / viewFrustum.mVerticalSlope;
	float Q = viewFrustum.mFar / (viewFrustum.mFar - viewFrustum.mNear);
	mat(2,2) = -Q;
	mat(2,3) = -1;
	mat(3,2) = 2.0f * -Q * viewFrustum.mNear;
	mat(3,3) = 0;
	mDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION,&mat);

	//Now that the matrices are set up, the scene can be drawn in world coordinates.
	mScene->Draw(mPosition, mDirection, viewFrustum, geoLod);
}

GAUGE3D_DECLARE_INTERFACES(GpD3D7Camera)
	GAUGE3D_DECLARE_BASECLASS(GCamera)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} //namespace GAUGE3D
