/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Direct3D 7 Rendering Plugin
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7display.cpp
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: Display
 *  Description: Represents the rendering window.
 * Dependencies: GpD3D7Scene, GpD3D7Sprite, GpD3D7Light
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "d3d7display.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_IOSTREAM__INCLUDED__
#include <iostream>
#define __SYSTEM_IOSTREAM__INCLUDED__
#endif

#ifndef __SYSTEM_VECTOR__INCLUDED__
#include <vector>
#define __SYSTEM_VECTOR__INCLUDED__
#endif

#ifndef __SYSTEM_D3DX_H__INCLUDED__
#include <d3dx.h>
#define __SYSTEM_D3DX_H__INCLUDED__
#endif //redundant include guard

#include <iostream>
using namespace std;

// GAUGE^3D ----------------------------------------------------------

#undef ABSOLUTE  //temporal no like win32 API
#undef RELATIVE

#include "gauge3d/input/messagequeue.h"
#include "gauge3d/scheduler.h"

// local package -----------------------------------------------------
#include "d3d7scene.h"
#include "d3d7camera.h"

// Windows functions==================================================
// The display class uses the current DLL instance when creating its window,
// so we obtain it in the DLL's entry point function.
BOOL WINAPI DllMain(HINSTANCE hinstDLL,DWORD fdwReason,LPVOID lpvReserved)
{
	if(fdwReason == DLL_PROCESS_ATTACH) {
		GAUGE3D::Plugins::GpD3D7Display::mInstance = hinstDLL;
	}
	return TRUE;
}

#define RELEASE_INTERFACE(int) if(int) { int->Release(); int = NULL; }

// Debug macro to send a specified error to all available locations.
#define SEND_ERROR(buf) cerr << buf; OutputDebugString(buf); MessageBox(NULL,buf,NULL,MB_OK | MB_ICONERROR);

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

const UINT WM_SYNCACQUIRE = WM_USER;

HINSTANCE GpD3D7Display::mInstance;

// A window procedure for the window that the display creates.  At the moment it is
// used to tell the display class about window position and size changes.
LRESULT CALLBACK GpD3D7WindowProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	// A pointer to the window's owner class is stored in the user data section of the window (which
	// is just a 32bit value).
	GpD3D7Display *display = reinterpret_cast<GpD3D7Display *>(GetWindowLong(hwnd,GWL_USERDATA));

	if(display)
		return display->WindowProc(hwnd,msg,wparam,lparam);
	return DefWindowProc(hwnd,msg,wparam,lparam);
}

HRESULT WINAPI GpD3D7EnumModesCallback(LPDDSURFACEDESC2 lpDDSurfaceDesc,LPVOID lpContext)
{
	std::vector<GDisplay::DisplayProps> *Displays = static_cast<std::vector<GDisplay::DisplayProps> *>(lpContext);
	GDisplay::DisplayProps ThisDisplay;
	
	ThisDisplay.mWidth = lpDDSurfaceDesc->dwWidth;
	ThisDisplay.mHeight = lpDDSurfaceDesc->dwHeight;
	ThisDisplay.mColorDepth = lpDDSurfaceDesc->ddpfPixelFormat.dwRGBBitCount;
	ThisDisplay.mFullScreen = true;
	Displays->push_back(ThisDisplay);
	return DDENUMRET_OK;
}

// Maps Direct Input scan codes to SDL matching names
static char *DIK_KeyNames[256];

// Initialises DIK_KeyNames so that the Direct Input scan codes DIK_ match up with the same
//  names as SDL provides for the same keys.
void GpD3D7InitKeyNames(void)
{
	DIK_KeyNames[DIK_ESCAPE] = "escape";
	DIK_KeyNames[DIK_1] = "1";
	DIK_KeyNames[DIK_2] = "2";
	DIK_KeyNames[DIK_3] = "3";
	DIK_KeyNames[DIK_4] = "4";
	DIK_KeyNames[DIK_5] = "5";
	DIK_KeyNames[DIK_6] = "6";
	DIK_KeyNames[DIK_7] = "7";
	DIK_KeyNames[DIK_8] = "8";
	DIK_KeyNames[DIK_9] = "9";
	DIK_KeyNames[DIK_0] = "0";
	DIK_KeyNames[DIK_MINUS] = "-";
	DIK_KeyNames[DIK_EQUALS] = "=";
	DIK_KeyNames[DIK_BACK] = "backspace";
	DIK_KeyNames[DIK_TAB] = "tab";
	DIK_KeyNames[DIK_Q] = "q";
	DIK_KeyNames[DIK_W] = "w";
	DIK_KeyNames[DIK_E] = "e";
	DIK_KeyNames[DIK_R] = "r";
	DIK_KeyNames[DIK_T] = "t";
	DIK_KeyNames[DIK_Y] = "y";
	DIK_KeyNames[DIK_U] = "u";
	DIK_KeyNames[DIK_I] = "i";
	DIK_KeyNames[DIK_O] = "o";
	DIK_KeyNames[DIK_P] = "p";
	DIK_KeyNames[DIK_LBRACKET] = "[";
	DIK_KeyNames[DIK_RBRACKET] = "]";
	DIK_KeyNames[DIK_RETURN] = "return";
	DIK_KeyNames[DIK_LCONTROL] = "left ctrl";
	DIK_KeyNames[DIK_A] = "a";
	DIK_KeyNames[DIK_S] = "s";
	DIK_KeyNames[DIK_D] = "d";
	DIK_KeyNames[DIK_F] = "f";
	DIK_KeyNames[DIK_G] = "g";
	DIK_KeyNames[DIK_H] = "h";
	DIK_KeyNames[DIK_J] = "j";
	DIK_KeyNames[DIK_K] = "k";
	DIK_KeyNames[DIK_L] = "l";
	DIK_KeyNames[DIK_SEMICOLON] = ";";
	DIK_KeyNames[DIK_APOSTROPHE] = "'";
	DIK_KeyNames[DIK_GRAVE] = "`";
	DIK_KeyNames[DIK_LSHIFT] = "left shift";
	DIK_KeyNames[DIK_BACKSLASH] = "\\";
	DIK_KeyNames[DIK_Z] = "z";
	DIK_KeyNames[DIK_X] = "x";
	DIK_KeyNames[DIK_C] = "c";
	DIK_KeyNames[DIK_V] = "v";
	DIK_KeyNames[DIK_B] = "b";
	DIK_KeyNames[DIK_N] = "n";
	DIK_KeyNames[DIK_M] = "m";
	DIK_KeyNames[DIK_COMMA] = ",";
	DIK_KeyNames[DIK_PERIOD] = ".";
	DIK_KeyNames[DIK_SLASH] = "/";
	DIK_KeyNames[DIK_RSHIFT] = "right shift";
	DIK_KeyNames[DIK_MULTIPLY] = "[*]";
	DIK_KeyNames[DIK_LMENU] = "left alt";
	DIK_KeyNames[DIK_SPACE] = "space";
	DIK_KeyNames[DIK_CAPITAL] = "caps lock";
	DIK_KeyNames[DIK_F1] = "f1";
	DIK_KeyNames[DIK_F2] = "f2";
	DIK_KeyNames[DIK_F3] = "f3";
	DIK_KeyNames[DIK_F4] = "f4";
	DIK_KeyNames[DIK_F5] = "f5";
	DIK_KeyNames[DIK_F6] = "f6";
	DIK_KeyNames[DIK_F7] = "f7";
	DIK_KeyNames[DIK_F8] = "f8";
	DIK_KeyNames[DIK_F9] = "f9";
	DIK_KeyNames[DIK_F10] = "f10";
	DIK_KeyNames[DIK_NUMLOCK] = "numlock";
	DIK_KeyNames[DIK_SCROLL] = "scroll lock";
	DIK_KeyNames[DIK_NUMPAD7] = "[7]";
	DIK_KeyNames[DIK_NUMPAD8] = "[8]";
	DIK_KeyNames[DIK_NUMPAD9] = "[9]";
	DIK_KeyNames[DIK_SUBTRACT] = "[-]";
	DIK_KeyNames[DIK_NUMPAD4] = "[4]";
	DIK_KeyNames[DIK_NUMPAD5] = "[5]";
	DIK_KeyNames[DIK_NUMPAD6] = "[6]";
	DIK_KeyNames[DIK_ADD] = "[+]";
	DIK_KeyNames[DIK_NUMPAD1] = "[1]";
	DIK_KeyNames[DIK_NUMPAD2] = "[2]";
	DIK_KeyNames[DIK_NUMPAD3] = "[3]";
	DIK_KeyNames[DIK_NUMPAD0] = "[0]";
	DIK_KeyNames[DIK_DECIMAL] = "[.]";
	DIK_KeyNames[DIK_F11] = "f11";
	DIK_KeyNames[DIK_F12] = "f12";
	DIK_KeyNames[DIK_F13] = "f13";
	DIK_KeyNames[DIK_F14] = "f14";
	DIK_KeyNames[DIK_F15] = "f15";
	DIK_KeyNames[DIK_NUMPADEQUALS] = "equals";
	DIK_KeyNames[DIK_NUMPADENTER] = "enter";
	DIK_KeyNames[DIK_RCONTROL] = "right ctrl";
	DIK_KeyNames[DIK_DIVIDE] = "[/]";
	DIK_KeyNames[DIK_SYSRQ] = "sys req";
	DIK_KeyNames[DIK_RMENU] = "right alt";
	DIK_KeyNames[DIK_HOME] = "home";
	DIK_KeyNames[DIK_UP] = "up";
	DIK_KeyNames[DIK_PRIOR] = "page up";
	DIK_KeyNames[DIK_LEFT] = "left";
	DIK_KeyNames[DIK_RIGHT] = "right";
	DIK_KeyNames[DIK_END] = "end";
	DIK_KeyNames[DIK_DOWN] = "down";
	DIK_KeyNames[DIK_NEXT] = "page down";
	DIK_KeyNames[DIK_INSERT] = "insert";
	DIK_KeyNames[DIK_DELETE] = "delete";
	DIK_KeyNames[DIK_LWIN] = "left meta";
	DIK_KeyNames[DIK_RWIN] = "right meta";
	DIK_KeyNames[DIK_APPS] = "menu";
}

// class state definition ============================================
// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpD3D7Display::GpD3D7Display(IDirectDraw7 *DD) : mDD(DD)
{
	mError = NULL;
	mInit = false;
	mPrimarySurface = NULL;
	mBackSurface = NULL;
	mDevice = NULL;
	mZBuffer = NULL;
	mDInput = NULL;
	mDIMouse = NULL;
	mDIKeyboard = NULL;
	mWindow = NULL;
	mFullScreen = false;
	mQuitting = false;
	mGeoLod = 0.25f / 16384;
	GpD3D7InitKeyNames();

	gVideoScheduler->CallEveryFrame(this, 0);
}

GpD3D7Display::~GpD3D7Display()
{
	gVideoScheduler->StopCallingEveryFrame(this);

	if(IsWindow(mWindow))
		SetWindowLong(mWindow,GWL_USERDATA,0);
	if(mInit) {
		RELEASE_INTERFACE(mDevice);
		if(mZBuffer)
			mBackSurface->DeleteAttachedSurface(0,mZBuffer);
		RELEASE_INTERFACE(mZBuffer);
		RELEASE_INTERFACE(mPrimarySurface);
		RELEASE_INTERFACE(mBackSurface);
		SetMouseAcquire(false);
		RELEASE_INTERFACE(mDIMouse);
		mDIKeyboard->Unacquire();
		RELEASE_INTERFACE(mDIKeyboard);
		RELEASE_INTERFACE(mDInput);
		if(IsWindow(mWindow))
			DestroyWindow(mWindow);
	}
	RELEASE_ERROR(mError);
	mInit = false;
}

// operations --------------------------------------------------------
//Open resets the display resolution and color depth.
//This function must be called once before any rendering can be done.
//If the display could not be opened, false is returned.
//Passing zero for colorDepth will cause the current display depth to be used.
bool
GpD3D7Display::Open(int width, int height, int colorDepth, bool fullScreen)
{
	const char ClassName[] = "GpD3D7Display::Window";
	const char WindowName[] = "Gauge3D Direct3D Renderer Window";
	HRESULT hr;
	WNDCLASSEX wc;

	mFullScreen = fullScreen;
	// mDD is initialised by the rendering interface so that only one IDirectDraw7
	// interface is initialised for all drawing contexts.
	if(mDD == NULL) {
		CREATE_GAUGE3D_ERROR(mError,GGaugeError::INVALID_ARG,"IDirectDraw7 Interface not defined");
		return false;
	}
	if(mDInput == NULL) {
		hr = DirectInputCreateEx(mInstance,DIRECTINPUT_VERSION,IID_IDirectInput7,reinterpret_cast<LPVOID *>(&mDInput),NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating DirectInput7 interface");
			return false;
		}
	}
	// Preliminary: if a window is already created for this class, then just resize it.
	// Should handle changes for fullscreen before it is considered complete.
	if(IsWindow(mWindow)) {
		SetWindowPos(mWindow,NULL,0,0,width,height,SWP_NOMOVE | SWP_NOZORDER);
		return true;
	}
	// Setup a window class.
	ZeroMemory(&wc,sizeof(WNDCLASSEX));
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = GpD3D7WindowProc;
	wc.hInstance = mInstance;
	wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = static_cast<HBRUSH>(GetStockObject(BLACK_BRUSH));
	wc.lpszClassName = ClassName;
	wc.hIconSm = wc.hIcon;
	RegisterClassEx(&wc);
	mWindow = CreateWindowEx(0,ClassName,WindowName,WS_OVERLAPPEDWINDOW | WS_VISIBLE,CW_USEDEFAULT,CW_USEDEFAULT,width,height,NULL,NULL,mInstance,NULL);
	if(!IsWindow(mWindow)) {
		CREATE_WIN32_ERROR(mError,"Creating rendering window");
		return false;
	}
	GetClientRect(mWindow,&mPosition);
	mSize.cx = mPosition.right;
	mSize.cy = mPosition.bottom;
	ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.left));
	ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.right));
	// Provide the window with its owner class.
	SetWindowLong(mWindow,GWL_USERDATA,reinterpret_cast<LONG>(this));
	if(fullScreen) {
		if(colorDepth == 0) {
			DDSURFACEDESC2 ddsd = { sizeof(DDSURFACEDESC2) };
			mDD->GetDisplayMode(&ddsd);
			colorDepth = ddsd.ddpfPixelFormat.dwRGBBitCount;
		}
		hr = mDD->SetCooperativeLevel(mWindow,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Setting exclusive cooperative mode");
			return false;
		}
		hr = mDD->SetDisplayMode(width,height,colorDepth,0,0);
		if(FAILED(hr)) {
			mDD->RestoreDisplayMode();
			mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
			CREATE_DIRECTX_ERROR(mError,hr,"Changing display mode");
			return false;
		}
		if(!CreateFullScreen())
			return false;
	}
	else {
		DDSURFACEDESC2 ddsd;
		LPDIRECTDRAWCLIPPER clipper = NULL;

		hr = mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Setting normal cooperative mode");
			return false;
		}
		ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
		// Create the primary surface that matches the window we just created.
		ddsd.dwSize = sizeof(DDSURFACEDESC2);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
		hr = mDD->CreateSurface(&ddsd,&mPrimarySurface,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating primary rendering surface");
			return false;
		}
		// Create the clipper for the primary surface.
		hr = mDD->CreateClipper(0,&clipper,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating clipper for primary rendering surface");
			return false;
		}
		clipper->SetHWnd(0,mWindow);
		mPrimarySurface->SetClipper(clipper);
		RELEASE_INTERFACE(clipper);
		if(!CreateBackSurface())
			return false;
	}
	if(!CreateZBuffer())
		return false;
	if(!CreateDevice())
		return false;
	SetViewport();
	hr = mDInput->CreateDeviceEx(GUID_SysMouse,IID_IDirectInputDevice7,reinterpret_cast<LPVOID *>(&mDIMouse),NULL);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Obtaining device for system mouse");
		return false;
	}
	hr = mDIMouse->SetDataFormat(&c_dfDIMouse);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting data format for system mouse DI device");
		return false;
	}
	hr = mDIMouse->SetCooperativeLevel(mWindow,DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting system mouse cooperative level");
		return false;
	}
	DIPROPDWORD dipd;
	dipd.diph.dwSize = sizeof(DIPROPDWORD);
	dipd.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	dipd.diph.dwObj = 0;
	dipd.diph.dwHow = DIPH_DEVICE;
	dipd.dwData = 32;
	hr = mDIMouse->SetProperty(DIPROP_BUFFERSIZE,&dipd.diph);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting buffer size property for system mouse");
		return false;
	}
	PostMessage(mWindow,WM_SYNCACQUIRE,1,0);
	hr = mDInput->CreateDeviceEx(GUID_SysKeyboard,IID_IDirectInputDevice7,reinterpret_cast<LPVOID *>(&mDIKeyboard),NULL);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Obtaining device for system keyboard");
		return false;
	}
	hr = mDIKeyboard->SetDataFormat(&c_dfDIKeyboard);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting data format for system keyboard DI device");
		return false;
	}
	hr = mDIKeyboard->SetCooperativeLevel(mWindow,DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting system keyboard cooperative level");
		return false;
	}
	dipd.diph.dwSize = sizeof(DIPROPDWORD);
	dipd.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	dipd.diph.dwObj = 0;
	dipd.diph.dwHow = DIPH_DEVICE;
	dipd.dwData = 32;
	hr = mDIKeyboard->SetProperty(DIPROP_BUFFERSIZE,&dipd.diph);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting buffer size property for system keyboard");
		return false;
	}
	mDIKeyboard->Acquire();
	mInit = true;
	return true;
}

GArray<GDisplay::DisplayProps>
GpD3D7Display::ListAvailableDisplays()
{
	HRESULT hr;
	std::vector<GDisplay::DisplayProps> VecDisplays;

	hr = mDD->EnumDisplayModes(0,NULL,&VecDisplays,GpD3D7EnumModesCallback);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Enumerating display modes");
		return NULL;
	}
	GDisplay::DisplayProps *pDisplays = new GDisplay::DisplayProps[VecDisplays.size()];
	GDisplay::DisplayProps *pThisDisplay = pDisplays;
	for(std::vector<GDisplay::DisplayProps>::iterator i=VecDisplays.begin();i!=VecDisplays.end();i++)
		*(pThisDisplay++) = *i;
	GArray<GDisplay::DisplayProps> r(VecDisplays.size());
	memcpy(&r[0],pDisplays,sizeof(GDisplay::DisplayProps) * VecDisplays.size());
	delete [] pDisplays;
	return r;
}

void 
GpD3D7Display::Camera(pGCamera camera)
{
	mCamera = (pGpD3D7Camera)camera;
}

void 
GpD3D7Display::DoFrame(tTime time, tTime frametime)
{
	mDevice->Clear(0,NULL,D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,0x0,1.0f,0);

	mDevice->BeginScene();
	mCamera->Draw(mGeoLod);
	mDevice->EndScene();

	ProcessInputs();

	if(ProcessEvents()) {
		if(mFullScreen) {
			if(mPrimarySurface->Flip(NULL,0) == DDERR_SURFACELOST)
				mDD->RestoreAllSurfaces();
		}
		else {
			mDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
			if(mPrimarySurface->Blt(&mPosition,mBackSurface,NULL,DDBLT_WAIT,NULL) == DDERR_SURFACELOST)
				mDD->RestoreAllSurfaces();
		}
	}
}

void
GpD3D7Display::CaptureMouse(bool enable)
{
	if(mDIMouse) {
		if(enable) {
			mDIMouse->SetCooperativeLevel(mWindow,DISCL_EXCLUSIVE | DISCL_FOREGROUND);
		}
		else {
			mDIMouse->SetCooperativeLevel(mWindow,DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
		}
	}
}

// operations --------------------------------------------------------
bool
GpD3D7Display::ProcessEvents()
{
	MSG msg;

	while(PeekMessage(&msg,NULL,0,0,PM_NOREMOVE)) {
		if(!GetMessage(&msg,NULL,0,0)) {
			SetWindowLong(mWindow,GWL_USERDATA,0);
			mQuitting = true;
			if(mFullScreen) {
				mDD->RestoreDisplayMode();
				mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
			}
			gMainMessageQueue->EnqueueMessage(new GQuitMessage(this));
			return false;
		}
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return true;
}

void
GpD3D7Display::ProcessInputs()
{
	HRESULT hr;
	DIDEVICEOBJECTDATA DInputData[32];
	DWORD DataCount = 32;

	hr = mDIMouse->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),DInputData,&DataCount,0);
	if(hr == DIERR_INPUTLOST) {
		PostMessage(mWindow,WM_SYNCACQUIRE,1,0);
		return;
	}
	else if(hr == DIERR_NOTACQUIRED) {
		return;
	}
	else if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Getting system mouse device data");
		return;
	}
	for(int c=0;c<DataCount;c++) {
		switch(DInputData[c].dwOfs) {
		case DIMOFS_BUTTON0:
			gMainMessageQueue->EnqueueMessage(new GMouseButtonMessage(this,"mouse 0",(DInputData[c].dwData & 0x80) == 0x80));
			break;
		case DIMOFS_BUTTON1:
			gMainMessageQueue->EnqueueMessage(new GMouseButtonMessage(this,"mouse 1",(DInputData[c].dwData & 0x80) == 0x80));
			break;
		case DIMOFS_BUTTON2:
			gMainMessageQueue->EnqueueMessage(new GMouseButtonMessage(this,"mouse 2",(DInputData[c].dwData & 0x80) == 0x80));
			break;
		case DIMOFS_BUTTON3:
			gMainMessageQueue->EnqueueMessage(new GMouseButtonMessage(this,"mouse 3",(DInputData[c].dwData & 0x80) == 0x80));
			break;
		case DIMOFS_X:
			gMainMessageQueue->EnqueueMessage(new GMouseMotionMessage(this,"mouse x",static_cast<float>(static_cast<long>(DInputData[c].dwData)) / static_cast<float>(mSize.cx),0));
			break;
		case DIMOFS_Y:
			gMainMessageQueue->EnqueueMessage(new GMouseMotionMessage(this,"mouse y",static_cast<float>(static_cast<long>(DInputData[c].dwData)) / static_cast<float>(mSize.cy),0));
			break;
		case DIMOFS_Z:
			gMainMessageQueue->EnqueueMessage(new GMouseMotionMessage(this,"mouse z",DInputData[c].dwData,0));
			break;
		}
	}
	DataCount = 32;
	hr = mDIKeyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),DInputData,&DataCount,0);
	if(hr == DIERR_INPUTLOST) {
		PostMessage(mWindow,WM_SYNCACQUIRE,1,0);
		return;
	}
	else if(hr == DIERR_NOTACQUIRED) {
		return;
	}
	else if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Getting system keyboard device data");
		return;
	}
	BYTE KeyState[256];
	GetKeyboardState(KeyState);
	for(c=0;c<DataCount;c++) {
		char ch[2] = "\0";

		UINT vkey = MapVirtualKey(DInputData[c].dwOfs,1);
		GetKeyboardState(KeyState);
		ToAscii(vkey,DInputData[c].dwOfs,KeyState,reinterpret_cast<WORD *>(ch),0);
		gMainMessageQueue->EnqueueMessage(new GKeyboardMessage(this,DIK_KeyNames[DInputData[c].dwOfs],ch,(DInputData[c].dwData & 0x80) == 0x80));
	}
}

HRESULT
GpD3D7Display::SetMouseAcquire(bool set)
{
	if(mDIMouse == NULL)
		return S_FALSE;
	if(set)
		mDIMouse->Acquire();
	else
		mDIMouse->Unacquire();
	return S_OK;
}

bool
GpD3D7Display::CreateFullScreen()
{
	HRESULT hr;
	DDSURFACEDESC2 ddsd = { sizeof(DDSURFACEDESC2) };
	DDSCAPS2 ddsc;

	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_COMPLEX | DDSCAPS_FLIP | DDSCAPS_PRIMARYSURFACE | DDSCAPS_3DDEVICE;
	ddsd.dwBackBufferCount = 1;
	hr = mDD->CreateSurface(&ddsd,&mPrimarySurface,NULL);
	if(FAILED(hr)) {
		mDD->RestoreDisplayMode();
		mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		CREATE_DIRECTX_ERROR(mError,hr,"Creating fullscreen rendering surfaces");
		return false;
	}
	ZeroMemory(&ddsc,sizeof(DDSCAPS2));
	ddsc.dwCaps = DDSCAPS_BACKBUFFER;
	hr = mPrimarySurface->GetAttachedSurface(&ddsc,&mBackSurface);
	if(FAILED(hr)) {
		mDD->RestoreDisplayMode();
		mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		CREATE_DIRECTX_ERROR(mError,hr,"Obtaining back buffer from rendering surface chain");
		return false;
	}	
	return true;
}

bool
GpD3D7Display::CreateBackSurface()
{
	HRESULT hr;
	DDSURFACEDESC2 ddsd;

	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
	ddsd.dwWidth = mSize.cx;
	ddsd.dwHeight = mSize.cy;
	hr = mDD->CreateSurface(&ddsd,&mBackSurface,NULL);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Creating back buffer for rendering");
		return false;
	}
	return true;
}

bool
GpD3D7Display::CreateDevice()
{
	HRESULT hr;
	DDSURFACEDESC2 ddsd;
	D3DMATERIAL7 mtrl;
	D3DMATRIX mat;
	LPDIRECT3D7 ID3D;
	
	hr = mDD->QueryInterface(IID_IDirect3D7,reinterpret_cast<void **>(&ID3D));
	if(FAILED(hr)) {
		if(mFullScreen) {
			mDD->RestoreDisplayMode();
			mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		}
		CREATE_DIRECTX_ERROR(mError,hr,"Querying IDirect3D7 interface");
		return false;
	}
	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	mDD->GetDisplayMode(&ddsd);
	if(ddsd.ddpfPixelFormat.dwRGBBitCount <= 8) {
		if(mFullScreen) {
			mDD->RestoreDisplayMode();
			mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		}
		CREATE_GAUGE3D_ERROR(mError,GGaugeError::INVALID_DISPLAY,"Checking display mode");
		return false;
	}
	// Try T&L device first
	hr = ID3D->CreateDevice(IID_IDirect3DTnLHalDevice,mBackSurface,&mDevice);
	if(FAILED(hr)) {
		// No T&L device, so get the standard HAL device.  Fallback to software rendering.
		hr = ID3D->CreateDevice(IID_IDirect3DHALDevice,mBackSurface,&mDevice);
		if(FAILED(hr)) {
			RELEASE_INTERFACE(ID3D);
			if(mFullScreen) {
				mDD->RestoreDisplayMode();
				mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
			}
			CREATE_DIRECTX_ERROR(mError,hr,"Creating rendering device");
			return false;
		}
	}
	RELEASE_INTERFACE(ID3D);
	ZeroMemory(&mtrl,sizeof(D3DMATERIAL7));
	mtrl.ambient.r = mtrl.ambient.g = mtrl.ambient.b = 1.0f;
	mtrl.diffuse.r = mtrl.diffuse.g = mtrl.diffuse.b = 1.0f;
	mDevice->SetMaterial(&mtrl);
	mDevice->SetRenderState(D3DRENDERSTATE_AMBIENT,0xFFFFFFFF);
	mDevice->SetRenderState(D3DRENDERSTATE_LIGHTING,TRUE);
	mDevice->SetRenderState(D3DRENDERSTATE_CULLMODE,D3DCULL_CCW);
	mDevice->SetRenderState(D3DRENDERSTATE_ZENABLE,TRUE);
	mDevice->SetRenderState(D3DRENDERSTATE_SPECULARENABLE,TRUE);
	mDevice->SetRenderState(D3DRENDERSTATE_DIFFUSEMATERIALSOURCE,D3DMCS_MATERIAL);
	mDevice->SetRenderState(D3DRENDERSTATE_SPECULARMATERIALSOURCE,D3DMCS_MATERIAL);
	mDevice->SetRenderState(D3DRENDERSTATE_EMISSIVEMATERIALSOURCE,D3DMCS_MATERIAL);
	mDevice->SetRenderState(D3DRENDERSTATE_AMBIENTMATERIALSOURCE,D3DMCS_MATERIAL);
    mat._11 = mat._22 = mat._33 = mat._44 = 1.0f;
    mat._12 = mat._13 = mat._14 = mat._41 = 0.0f;
    mat._21 = mat._23 = mat._24 = mat._42 = 0.0f;
    mat._31 = mat._32 = mat._34 = mat._43 = 0.0f;
	mDevice->SetTransform(D3DTRANSFORMSTATE_WORLD,&mat);
	mDevice->SetTransform(D3DTRANSFORMSTATE_VIEW,&mat);
	mDevice->SetTransform(D3DTRANSFORMSTATE_PROJECTION,&mat);
   	return true;
}

bool
GpD3D7Display::CreateZBuffer()
{
	HRESULT hr;
	DDSURFACEDESC2 ddsd;

	ZeroMemory(&ddsd,sizeof(DDSURFACEDESC2));
	ddsd.dwSize = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | DDSCAPS_VIDEOMEMORY;
	if(mFullScreen) {
		DDSURFACEDESC2 fs_ddsd = { sizeof(DDSURFACEDESC2) };
		mBackSurface->GetSurfaceDesc(&fs_ddsd);
		ddsd.dwWidth = fs_ddsd.dwWidth;
		ddsd.dwHeight = fs_ddsd.dwHeight;
	}
	else {
		ddsd.dwWidth = mSize.cx;
		ddsd.dwHeight = mSize.cy;
	}
	// Preliminary: assuming 16-bit ZBuffer at the moment
	ddsd.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	ddsd.ddpfPixelFormat.dwFlags = DDPF_ZBUFFER;
	ddsd.ddpfPixelFormat.dwZBufferBitDepth = 16;
	ddsd.ddpfPixelFormat.dwZBitMask = 0xFFFF;
	hr = mDD->CreateSurface(&ddsd,&mZBuffer,NULL);
	if(FAILED(hr)) {
		if(mFullScreen) {
			mDD->RestoreDisplayMode();
			mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		}
		CREATE_DIRECTX_ERROR(mError,hr,"Creating ZBuffer surface");
		return false;
	}
	hr = mBackSurface->AddAttachedSurface(mZBuffer);
	if(FAILED(hr)) {
		RELEASE_INTERFACE(mZBuffer);
		if(mFullScreen) {
			mDD->RestoreDisplayMode();
			mDD->SetCooperativeLevel(mWindow,DDSCL_NORMAL);
		}
		CREATE_DIRECTX_ERROR(mError,hr,"Attaching Z-Buffer to rendering surface");
		return false;
	}
	return true;
}

void
GpD3D7Display::SetViewport()
{
	D3DVIEWPORT7 view;

	view.dwX = 0;
	view.dwY = 0;
	if(mFullScreen) {
		DDSURFACEDESC2 ddsd = { sizeof(DDSURFACEDESC2) };
		mBackSurface->GetSurfaceDesc(&ddsd);
		view.dwWidth = ddsd.dwWidth;
		view.dwHeight = ddsd.dwHeight;
	}
	else {
		view.dwWidth = mSize.cx;
		view.dwHeight = mSize.cy;
	}
	view.dvMinZ = 0.0f;
	view.dvMaxZ = 1.0f;
	mDevice->SetViewport(&view);
}

LRESULT
GpD3D7Display::WindowProc(HWND hWnd,UINT Msg,WPARAM wParam,LPARAM lParam)
{
	switch(Msg) {
	case WM_SYNCACQUIRE:
		SetMouseAcquire(wParam == 1);
		return 0;
	case WM_ACTIVATE:
		SetMouseAcquire(wParam != WA_INACTIVE);
		return 1;
	case WM_ENTERMENULOOP:
	case WM_ENTERSIZEMOVE:
		SetMouseAcquire(false);
		break;
	case WM_EXITMENULOOP:
	case WM_EXITSIZEMOVE:
		if(GetActiveWindow() == hWnd || !IsIconic(hWnd))
			SetMouseAcquire(true);
		else
			SetMouseAcquire(false);
		break;
	case WM_KEYDOWN:
		switch(wParam) {
		case VK_PRIOR:
			mGeoLod /= 4.0f;
			break;
		case VK_NEXT:
			mGeoLod *= 4.0f;
			break;
		case VK_ESCAPE:
			SendMessage(hWnd,WM_CLOSE,0,0);
			break;
		}
		return 0;
	case WM_MOVE:
		GetClientRect(mWindow,&mPosition);
		mSize.cx = mPosition.right;
		mSize.cy = mPosition.bottom;
		ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.left));
		ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.right));
		return 0;
	case WM_SIZE:
		GetClientRect(mWindow,&mPosition);
		mSize.cx = mPosition.right;
		mSize.cy = mPosition.bottom;
		ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.left));
		ClientToScreen(mWindow,reinterpret_cast<POINT *>(&mPosition.right));
		if(!mQuitting && mBackSurface && mZBuffer) {
			mBackSurface->DeleteAttachedSurface(0,mZBuffer);
			RELEASE_INTERFACE(mZBuffer);
			if(mFullScreen) {
				RELEASE_INTERFACE(mBackSurface);
				RELEASE_INTERFACE(mPrimarySurface);
				CreateFullScreen();
			}
			else {
				RELEASE_INTERFACE(mBackSurface);
				CreateBackSurface();
			}
			CreateZBuffer();
			if(mDevice) {
				mDevice->SetRenderTarget(mBackSurface,0);
				SetViewport();
			}
		}
		return 0;
	case WM_CLOSE:
		mQuitting = true;
		DestroyWindow(hWnd);
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hWnd,Msg,wParam,lParam);
}

GAUGE3D_DECLARE_INTERFACES(GpD3D7Display)
	GAUGE3D_DECLARE_BASECLASS(GDisplay)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
