/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Direct3D 7 Rendering Plugin
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7sprite.cpp
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: Sprite
 *  Description: D3D implementation of a sprite
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "d3d7sprite.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_WINDOWS_H__INCLUDED__
#include <windows.h>
#define __SYSTEM_WINDOWS_H__INCLUDED__
#endif

#ifndef __SYSTEM_D3DX_H__INCLUDED__
#include <d3dx.h>
#define __SYSTEM_D3DX_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpD3D7Sprite::GpD3D7Sprite(pGpD3D7Sprite* link,LPDIRECT3DDEVICE7 device)
	: mModel(NULL),
	  mPosition(0.0f, 0.0f, 0.0f),
	  mAngle(GVECTOR_I, 0.0f),
	  mFullBright(false),
	  mShow(false),
	  mScaleFactor(1.0f),
	  mNext(*link),
	  mPrev(link),
	  mDevice(device)
{
	mError = NULL;
	mColor[0] = 1.0;
	mColor[1] = 1.0;
	mColor[2] = 1.0;
	mColor[3] = 1.0;
	mLightProps[0] = 0.0;
	mLightProps[1] = 0.0;
	mLightProps[2] = 0.0;

	mVertexData = NULL;
	mTexcoordData = NULL;
	mIndexData = NULL;
	mVertexDataSize = 0;
	mTexcoordDataSize = 0;
	mIndexDataSize = 0;

	AddToList(link);
}

GpD3D7Sprite::~GpD3D7Sprite()
{
	if(mVertexData)
		delete [] mVertexData;
	if(mTexcoordData)
		delete [] mTexcoordData;
	if(mIndexData)
		delete [] mIndexData;
	RELEASE_ERROR(mError);
}

// operations --------------------------------------------------------	
void
GpD3D7Sprite::AddToList(pGpD3D7Sprite* link)
{
	mNext = *link;
	mPrev = link;
	
	*mPrev = this;
	if(mNext != NULL)
		mNext->mPrev = &mNext;
}

//Draw sprite via Direct3D
void
GpD3D7Sprite::Draw(GVector viewpoint, GQuaternion viewangle, GFrustum viewFrustum, float geoLod)
{
	if(mShow)
	{
		D3DMATRIX mat;
		mDevice->GetTransform(D3DTRANSFORMSTATE_WORLD,&mat);
	
		ApplyOrigin(viewpoint, viewangle, viewFrustum);
		ApplyAngle(viewpoint, viewangle, viewFrustum);
		ApplyScale(viewpoint, viewangle, viewFrustum);
		ApplyLightProps();
		ApplyColor();
		
		DrawGeometry(viewpoint, viewangle, viewFrustum, geoLod);
		
		mDevice->SetTransform(D3DTRANSFORMSTATE_WORLD,&mat);
	}
}

void
GpD3D7Sprite::ApplyOrigin(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	D3DMATRIX mat(1,0,0,0,
		          0,1,0,0,
				  0,0,1,0,
				  mPosition.x,mPosition.y,mPosition.z,1);

	mDevice->MultiplyTransform(D3DTRANSFORMSTATE_WORLD,&mat);

	viewpoint -= mPosition;
}

void
GpD3D7Sprite::ApplyAngle(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	D3DMATRIX mat;
	GVector axis;
	GCoordinate angle;
	D3DXVECTOR3 d3daxis;
	
	angle = mAngle.Angle();
	if(!IsApproxZero(angle))
	{
		axis = mAngle.Axis();
		d3daxis.x = axis.x;
		d3daxis.y = axis.y;
		d3daxis.z = axis.z;
		
		D3DXMatrixRotationAxis(reinterpret_cast<D3DXMATRIX *>(&mat), &d3daxis, angle);
		mDevice->MultiplyTransform(D3DTRANSFORMSTATE_WORLD,&mat);
	}

	viewangle = mAngle * viewangle;
}

void
GpD3D7Sprite::ApplyScale(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	if(!IsApproxZero(mScaleFactor - 1.0f)) {
		D3DMATRIX mat;
		D3DXMatrixScaling(reinterpret_cast<D3DXMATRIX *>(&mat),mScaleFactor,mScaleFactor,mScaleFactor);
		mDevice->MultiplyTransform(D3DTRANSFORMSTATE_WORLD,&mat);
		mDevice->SetRenderState(D3DRENDERSTATE_NORMALIZENORMALS,TRUE);

		//rather than grow the object, we shrink the viewspace.
		GCoordinate shrinkFactor = 1.0f / mScaleFactor;
		
		viewpoint *= shrinkFactor;
		viewFrustum.mWidth    *= shrinkFactor;
		viewFrustum.mHeight   *= shrinkFactor;
		viewFrustum.mNear     *= shrinkFactor;
		viewFrustum.mFar      *= shrinkFactor;
	}
	else {
		mDevice->SetRenderState(D3DRENDERSTATE_NORMALIZENORMALS,FALSE);
	}
}

void
GpD3D7Sprite::ApplyLightProps()
{
	D3DMATERIAL7 mtrl;

	ZeroMemory(&mtrl,sizeof(D3DMATERIAL7));
	mtrl.specular.r = mLightProps[0];
	mtrl.specular.g = mLightProps[0];
	mtrl.specular.b = mLightProps[0];
	mtrl.specular.a = mColor[3];
	mtrl.power = mLightProps[1];
	mtrl.emissive.r = mColor[0] * mLightProps[2];
	mtrl.emissive.g = mColor[1] * mLightProps[2];
	mtrl.emissive.b = mColor[2] * mLightProps[2];
	mtrl.emissive.a = mColor[3];
	mDevice->SetMaterial(&mtrl);
}

void
GpD3D7Sprite::ApplyColor()
{
	D3DMATERIAL7 mtrl;

	mDevice->GetMaterial(&mtrl);
	mtrl.diffuse.r = mColor[0];
	mtrl.diffuse.g = mColor[1];
	mtrl.diffuse.b = mColor[2];
	mtrl.diffuse.a = mColor[3];
	if(mFullBright) {
		mtrl.ambient.r = mColor[0];
		mtrl.ambient.g = mColor[1];
		mtrl.ambient.b = mColor[2];
		mtrl.ambient.a = mColor[3];
	}
	else {
		ZeroMemory(&mtrl.ambient,sizeof(D3DCOLORVALUE));
	}
	mDevice->SetMaterial(&mtrl);
}

void
GpD3D7Sprite::DrawGeometry(GVector& viewpoint, GQuaternion& viewangle,
                           GFrustum& viewFrustum, float geoLod)
{
	GModel::RenderInfo::RenderSet set;
	bool hasAlpha = false;
	int numSets;
	int setNum;
	
	numSets = mRenderInfo->StartFrame(mAnimState, viewpoint, viewangle, viewFrustum, geoLod);
	
	HRESULT hr;
	LPDIRECT3D7 lpD3D;
	LPDIRECT3DVERTEXBUFFER7 lpVertBuffer;
	D3DVERTEXBUFFERDESC vbd = { sizeof(D3DVERTEXBUFFERDESC) };
	D3DDRAWPRIMITIVESTRIDEDDATA stridedData;
	DWORD vertexFormat = D3DFVF_XYZ;

	hasAlpha = HasAlpha();
	if(hasAlpha)
	{
		mDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE);
		mDevice->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_SRCALPHA);
		mDevice->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCALPHA);
	}
		
	for(setNum = 0; setNum < numSets; setNum++)
	{
		mRenderInfo->GetStaticInfo(&set, setNum);
		
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::VERTEX_POSITION)
		{
			if(mVertexDataSize < set.mNumVertices)
			{
				if(mVertexDataSize > 0)
					delete [] mVertexData;
				mVertexData = new GVector[set.mNumVertices * 2];
				mVertexDataSize = set.mNumVertices;
			}
			set.mVertices = mVertexData;
			set.mNormals  = mVertexData + set.mNumVertices;
		}
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::VERTEX_TEXCOORD)
		{
			if(mTexcoordDataSize < set.mNumVertices)
			{
				if(mTexcoordDataSize > 0)
					delete [] mTexcoordData;
				mTexcoordData = new GVector[set.mNumVertices];
				mTexcoordDataSize = set.mNumVertices;
			}
			set.mTexCoords = mTexcoordData;
		}
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::INDEX_LIST)
		{
			if(mIndexDataSize < set.mNumIndices)
			{
				if(mIndexDataSize > 0)
					delete [] mIndexData;
				mIndexData = new uint16[set.mNumIndices];
				mIndexDataSize = set.mNumIndices;
			}
			set.mIndices = mIndexData;
		}
		
		mRenderInfo->GetDynamicInfo(&set, setNum);
		
		ZeroMemory(&stridedData,sizeof(D3DDRAWPRIMITIVESTRIDEDDATA));
		stridedData.position.lpvData = &set.mVertices[0];
		stridedData.position.dwStride = sizeof(GVector);
		
		//if the object is not fullbright, normals are needed for lighting
		if(!mFullBright)
		{
			stridedData.normal.lpvData = &set.mNormals[0];
			stridedData.normal.dwStride = sizeof(GVector);
			vertexFormat |= D3DFVF_NORMAL;
		}
		else
		{
			DWORD diffuseColor = (static_cast<DWORD>(mColor[0] * 255.0f) << 16)
			                   | (static_cast<DWORD>(mColor[1] * 255.0f) << 8) 
			                   | (static_cast<DWORD>(mColor[2] * 255.0f) << 0) 
			                   | (static_cast<DWORD>(mColor[3] * 255.0f) << 24);
			stridedData.diffuse.lpvData = &diffuseColor;
			stridedData.diffuse.dwStride = 0;
			vertexFormat |= D3DFVF_DIFFUSE;
		}
		
		if((mTexture != NULL) && (set.mTexCoords != NULL))
		{
			stridedData.textureCoords[0].lpvData = &set.mTexCoords[0];
			stridedData.textureCoords[0].dwStride = sizeof(GVector);
			vertexFormat |= D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0);
			mDevice->SetTexture(0,*mTexture);
			mDevice->SetTextureStageState(0,D3DTSS_TEXCOORDINDEX,0);
			mDevice->SetTextureStageState(0,D3DTSS_COLORARG1,D3DTA_TEXTURE);
			mDevice->SetTextureStageState(0,D3DTSS_COLORARG2,D3DTA_DIFFUSE);
			mDevice->SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
			mDevice->SetTextureStageState(0,D3DTSS_ALPHAARG1,D3DTA_TEXTURE);
			mDevice->SetTextureStageState(0,D3DTSS_ALPHAOP,D3DTOP_SELECTARG1);
			mDevice->SetTextureStageState(0,D3DTSS_MINFILTER,D3DFILTER_LINEAR);
			mDevice->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DFILTER_LINEAR);
		}
		
		vbd.dwCaps = D3DVBCAPS_WRITEONLY;
		vbd.dwFVF = D3DFVF_TLVERTEX;
		vbd.dwNumVertices = set.mNumVertices;
		mDevice->GetDirect3D(&lpD3D);
		hr = lpD3D->CreateVertexBuffer(&vbd,&lpVertBuffer,0);
		lpD3D->Release();
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating vertex buffer");
			return;
		}
		DWORD VertexOp = D3DVOP_TRANSFORM | D3DVOP_CLIP;
		if(!mFullBright)
			VertexOp |= D3DVOP_LIGHT;
		hr = lpVertBuffer->ProcessVerticesStrided(VertexOp,0,set.mNumVertices,&stridedData,vertexFormat,mDevice,0);
		if(FAILED(hr)) {
			lpVertBuffer->Release();
			CREATE_DIRECTX_ERROR(mError,hr,"Processing vertex buffer");
			return;
		}

		D3DPRIMITIVETYPE type;

		switch(set.mIndexType) {
		default:
		case GModel::RenderInfo::RenderSet::TRIANGLES:
			type = D3DPT_TRIANGLELIST;
			break;
		case GModel::RenderInfo::RenderSet::LINES:
			type = D3DPT_LINELIST;
			break;
		case GModel::RenderInfo::RenderSet::POINTS:
			type = D3DPT_POINTLIST;
			break;
		}

		mDevice->DrawIndexedPrimitiveVB(type,lpVertBuffer,0,set.mNumVertices,set.mIndices,set.mNumIndices,0);

		lpVertBuffer->Release();
		if((mTexture != NULL) && (set.mTexCoords != NULL))
			mDevice->SetTexture(0,NULL);
	}

	if(hasAlpha)
		mDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE,FALSE);
}

// operations ------------------------------------------------------
//Call Show to tell the renderer to start drawing this object whenever it is visible.
void
GpD3D7Sprite::Show()
{
	mShow = true;
}
//Call Hide to tell the renderer not to draw this object at all, and not to use it in
// any calculations such as shadow casting.
void
GpD3D7Sprite::Hide()
{
	mShow = false;
}

//Call Remove to permenantly remove the sprite from the scene.
void
GpD3D7Sprite::Remove()
{
	if(mPrev != NULL)
		*mPrev = mNext;
	if(mNext != NULL)
		mNext->mPrev = mPrev;
	mPrev = NULL;
	mNext = NULL;
}

// access ----------------------------------------------------------
//A GModel is used to represent the geometry of the object.
void
GpD3D7Sprite::Model(pGModel model)
{
	mModel = model;
	if(model == NULL)
		mRenderInfo = NULL;
	else
		mRenderInfo = model->CreateRenderInfo();
}

void
GpD3D7Sprite::AnimState(GModel::pAnimState anim)
{
	mAnimState = anim;
}

//Set the current texture like so.  The texture object must have come from
// the same renderer object as the sprite.
void
GpD3D7Sprite::Texture(pGTexture texture)
{
	mTexture = texture;
}

//The origin is the location of the object.
// Angles is the direction the object is pointing.
void
GpD3D7Sprite::Position(const GVector& position)
{
	mPosition = position;
}
void
GpD3D7Sprite::Angle(const GQuaternion& angle)
{
	mAngle = angle;
}

//Use these to set the solid color of the object.  Set to full white to ignore.
void
GpD3D7Sprite::Color(float red, float green, float blue)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
}
void
GpD3D7Sprite::Color(float red, float green, float blue, float alpha)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
	mColor[3] = alpha;
}
void
GpD3D7Sprite::Color3v(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
}
void
GpD3D7Sprite::Color4v(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
	mColor[3] = color[3];
}
void
GpD3D7Sprite::Alpha(float alpha)
{
	mColor[3] = alpha;
}

//LightProps sets the light reflectiveness properties of the object.
void
GpD3D7Sprite::LightProps(float specular, float shininess, float emission)
{
	mLightProps[0] = specular;
	mLightProps[1] = shininess;
	mLightProps[2] = emission;
}
//If fullbright is true, lighting will be ignored and the object will be drawn
// at maximum brightness.
void
GpD3D7Sprite::FullBright(bool fullBright)
{
	mFullBright = fullBright;
}

void
GpD3D7Sprite::Scale(GCoordinate magnification)
{
	mScaleFactor = magnification;
}

bool
GpD3D7Sprite::HasAlpha()
{
	if(mColor[3] != 1.0)
		return true;
	return false;
}

GAUGE3D_DECLARE_INTERFACES(GpD3D7Sprite)
	GAUGE3D_DECLARE_BASECLASS(GSprite)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
