/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7texture.cpp
 *      Package: Direct3D Rendering Plugin
 *    Component: Texture
 *  Description: Represents a cached texture
 * Dependencies: GTexture
 *               GImage
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "d3d7texture.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_IOSTREAM__INCLUDED__
#include <iostream>
#define __SYSTEM_IOSTREAM__INCLUDED__
#endif

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

static HRESULT CALLBACK GpD3D7EnumTextureFormats(LPDDPIXELFORMAT lpDDPixFmt,LPVOID lpContext)
{
	std::vector<DDPIXELFORMAT> *Formats = static_cast<std::vector<DDPIXELFORMAT> *>(lpContext);
	char buf[1024];

	sprintf(buf,"Found texture: dwFlags = 0x%08X; dwRGBBitCount = %d.\n",lpDDPixFmt->dwFlags,lpDDPixFmt->dwRGBBitCount);
	OutputDebugString(buf);
	Formats->push_back(*lpDDPixFmt);
	return D3DENUMRET_OK;
}

// class member functions ============================================

// lifecycle ---------------------------------------------------------
GpD3D7Texture::GpD3D7Texture(IDirectDraw7 *DD,IDirect3DDevice7 *Device) : 
	mDD(DD), mDevice(Device), mSurface(NULL), mTextureDepthLimit(32)
{
	mError = NULL;
	mDevice->EnumTextureFormats(GpD3D7EnumTextureFormats,&mTextureFormats);
}

GpD3D7Texture::~GpD3D7Texture()
{
	if(mSurface) {
		mSurface->Release();
		mSurface = NULL;
	}
	RELEASE_ERROR(mError);
}

// operations --------------------------------------------------------
void
GpD3D7Texture::Image(pGImage image)
{
	GImage::FormatDesc ImageFormat;

	if(mSurface) {
		mSurface->Release();
		mSurface = NULL;
	}
	ImageFormat = image->Format();
	//check if image's width and height are powers of two.
	int32 w, h;
	bool needScale = false;
	for(w=1 << 30;w>ImageFormat.mWidth;w >>= 1);
	if(w < ImageFormat.mWidth) {
		w <<= 1;        //lowest power of 2 that is greater than the width.
		needScale = true;
	}
	for(h=1 << 30;h>ImageFormat.mHeight;h >>= 1);
	if(h < ImageFormat.mHeight) {
		h <<= 1;        //lowest power of 2 that is greater than the height.
		needScale = true;
	}
	if(needScale) {
		HRESULT hr;
		DDSURFACEDESC2 ddsd = { sizeof(DDSURFACEDESC2) };
		LPDIRECTDRAWSURFACE7 UnscaledSurface;
		RECT ImageRect;

		LPDDPIXELFORMAT SurfaceFormat = SelectTextureFormat(ImageFormat);
		if(SurfaceFormat == NULL) {
			CREATE_GAUGE3D_ERROR(mError,GGaugeError::NOT_FOUND,"Obtaining texture format for image");
			return;
		}
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth = ImageFormat.mWidth;
		ddsd.dwHeight = ImageFormat.mHeight;
		memcpy(&ddsd.ddpfPixelFormat,SurfaceFormat,sizeof(DDPIXELFORMAT));
		hr = mDD->CreateSurface(&ddsd,&UnscaledSurface,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating unscaled offscreen surface for texture");
			return;
		}
		ImageRect.left = ImageRect.top = 0;
		ImageRect.right = ImageFormat.mWidth;
		ImageRect.bottom = ImageFormat.mHeight;
		GArray<uint8> ImageData(image->Data());
		hr = UnscaledSurface->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Locking offscreen texture surface");
			return;
		}
		CopyImageToTexture(ImageFormat,ddsd,ImageRect,ImageData);
		UnscaledSurface->Unlock(NULL);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT | DDSD_TEXTURESTAGE;
		ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
		ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
		ddsd.dwWidth = w;
		ddsd.dwHeight = h;
		ddsd.dwTextureStage = 0;
		memcpy(&ddsd.ddpfPixelFormat,SurfaceFormat,sizeof(DDPIXELFORMAT));
		hr = mDD->CreateSurface(&ddsd,&mSurface,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating scaled texture surface");
			return;
		}
		DDBLTFX bfx = { sizeof(DDBLTFX) };
		hr = mSurface->Blt(NULL,UnscaledSurface,NULL,DDBLT_WAIT,&bfx);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Blitting texture image from offscreen to texture surface");
		}
		UnscaledSurface->Release();
	}
	else {
		DDSURFACEDESC2 ddsd = { sizeof(DDSURFACEDESC2) };
		HRESULT hr;
		DWORD width,height;
		GImage::FormatDesc ImageFormat;
		RECT ImageRect;
	
		if(mSurface) {
			mSurface->Release();
			mSurface = NULL;
		}
		ImageFormat = image->Format();
		LPDDPIXELFORMAT SurfaceFormat = SelectTextureFormat(ImageFormat);
		if(SurfaceFormat == NULL) {
			CREATE_GAUGE3D_ERROR(mError,GGaugeError::NOT_FOUND,"Obtaining texture format for image");
			return;
		}
		width = ImageFormat.mWidth;
		height = ImageFormat.mHeight;
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT | DDSD_TEXTURESTAGE;
		ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
		ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
		ddsd.dwWidth = width;
		ddsd.dwHeight = height;
		ddsd.dwTextureStage = 0;
		memcpy(&ddsd.ddpfPixelFormat,SurfaceFormat,sizeof(DDPIXELFORMAT));
		hr = mDD->CreateSurface(&ddsd,&mSurface,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Creating texture surface");
			return;
		}
		hr = mSurface->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);
		if(FAILED(hr)) {
			CREATE_DIRECTX_ERROR(mError,hr,"Locking texture surface");
			return;
		}
		ImageRect.left = ImageRect.top = 0;
		ImageRect.right = ImageFormat.mWidth;
		ImageRect.bottom = ImageFormat.mHeight;
		GArray<uint8> ImageData(image->Data());
		CopyImageToTexture(ImageFormat,ddsd,ImageRect,ImageData);
		mSurface->Unlock(NULL);
	}
}

int
GpD3D7Texture::CountBits(DWORD Mask)
{
	DWORD TestMask = 1;
	int Bits = 0;

	for(int i=0;i<32;i++) {
		if(Mask & TestMask)
			++Bits;
		TestMask <<= 1;
	}
	return Bits;
}

int
GpD3D7Texture::CalcShift(DWORD Mask)
{
	int Bits = 0;

	while((Mask & 1) == 0) {
		Mask >>= 1;
		++Bits;
	}
	return Bits;
}

LPDDPIXELFORMAT
GpD3D7Texture::SelectTextureFormat(GImage::FormatDesc& SrcFormat)
{
	int TotalBits;
	LPDDPIXELFORMAT ddpf = NULL;

	TotalBits = min(SrcFormat.mNumChannels * SrcFormat.mBitsPerChannel,mTextureDepthLimit);
	if(TotalBits > 16)
		TotalBits = 32;
	switch(SrcFormat.mNumChannels) {
	case 3:  //RGB
		ddpf = FindTextureFormat(DDPF_RGB,TotalBits);
		if((ddpf == NULL) && (TotalBits == 32))
			ddpf = FindTextureFormat(DDPF_RGB,16);
		break;
	case 4:  //RGBA
		ddpf = FindTextureFormat(DDPF_RGB | DDPF_ALPHAPIXELS,TotalBits);
		if((ddpf == NULL) && (TotalBits == 32))
			ddpf = FindTextureFormat(DDPF_RGB | DDPF_ALPHAPIXELS,16);
		break;
	}
	return ddpf;
}

LPDDPIXELFORMAT
GpD3D7Texture::FindTextureFormat(DWORD Flags,DWORD BitCount)
{
	for(std::vector<DDPIXELFORMAT>::iterator i=mTextureFormats.begin();i!=mTextureFormats.end();i++) {
		if((i->dwFlags == Flags) && (i->dwRGBBitCount == BitCount))
			return &(*i);
	}
	return NULL;
}

void
GpD3D7Texture::CopyImageToTexture(const GImage::FormatDesc& ImageFormat,const DDSURFACEDESC2& SurfaceFormat,const RECT& ImageRect,GArray<uint8> ImageData)
{
	RECT SurfaceRect,CopyRect;
	SurfaceRect.left = SurfaceRect.top = 0;
	SurfaceRect.right = SurfaceFormat.dwWidth;
	SurfaceRect.bottom = SurfaceFormat.dwHeight;
	IntersectRect(&CopyRect,&ImageRect,&SurfaceRect);
	int ImageWidth = ImageRect.right - ImageRect.left;
	int CopyWidth = CopyRect.right - CopyRect.left;

	switch(ImageFormat.mNumChannels) {
	case 4:  //RGBA
		if(ImageFormat.mBitsPerChannel == 8) {
			int RShift,GShift,BShift,AShift;
			int RBitDiff,GBitDiff,BBitDiff,ABitDiff;

			RShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwRBitMask);
			GShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwGBitMask);
			BShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwBBitMask);
			AShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwRGBAlphaBitMask);
			RBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwRBitMask);
			GBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwGBitMask);
			BBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwBBitMask);
			ABitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwRGBAlphaBitMask);
			for(int y=CopyRect.top;y<CopyRect.bottom;y++) {
				uint8 *Image = &ImageData[(y - CopyRect.top) * ImageWidth * 4];
				switch(SurfaceFormat.ddpfPixelFormat.dwRGBBitCount) {
				case 16:
					{
						uint16 *Surface = reinterpret_cast<uint16 *>(static_cast<uint8 *>(SurfaceFormat.lpSurface) + SurfaceFormat.lPitch * y) + CopyRect.left;
						for(int x=0;x<CopyWidth;x++) {
							uint16 *Pixel = Surface++;
							*Pixel = static_cast<uint16>(*(Image++) >> RBitDiff) << RShift;
							*Pixel |= static_cast<uint16>(*(Image++) >> GBitDiff) << GShift;
							*Pixel |= static_cast<uint16>(*(Image++) >> BBitDiff) << BShift;
							*Pixel |= static_cast<uint16>(*(Image++) >> ABitDiff) << AShift;
						}
					}
					break;
				case 32:
					{
						uint32 *Surface = reinterpret_cast<uint32 *>(static_cast<uint8 *>(SurfaceFormat.lpSurface) + SurfaceFormat.lPitch * y) + CopyRect.left;
						for(int x=0;x<CopyWidth;x++) {
							uint32 *Pixel = Surface++;
							*Pixel = static_cast<uint32>(*(Image++)) << RShift;
							*Pixel |= static_cast<uint32>(*(Image++)) << GShift;
							*Pixel |= static_cast<uint32>(*(Image++)) << BShift;
							*Pixel |= static_cast<uint32>(*(Image++)) << AShift;
						}
					}
					break;
				}
			}
		}
		else {
			CREATE_GAUGE3D_ERROR(mError,GGaugeError::INVALID_ARG,"Unable to handle non 32-bit RGBA source image format");
			return;
		}
		break;
	case 3:  //RGB
		if(ImageFormat.mBitsPerChannel == 8) {
			int RShift,GShift,BShift;
			int RBitDiff,GBitDiff,BBitDiff;

			RShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwRBitMask);
			GShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwGBitMask);
			BShift = CalcShift(SurfaceFormat.ddpfPixelFormat.dwBBitMask);
			RBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwRBitMask);
			GBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwGBitMask);
			BBitDiff = ImageFormat.mBitsPerChannel - CountBits(SurfaceFormat.ddpfPixelFormat.dwBBitMask);
			for(int y=CopyRect.top;y<CopyRect.bottom;y++) {
				uint8 *Image = &ImageData[(y - CopyRect.top) * ImageWidth * 3];
				switch(SurfaceFormat.ddpfPixelFormat.dwRGBBitCount) {
				case 16:
					{
						uint16 *Surface = reinterpret_cast<uint16 *>(static_cast<uint8 *>(SurfaceFormat.lpSurface) + SurfaceFormat.lPitch * y) + CopyRect.left;
						for(int x=0;x<CopyWidth;x++) {
							uint16 *Pixel = Surface++;
							*Pixel = static_cast<uint16>(*(Image++) >> RBitDiff) << RShift;
							*Pixel |= static_cast<uint16>(*(Image++) >> GBitDiff) << GShift;
							*Pixel |= static_cast<uint16>(*(Image++) >> BBitDiff) << BShift;
						}
					}
					break;
				case 32:
					{
						uint32 *Surface = reinterpret_cast<uint32 *>(static_cast<uint8 *>(SurfaceFormat.lpSurface) + SurfaceFormat.lPitch * y) + CopyRect.left;
						for(int x=0;x<CopyWidth;x++) {
							uint32 *Pixel = Surface++;
							*Pixel = static_cast<uint32>(*(Image++)) << RShift;
							*Pixel |= static_cast<uint32>(*(Image++)) << GShift;
							*Pixel |= static_cast<uint32>(*(Image++)) << BShift;
						}
					}
					break;
				}
			}
		}
		else {
			CREATE_GAUGE3D_ERROR(mError,GGaugeError::INVALID_ARG,"Unable to handle non 24-bit RGB source image format");
			return;
		}
		break;
	default:
		CREATE_GAUGE3D_ERROR(mError,GGaugeError::INVALID_ARG,"Unable to handle non RGB[A] source image format");
		return;
	}
	return;
}

GAUGE3D_DECLARE_INTERFACES(GpD3D7Texture)
	GAUGE3D_DECLARE_BASECLASS(GTexture)
GAUGE3D_END_INTERFACES

} //namespace Plugins
} //namespace GAUGE3D
