/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: glcamera.cpp
 *      Package: OpenGL Rendering Plugin
 *    Component: Display
 *  Description: Represents the rendering window.
 * Dependencies: GpGLScene, GpGLSprite, GpGLLight
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "glcamera.h"

// system ------------------------------------------------------------
#ifdef PLATFORM_WINDOWS
#include <windows.h>        //needed for gl.h  :(
#endif

#ifndef __SYSTEM_GL_GL_H__INCLUDED__
#include <GL/gl.h>
#define __SYSTEM_GL_GL_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
// lifecycle -------------------------------------------------------
GpGLCamera::GpGLCamera(pGpGLScene scene)
	: mScene(scene),
	  mPosition(GVECTOR_0),
	  mDirection(GVECTOR_I, 0.0f),
	  mZoom(1.0f)
{}

GpGLCamera::~GpGLCamera()
{
}

// operations ------------------------------------------------------
//Position sets the position of the camera.
void
GpGLCamera::Position(const GVector& position)
{
	mPosition = position;
}

//Direction sets the direction in which the camera is facing.
void
GpGLCamera::Direction(const GQuaternion& direction)
{
	mDirection = direction;
}

//Zoom!  1.0 is normal, higher is zoomed in.  Setting the zoom below
// 1.0 is probably a bad idea, as this will increase the FOV beyond
// 90 degrees, which could cause distortion.
void
GpGLCamera::Zoom(GCoordinate zoom)
{
	mZoom = zoom;
}

//Render the scene from the point of view of the camera.
void
GpGLCamera::Draw(float geoLod)
{
	GFrustum viewFrustum;
	viewFrustum.mNear = 0.1;
	viewFrustum.mFar  = 100.0;
	viewFrustum.mHorizontalSlope = 1.0f  / mZoom; //90 degree FOV
	viewFrustum.mVerticalSlope   = 0.75f / mZoom; //4/3 aspect ratio
	viewFrustum.mWidth  = viewFrustum.mHorizontalSlope * viewFrustum.mNear * 2.0f;
	viewFrustum.mHeight = viewFrustum.mVerticalSlope   * viewFrustum.mNear * 2.0f;
	
	//Apply projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-viewFrustum.mWidth  * 0.5f,
	           viewFrustum.mWidth  * 0.5f,
	          -viewFrustum.mHeight * 0.5f,
	           viewFrustum.mHeight * 0.5f,
	           viewFrustum.mNear,
	           viewFrustum.mFar);
	
	//Switch to modelview
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	//Apply camera direction to modelview matrix
	//first, we need to rotate such that the base camera view is pointing in the
	// GVECTOR_FORWARD direction.
	glRotatef(90.0f, 1.0, 0.0, 0.0);
	//Then, we rotate by the camera angle.
	//FIXME:  Should this rotation be inverted?
	GVector axis = mDirection.Axis();
	GCoordinate angle = mDirection.Angle();
	if(!IsApproxZero(angle))
		glRotatef(angle * 180.0f / PI, axis.x, axis.y, axis.z);
	
	//Apply camera position to modelview matrix
	glTranslatef(-mPosition.x, -mPosition.y, -mPosition.z);
	
	//Now that the matrices are set up, the scene can be draw in world coordinates.
	mScene->Draw(mPosition, mDirection, viewFrustum, geoLod);
}


GAUGE3D_DECLARE_INTERFACES(GpGLCamera)
	GAUGE3D_DECLARE_BASECLASS(GCamera)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} //namespace GAUGE3D
