/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: internalobj.cpp
 *      Package: Renterer
 *    Component: Internal Object
 *  Description: The actual implementation of a render object.  This is
 *               kept separate from the interface object (GRenderObject)
 *               in the interests of insulation and encapsulation.
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "glsprite.h"

// system ------------------------------------------------------------
#ifdef PLATFORM_WINDOWS
#include <windows.h>        //needed for gl.h  :(
#endif

#ifndef __SYSTEM_GL_GL_H__INCLUDED__
#include <GL/gl.h>
#define __SYSTEM_GL_GL_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
#ifndef __GAUGE3D_3DMATH_ANGLES_H_INCLUDED__
#include "gauge3d/3dmath/angles.h"
#endif

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpGLSprite::GpGLSprite(pGpGLSprite* link)
	: mModel(NULL),
	  mPosition(0.0f, 0.0f, 0.0f),
	  mAngle(GVECTOR_I, 0.0f),
	  mFullBright(false),
	  mShow(false),
	  mScaleFactor(1.0f)
{
	mColor[0] = 1.0;
	mColor[1] = 1.0;
	mColor[2] = 1.0;
	mColor[3] = 1.0;
	mLightProps[0] = 0.0;
	mLightProps[1] = 0.0;
	mLightProps[2] = 0.0;
	
	AddToList(link);
}

GpGLSprite::~GpGLSprite()
{
}

// operations --------------------------------------------------------	
void
GpGLSprite::AddToList(pGpGLSprite* link)
{
	mNext = *link;
	mPrev = link;
	
	*mPrev = this;
	if(mNext != NULL)
		mNext->mPrev = &mNext;
}

//Draw sprite via OpenGL
void
GpGLSprite::Draw(GVector viewpoint, GQuaternion viewangle, GFrustum viewFrustum, float geoLod)
{
	if(mShow)
	{
		glPushMatrix();
	
		ApplyOrigin(viewpoint, viewangle, viewFrustum);
		ApplyAngle(viewpoint, viewangle, viewFrustum);
		ApplyScale(viewpoint, viewangle, viewFrustum);
		ApplyLightProps();
		ApplyColor();
		
		DrawGeometry(viewpoint, viewangle, viewFrustum, geoLod);
		
		glPopMatrix();
	}
}

void
GpGLSprite::ApplyOrigin(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	//no gTranslatefv?  Odd...
	glTranslatef(mPosition.x, mPosition.y, mPosition.z);
	
	viewpoint -= mPosition;
}

void
GpGLSprite::ApplyAngle(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	GVector axis = mAngle.Axis();
	GCoordinate angle = mAngle.Angle();
	if(!IsApproxZero(angle))
		glRotatef(GAngles::Radians2Degrees(angle), axis.x, axis.y, axis.z);
	
	viewangle = mAngle * viewangle;
}

void
GpGLSprite::ApplyScale(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum)
{
	if(!IsApproxZero(mScaleFactor - 1.0f))
	{
		glScalef(mScaleFactor, mScaleFactor, mScaleFactor);
		glEnable(GL_NORMALIZE);
		
		//rather than grow the object, we shrink the viewspace.
		GCoordinate shrinkFactor = 1.0f / mScaleFactor;
		
		viewpoint *= shrinkFactor;
		viewFrustum.mWidth    *= shrinkFactor;
		viewFrustum.mHeight   *= shrinkFactor;
		viewFrustum.mNear     *= shrinkFactor;
		viewFrustum.mFar      *= shrinkFactor;
	}
	else
		glDisable(GL_NORMALIZE);
}

void
GpGLSprite::ApplyLightProps()
{
	GLfloat color[4];
	int i;

	if(mFullBright)
	{
		glDisable(GL_LIGHTING);
		return;
	}
	
	glEnable(GL_LIGHTING);
	
	//specular (reflective, not affected by color)
	color[3] = mColor[3];          //alpha
	for(i = 0; i < 3; i++)
		color[i] = mLightProps[0];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mLightProps[1]);

	//emission	
	for(i = 0; i < 3; i++)
		color[i] = mColor[i] * mLightProps[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);
}

void
GpGLSprite::ApplyColor()
{
	glColor4fv(mColor);

	float color[4];

	for(int i = 0; i < 3; i++)
		color[i] = mColor[i] * mLightProps[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);
}

void
GpGLSprite::DrawGeometry(GVector& viewpoint, GQuaternion& viewangle,
                         GFrustum& viewFrustum, float geoLod)
{
	GModel::RenderInfo::RenderSet set;
	GLenum format;
	GVector* vertexData = NULL;
	GVector* texcoordData = NULL;
	uint16* indexData = NULL;
	bool hasAlpha = false;
	int numSets;
	int setNum;
	
	numSets = mRenderInfo->StartFrame(mAnimState, viewpoint, viewangle, viewFrustum, geoLod);
	
	for(setNum = 0; setNum < numSets; setNum++)
	{
		mRenderInfo->GetStaticInfo(&set, setNum);
		
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::VERTEX_POSITION)
		{
			//FIXME:  Keep a buffer laying around so that we don't have to constantly
			//        call new.
			vertexData    = new GVector[set.mNumVertices * 2];
			set.mVertices = vertexData;
			set.mNormals  = vertexData + set.mNumVertices;
		}
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::VERTEX_TEXCOORD)
		{
			//FIXME:  Keep a buffer laying around so that we don't have to constantly
			//        call new.
			texcoordData   = new GVector[set.mNumVertices];
			set.mTexCoords = texcoordData;
		}
		if(set.mDynamicFlags & GModel::RenderInfo::RenderSet::INDEX_LIST)
		{
			//FIXME:  Keep a buffer laying around so that we don't have to constantly
			//        call new.
			indexData = new uint16[set.mNumIndices];
			set.mIndices = indexData;
		}
		
		mRenderInfo->GetDynamicInfo(&set, setNum);
		
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, &set.mVertices[0].x);
		
		//if the object is not fullbright, normals are needed for lighting
		if(!mFullBright)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, 0, &set.mNormals[0].x);
		}
		
		if((mTexture != NULL) && (set.mTexCoords != NULL))
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glEnable(GL_TEXTURE_2D);
			mTexture->Bind();
			glTexCoordPointer(2, GL_FLOAT, sizeof(GVector), &set.mTexCoords[0].x);
		}
		
		hasAlpha = HasAlpha();
		if(hasAlpha)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}
		
		// TODO: detect and enable OpenGL extensions, such as compiled vertex arrays.
		//	glLockArraysEXT(0, set->mNumVertices);
		
		switch(set.mIndexType)
		{
		case GModel::RenderInfo::RenderSet::POINTS:
			format = GL_POINTS;
			break;
		case GModel::RenderInfo::RenderSet::LINES:
			format = GL_LINES;
			break;
		default:
		case GModel::RenderInfo::RenderSet::TRIANGLES:
			format = GL_TRIANGLES;
			break;
		}
		glDrawElements(format, set.mNumIndices, GL_UNSIGNED_SHORT,
		               set.mIndices);
		
		//	glUnlockArraysEXT();
		
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisable(GL_TEXTURE_2D);
		
		if(vertexData)
			delete [] vertexData;
		if(texcoordData)
			delete [] texcoordData;
		if(indexData)
			delete [] indexData;
		
		if(hasAlpha)
		{
			glDisable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ZERO);
		}
	}
}

// operations ------------------------------------------------------
//Call Show to tell the renderer to start drawing this object whenever it is visible.
void
GpGLSprite::Show()
{
	mShow = true;
}
//Call Hide to tell the renderer not to draw this object at all, and not to use it in
// any calculations such as shadow casting.
void
GpGLSprite::Hide()
{
	mShow = false;
}

//Call Remove to permenantly remove the sprite from the scene.
void
GpGLSprite::Remove()
{
	if(mPrev != NULL)
		*mPrev = mNext;
	if(mNext != NULL)
		mNext->mPrev = mPrev;
	mPrev = NULL;
	mNext = NULL;
}

// access ----------------------------------------------------------
//A GModel is used to represent the geometry of the object.
void
GpGLSprite::Model(pGModel model)
{
	mModel = model;
	if(model == NULL)
		mRenderInfo = NULL;
	else
		mRenderInfo = model->CreateRenderInfo();
}
//The current animation state of the model must also be given.
void
GpGLSprite::AnimState(GModel::pAnimState anim)
{
	mAnimState = anim;
}
//Set the current texture like so.  The texture object must have come from
// the same renderer object as the sprite.
void
GpGLSprite::Texture(pGTexture texture)
{
	mTexture = texture;
}

//The origin is the location of the object.
void
GpGLSprite::Position(const GVector& position)
{
	mPosition = position;
}
//Angle is the direction the object is pointing.  Remember, you can convert
// GAngles to GQuaternion, so you can use a GAngles object here if you prefer.
void
GpGLSprite::Angle(const GQuaternion& angle)
{
	mAngle = angle;
}

//Use these to set the solid color of the object.  Set to full white to ignore.
void
GpGLSprite::Color(float red, float green, float blue)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
}
void
GpGLSprite::Color(float red, float green, float blue, float alpha)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
	mColor[3] = alpha;
}
void
GpGLSprite::Color3v(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
}
void
GpGLSprite::Color4v(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
	mColor[3] = color[3];
}
void
GpGLSprite::Alpha(float alpha)
{
	mColor[3] = alpha;
}
	
//LightProps sets the light reflectiveness properties of the object.
void
GpGLSprite::LightProps(float specular, float shininess, float emission)
{
	mLightProps[0] = specular;
	mLightProps[1] = shininess;
	mLightProps[2] = emission;
}
//If fullbright is true, lighting will be ignored and the object will be drawn
// at maximum brightness.
void
GpGLSprite::FullBright(bool fullBright)
{
	mFullBright = fullBright;
}

void
GpGLSprite::Scale(GCoordinate magnification)
{
	mScaleFactor = magnification;
}

bool
GpGLSprite::HasAlpha()
{
	if(mTexture)
		if(mTexture->HasAlpha())
			return true;
	if(mColor[3] != 1.0)
		return true;
	return false;
}


GAUGE3D_DECLARE_INTERFACES(GpGLSprite)
	GAUGE3D_DECLARE_BASECLASS(GSprite)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
