/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: gaugelogo.cpp
 *      Package:
 *    Component:
 *  Description: Displays the GAUGE^3D logo / splash screen.
 * Dependencies:
 */

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
#include <string.h>
#include <iostream>
using namespace std;

// external packages -------------------------------------------------
#include <gauge3d/engine.h>
#include <gauge3d/scheduler.h>
#include <gauge3d/osabstraction/messagebox.h>
#include <gauge3d/osabstraction/timer.h>
#include <gauge3d/files/diskfile.h>
#include <gauge3d/plugins/objectloader.h>
#include <gauge3d/renderer/renderer.h>
#include <gauge3d/renderer/display.h>
#include <gauge3d/renderer/sprite.h>
#include <gauge3d/renderer/light.h>
#include <gauge3d/input/messagequeue.h>

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

const float LIGHT_COLORS[3][3] = {{1.0f, 1.0f, 1.0f}, {1.0f, 1.0f, 1.0f}, {1.0f, 1.0f, 1.0f}};
const GCoordinate LIGHT_DISTANCES[] = {3.0F, 3.0F, 3.0F};
const GCoordinate LIGHT_SPEEDS[] = {0.5F, 0.7F, -1.1F};

class SplashScreen: public GObject
{
	enum
	{
		NUM_LIGHTS = 3
	};
	
	pGSprite mBigObj;
	pGLight  mLight[NUM_LIGHTS];
	
	tTime mStartTime;
	
	int mViewPoint;
	
public:
	SplashScreen(pGScene scene, pGModel logo, int viewpoint)
		: mBigObj     (scene->NewSprite()),
		  mViewPoint  (viewpoint)
	{
		mBigObj->Position(GVector(0.0F, 0.0F, -3.0F));
		mBigObj->Color(1.0F, 0.25F, 0.25F);
		
		mBigObj->FullBright(false);
		mBigObj->LightProps(0.8F, 30.0F, 0.0F);
		
		mBigObj->Model(logo);
		GModel::pAnimState anim = logo->CreateAnimState();
		mBigObj->AnimState(anim);
		
		for(int i = 0; i < NUM_LIGHTS; i++)
		{
			mLight[i] = scene->NewLight();
			mLight[i]->LightProps(0.0F, 0.7F, 0.7F);
			mLight[i]->Color(LIGHT_COLORS[i]);
			mLight[i]->Show();
		}
		
		if(mViewPoint == 1)
			mBigObj->Angle(GQuaternion(GAngles(-PI_4, 0, 0)) * GQuaternion(GAngles(0, -PI_4, 0)));
		else
			mBigObj->Angle(GQuaternion(GVECTOR_I, 0.0f));
		
		mBigObj->Show();
		
		mStartTime = GTimer::Time();
		
		gMiscScheduler->CallEveryFrame(this, 0);
	}
	
	~SplashScreen()
	{
		gMiscScheduler->StopCallingEveryFrame(this);
	}
	
	void DoFrame(tTime time, tTime frametime)
	{
		GVector pos;
		double angle;
		static bool firstUpdate = true;
		
		angle = time * LIGHT_SPEEDS[0];
		pos = LIGHT_DISTANCES[0] * GVector(
			cos(angle),
			0.0F,
			sin(angle)
			) - GVector(0.0F, 0.0F, 3.0F);
		mLight[0]->Position(pos);
		
		angle = time * LIGHT_SPEEDS[1];
		pos = LIGHT_DISTANCES[1] * GVector(
			cos(angle) / 2.0F,
			cos(angle) * 1.7320508F / 2.0F,
			sin(angle)
			) - GVector(0.0f, 0.0f, 3.0f);
		mLight[1]->Position(pos);
		
		angle = time * LIGHT_SPEEDS[2];
		pos = LIGHT_DISTANCES[2] * GVector(
			cos(angle) / 2.0f,
			cos(angle) * -1.7320508f / 2.0f,
			sin(angle)
			) - GVector(0.0f, 0.0f, 3.0f);
		mLight[2]->Position(pos);
		
		if(mViewPoint == 0)
		{
			GQuaternion angle;
			
			if(firstUpdate)
			{
				mStartTime = time;
				firstUpdate = false;
			}
			
			time -= mStartTime;
			
			if(time >= 8.0f)
			{
				time -= 8.0f;
				mStartTime += 8.0f;
			}
			
			if(time >= 0.0f && time < 2.0f)
			{
				angle =  GQuaternion(GVECTOR_J, -sin(time * PI) / PI);
				angle *= GQuaternion(GVECTOR_I, -time * PI_2);
			}
			else if(time >= 2.0f && time < 4.0f)
			{
				angle = GQuaternion(GVECTOR_I, -sin(time * PI) / PI);
				angle *= GQuaternion(GVECTOR_I, -PI);
				angle *= GQuaternion(GVECTOR_J, (time - 2.0f) * PI_2);
			}
			else if(time >= 4.0f && time < 6.0f)
			{
				angle = GQuaternion(GVECTOR_J, -sin(time * PI) / PI);
				angle *= GQuaternion(GVECTOR_K, -PI);
				angle *= GQuaternion(GVECTOR_I, (time - 4.0f) * PI_2);
			}
			else if(time >= 6.0f && time < 8.0f)
			{
				angle = GQuaternion(GVECTOR_I, -sin(time * PI) / PI);
				angle *= GQuaternion(GVECTOR_K, -PI);
				angle *= GQuaternion(GVECTOR_I, PI);
				angle *= GQuaternion(GVECTOR_J, -(time - 6.0f) * PI_2);
			}
			
			mBigObj->Angle(angle);
		}
	}
};

class SimpleMessageHandler: public GMessageQueue::Handler
{
public:
	virtual bool HandleMessage(pGMessage message)
	{
		pGInputMessage input = QueryInterface<GInputMessage>(message);
		if(input)
		{
			if(input->Name() == "escape")
				GEngine::Quit();
		}
		else if(QueryInterface<GQuitMessage>(message))
		{
			GEngine::Quit();
			return true;
		}
		
		return false;
	};
};

void
ShowError(GString message)
{
	GMessageBox mbox(message, GMessageBox::ERROR, GMessageBox::OK);
	mbox.Run();
}

void
ShowSplashScreen(bool fullscreen, int viewpoint, int resX, int resY)
{
	gMainMessageQueue->AddHandler(new SimpleMessageHandler);
	
	pGModel logo;
	pGDiskFile logofile(new GDiskFile("gcube.gmdl", GDiskFile::NOCREATE));
	if(!logofile->IsOpen())
	{
		ShowError("Could not find model gcube.gmdl");
		return;
	}
	logo = (pGModel)gMainObjectLoader->LoadObjectForFile(logofile, "GModel");
	if(!logo)
	{
		ShowError("Could not load model gcube.gmdl");
		return;
	}
	
	pGRenderer renderer(gMainObjectLoader->LoadObjectOfType("GRenderer"));
	if(renderer == NULL)
	{
		ShowError("Could not find a rendering plugin.");
		return;
	}
	
	pGDisplay display(renderer->NewDisplay());
	if(display == NULL)
	{
		ShowError("Could not get a display.");
		return;
	}
	if(!display->Open(resX, resY, 0, fullscreen))
	{
		ShowError("Could not open display.");
		return;
	}
	
	pGScene      scene(renderer->NewScene());
	SplashScreen splashScreen(scene, logo, viewpoint);
	pGCamera     camera(scene->NewCamera());
	
	camera->Direction(GQuaternion(GVECTOR_I, -PI_2));
	display->Camera(camera);
	
	GEngine::MainLoop();
}

} // namespace GAUGE3D

int
main(int argc, char* argv[])
{
	bool fullscreen = false;
	int viewpoint = 0;
	int resX = 800, resY = 600;
	
	for(int i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "-f"))
			fullscreen = true;
		else if(!strcmp(argv[i], "-still"))
			viewpoint = 1;
		else if(!strcmp(argv[i], "-stillg"))
			viewpoint = 2;
		else if(!strcmp(argv[i], "-res"))
		{
			resX = int(GAUGE3D::GString(argv[++i]));
			resY = int(GAUGE3D::GString(argv[++i]));
		}
	}
	
	GAUGE3D::GEngine::Init(GAUGE3D::GEngine::VIDEO |
	                       GAUGE3D::GEngine::INPUT |
	                       GAUGE3D::GEngine::OBJECT_LOADER);
	
	GAUGE3D::ShowSplashScreen(fullscreen, viewpoint, resX, resY);
	
	GAUGE3D::GEngine::Shutdown();
	
	return 0;
}
