/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: test_backend.cpp
 *      Package: test progs
 *    Component: backend
 *  Description: backend (non-graphical) testing programs.
 * Dependencies:
 */

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <strstream>
using namespace std;

// external packages -------------------------------------------------
#include "gauge3d/common/error.h"

#include "gauge3d/files/istream.h"
#include "gauge3d/files/ostream.h"

#include "gauge3d/3dmath/vector.h"
#include "gauge3d/3dmath/quaternion.h"
#include "gauge3d/3dmath/angles.h"
#include "gauge3d/3dmath/line.h"
#include "gauge3d/3dmath/plane.h"

#include "gauge3d/osabstraction/filesystem.h"
#include "gauge3d/osabstraction/thread.h"
#include "gauge3d/osabstraction/timer.h"
#include "gauge3d/osabstraction/messagebox.h"

#include "gauge3d/files/diskfile.h"
#include "gauge3d/files/configfile.h"
#include "gauge3d/files/stringfile.h"

#include "gauge3d/input/inputmap.h"

#include "gauge3d/plugins/objectloader.h"
#include "gauge3d/plugins/fileloader.h"

#include "gauge3d/engine.h"

namespace GAUGE3D {

static ostream&
operator<<(ostream& os, const GVector& v) //stream insertion
{
	os << "[";
	if(IsApproxZero(v.x))
		os << 0.0;
	else
		os << v.x;
	os << ", ";
	if(IsApproxZero(v.y))
		os << 0.0;
	else
		os << v.y;
	os << ", ";
	if(IsApproxZero(v.z))
		os << 0.0;
	else
		os << v.z;
	os << "]";
	
	return os;
}

static ostream&
operator<<(ostream& os, GString str)
{
	str = str.ToAscii();
	return os << str.AsciiData();
}

static istream&
operator>>(istream& is, GString& str)
{
	char buffer[256];
	is >> buffer;
	str = buffer;
	return is;
}

/*
static istream&
operator>>(istream& is, GVector& v) //stream extraction
{
	v.w = 1;
	return is >> v.x >> v.y >> v.z;
}
*/

void
TestGString()
{
	GString str("test string");
	cout << str << endl;
	str = str.ToAscii();
	cout << str << endl;
}

void
Test3DMath()
{
	GVector ret;
	GVector vec111(1, 1, 1);
	GCoordinate res;
	bool found;

	cout << "Testing 3D Math package..." << endl;

	found = GPlane(GVECTOR_I, GVECTOR_I * -1).FindIntersection(&ret, GLine(GVECTOR_0, GVECTOR_I));
	if(found)
	{
		cout << ret;
		if(ret == GVector(1, 0, 0))
			cout << " (correct)"   << endl;
		else
			cout << " (incorrect)" << endl;
	}
	else
		cout << "no intersection (incorrect)" << endl;

	found = GLine(GVECTOR_0, GVECTOR_I).FindIntersection(&ret, GLine(vec111, vec111));
	if(found)
	{
		cout << ret;
		if(ret == GVector(0, 0, 0))
			cout << " (correct)"   << endl;
		else
			cout << " (incorrect)" << endl;
	}
	else
		cout << "no intersection (incorrect)" << endl;

	// test for intersection
	if(GLine(GVECTOR_0, vec111).Contains(vec111))
		cout << "intersect1 true  (correct)" << endl;
	else
		cout << "intersect1 false (incorrect)" << endl;
	
	if(GPlane(GVector(1, 1, 0), GVECTOR_I).Contains(vec111))
		cout << "intersect2 true  (correct)" << endl;
	else
		cout << "intersect2 false (incorrect)" << endl;

	// find distances
	cout << (res = GLine(GVECTOR_I, vec111).FindDistance(vec111));
	if(IsApproxZero(float(res - 0.816497)))
		cout << " (correct)"   << endl;
	else
		cout << " (incorrect)" << endl;

	cout << (res = GPlane(GVector(0, 1, 0), GVECTOR_I).FindDistance(vec111));
	if(IsApproxZero(res - 1.0))
		cout << " (correct)"   << endl;
	else
		cout << " (incorrect)" << endl;
	
	cout << (res = GLine(GVECTOR_0, GVECTOR_I).FindDistance(GLine(GVECTOR_J, GVECTOR_K)));
	if(IsApproxZero(res - 1.0))
		cout << " (correct)"   << endl;
	else
		cout << " (incorrect)" << endl;

	
	GAngles angles;
	GAngles::tAngle ang;

	cout << "Now let's play with angles..." << endl;
	cout << "Enter pitch (in degrees): ";
	cin >> ang;
	angles.mPitch = GAngles::Degrees2Radians(ang);
	cout << "Enter yaw   (in degrees): ";
	cin >> ang;
	angles.mYaw   = GAngles::Degrees2Radians(ang);
	cout << "Enter roll  (in degrees): ";
	cin >> ang;
	angles.mRoll  = GAngles::Degrees2Radians(ang);
	
	GVector forward;
	GVector up;
	GVector right;
	
	angles.MakeVectors(&forward, &up, &right);
	cout << "Forward = " << forward << endl;
	cout << "Up      = " << up      << endl;
	cout << "Right   = " << right   << endl;
	
	cout << "\nNow we'll attempt to re-create the original angles from these vectors." << endl;
	
	angles.SetDirection(forward);
	cout << "Using just the forward vector, the angles are:" << endl;
	cout << "Pitch = " << GAngles::Radians2Degrees(angles.mPitch) << endl;
	cout << "Yaw   = " << GAngles::Radians2Degrees(angles.mYaw)   << endl;
	cout << "Roll  = " << GAngles::Radians2Degrees(angles.mRoll)  << endl;
	
	angles.SetDirection(forward, up);
	cout << "Using all three:" << endl;
	cout << "Pitch = " << GAngles::Radians2Degrees(angles.mPitch) << endl;
	cout << "Yaw   = " << GAngles::Radians2Degrees(angles.mYaw)   << endl;
	cout << "Roll  = " << GAngles::Radians2Degrees(angles.mRoll)  << endl;
	
	GQuaternion q(GVECTOR_J, (GCoordinate)PI_2);
	ret = GVECTOR_I;
	q.Rotate(&ret);
	cout << ret << endl;
	
	cout << "Press enter to continue..." << endl;
	cin.get();
	cin.get();
}

#define FS_TEST_FAILED                             \
	{                                               \
		cout << "FAILED" << endl;                    \
		cout << "Press return to shut down the filesystem..." << endl; \
		cin.get();                                   \
		GFileSystem::ShutDown();                     \
		return;                                      \
	}

void TestGFileSystem()
{
	GString dir;
	GString data;
	GString filename;
	int i;
	
	cout << "Initializing filesystem with prog name \"gauge3d\"." << endl;
	GFileSystem::Init("gauge3d");
	
	cout << "Current working directory is:" << endl;
	cout << "   " << GFileSystem::WorkingDir() << endl;
	
	cout << "User config directory is:" << endl;
	cout << "   " << GFileSystem::ConfigDir() << endl;
	
	cout << "Resource directory is:" << endl;
	cout << "   " << GFileSystem::ResourceDir() << endl;
	
	cout << "Plugin directory is:" << endl;
	cout << "   " << GFileSystem::PluginDir() << endl;
	
	cout << "Press return to begin phase 2." << endl;
	cin.get();
	
	cout << "\nCreating file \"testfile\" in config directory with\nflags WRITE, CREATE, NOREPLACE." << endl;
	GFileSystem::File* testfile;
	filename = GFileSystem::ConfigDir();
	filename += "testfile";
	testfile = GFileSystem::Open(filename, GFileSystem::WRITE
	                                     | GFileSystem::CREATE
	                                     | GFileSystem::NOREPLACE);
	if(testfile == NULL)
		FS_TEST_FAILED
	
	cout << "Writing \"this is a test file\" to the file." << endl;
	data = "this is a test file";
	cout << "   " << GFileSystem::Write(testfile, data.AsciiData(), data.Length()) << " bytes written.  (should be 19)" << endl;
	
	cout << "Closing file." << endl;
	GFileSystem::Close(testfile);
	
	cout << "Checking if file exists." << endl;
	if(GFileSystem::Exists(filename))
		cout << "   yes" << endl;
	else
		cout << "   no (it should...)" << endl;
	
	cout << "Attempting to open the same file with flag READ." << endl;
	testfile = GFileSystem::Open(filename, GFileSystem::READ);
	if(testfile == NULL)
		FS_TEST_FAILED
	
	cout << "Reading file:" << endl;
	char buffer[100];
	int size;
	size = GFileSystem::Read(testfile, buffer, 99);
	buffer[size] = 0;
	cout << "   Read " << size << " bytes (should be 19).  Contents were:\n" << buffer << endl;
	
	cout << "Seeking to byte 8 (right before \"a\")." << endl;
	GFileSystem::Seek(testfile, 8, GFileSystem::BEGINNING);
	cout << "TellPos reports file position is " << GFileSystem::TellPos(testfile) << endl;
	
	cout << "Asking how big the file is (using file pointer)." << endl;
	cout << "   File size is " << GFileSystem::FileSize(testfile) << " bytes (should be 19)." << endl;
	
	cout << "Reading remainder of file:" << endl;
	size = GFileSystem::Read(testfile, buffer, 99);
	buffer[size] = 0;
	cout << "   Read " << size << " bytes (should be 11).  Contents were:\n" << buffer << endl;
	
	cout << "Closing file." << endl;
	GFileSystem::Close(testfile);
	
	cout << "Deleting file." << endl;
	GFileSystem::DeleteFile(filename);
	
	cout << "Attempting to open the file without the CREATE flag (should fail)" << endl;
	testfile = GFileSystem::Open(filename, GFileSystem::READ);
	if(testfile != NULL)
	{
		cout << "   It worked, but it shouldn't have." << endl;
		GFileSystem::Close(testfile);
	}
	else
		cout << "   Not able to open.  Good." << endl;
	
	
	cout << "Press return to begin phase 3." << endl;
	cin.get();
	
	cout << "\nTime to test memory mapped files!" << endl;
	cout << "Mapping a new 50-byte file for writing (same name) with\n";
	cout << "   flags WRITE and CREATE." << endl;
	GFileSystem::FileMap* map;
	map = GFileSystem::Map(filename, 0, 50, GFileSystem::WRITE
	                                      | GFileSystem::CREATE);
	if(map == NULL)
		FS_TEST_FAILED
	
	cout << "Filling map with numbers." << endl;
	for(i = 0; i < 50; i++)
		((char*)map->mMap)[i] = '0' + i % 10;
	
	cout << "Unmapping." << endl;
	GFileSystem::UnMap(map);
	
	cout << "Asking how big the file is (using file name)." << endl;
	cout << "   File size is " << GFileSystem::FileSize(filename) << " bytes (should be 50)." << endl;
	
	cout << "Mapping bytes 15-27 for reading." << endl;
	map = GFileSystem::Map(filename, 15, 13, GFileSystem::READ);
	if(map == NULL)
		FS_TEST_FAILED
	
	cout << "Printing contents. (should be \"5678901234567\")" << endl;
	for(i = 0; i < 13; i++)
		cout << ((char*)map->mMap)[i];
	cout << endl;
	
	cout << "Unmapping." << endl;
	GFileSystem::UnMap(map);
	
	
	cout << "Press return to begin phase 4." << endl;
	cin.get();
	
	cout << "\nEnter a directory to list:" << endl;
	cin >> dir;
	
	cout << "Openning directory." << endl;
	GFileSystem::Directory* testdir;
	testdir = GFileSystem::OpenDirectory(dir);
	if(testdir == NULL)
		FS_TEST_FAILED
	
	cout << "Listing contents, and testing if they are files or sub directories." << endl;
	cout << "Press enter to list the files." << endl;
	cin.get();
	cin.get();
	while(GFileSystem::ReadDirectory(&filename, testdir))
	{
		if(GFileSystem::IsDirectory(dir + '/' + filename))
			cout << "   subdir: " << filename << endl;
		else
			cout << "   file:   " << filename << endl;
	}
	
	cout << "Rewinding the directory for another go." << endl;
	GFileSystem::RewindDirectory(testdir);
	
	cout << "Press enter to list the files again." << endl;
	cin.get();
	while(GFileSystem::ReadDirectory(&filename, testdir))
	{
		if(GFileSystem::IsDirectory(dir + '/' + filename))
			cout << "   subdir: " << filename << endl;
		else
			cout << "   file:   " << filename << endl;
	}
	
	cout << "Closing the directory." << endl;
	GFileSystem::CloseDirectory(testdir);
	
	cout << "\nWe're done!" << endl;
	
	cout << "Press return to shut down the filesystem..." << endl;
	cin.get();
	GFileSystem::ShutDown();
}


bool gStopThreads;
int gTime;
pGMutex gTimerMutex;

void* OutThread(void* arg)
{
	int threadNum;
	int delay;

	if(arg)
	{
		threadNum = 1;
		delay = 2;
	}
	else
	{
		threadNum = 2;
		delay = 3;
	}
	
	while(!gStopThreads)
	{
		cout << "Output thread #" << threadNum << ": running with "
		     << delay << " second delay..." << endl;
		
		if(gTimerMutex->TryLock())
		{
			gTimerMutex->Unlock();
			cout << "                  (timer mutex is unlocked)" << endl;
		}
		else
			cout << "                  (timer mutex is locked)" << endl;
		
		GTimer::Sleep(delay);
	}
	
	return NULL;
}

void* TimerThread(void* arg)
{
	while(gTime > 0)
	{
		GTimer::Sleep(1);
		
		gTimerMutex->Lock();
		gTime--;
		gTimerMutex->Unlock();

		cout << "Timer thread: " << gTime << " seconds remaining." << endl;
	}
	
	gStopThreads = true;
	
	return NULL;
}

void* LockThread(void* arg)
{
	GTimer::Sleep(9);

	cout << "Lock thread: Locking timer mutex for 6 seconds..." << endl;
	gTimerMutex->Lock();
	GTimer::Sleep(6);
	cout << "Lock thread: Unlocking timer mutex..." << endl;
	gTimerMutex->Unlock();

	return NULL;
}

void
TestGThread()
{
	GThread outThread1(OutThread);
	GThread outThread2(OutThread);
	GThread timerThread(TimerThread);
	GThread lockThread(LockThread);

	gStopThreads = false;
	gTime = 18;
	gTimerMutex = new GMutex;
	
	outThread1.Start((void*)1);
	outThread2.Start(NULL);
	timerThread.Start(NULL);
	lockThread.Start(NULL);
	
	timerThread.Join();
	cout << "Waiting for output threads to stop..." << endl;
	outThread1.Join();
	outThread2.Join();
	
	gTimerMutex = NULL;
}

void
TestGTimer()
{
	char choice;
	GTimer::Init(0.0f);

	while(1)
	{
		cout << "\n(1) Get time\n"
		     <<   "(0) Quit\n"
		     <<   ":";
		cin >> choice;
	
		switch(choice)
		{
		case '1':
			cout << "Time is: "
			     << GTimer::Time() << endl;
			break;
		case '0':
			return;
			break;
		default:
			break;
		}
	}
}

char* TestGMessageBox_DecodeResult(GMessageBox::tEnum result)
{
	switch(result)
	{
	case GMessageBox::ABORT:
		return "abort";
	case GMessageBox::OK:
		return "OK";
	case GMessageBox::CANCEL:
		return "cancel";
	case GMessageBox::RETRY:
		return "retry";
	case GMessageBox::IGNORE:
		return "ignore";
	case GMessageBox::YES:
		return "yes";
	case GMessageBox::NO:
		return "no";
	default:
		return "unknown";
	}
}

void
TestGMessageBox()
{
	cout << "Attempting to display a message box..." << endl;
	
	GMessageBox* testBox;
	GMessageBox::tEnum result;
	
	testBox = new GMessageBox("Type: INFORMATION\nButtons: OK",
	                          GMessageBox::INFORMATION,
	                          GMessageBox::OK);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	testBox = new GMessageBox("Type: ERROR\nButtons: ABORT_RETRY_IGNORE",
	                          GMessageBox::ERROR,
	                          GMessageBox::ABORT_RETRY_IGNORE);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	testBox = new GMessageBox("Type: WARNING\nButtons: RETRY_CANCEL",
	                          GMessageBox::WARNING,
	                          GMessageBox::RETRY_CANCEL);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	testBox = new GMessageBox("Type: QUESTION\nButtons: YES_NO",
	                          GMessageBox::QUESTION,
	                          GMessageBox::YES_NO);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	testBox = new GMessageBox("Type: INFORMATION\nButtons: OK_CANCEL",
	                          GMessageBox::INFORMATION,
	                          GMessageBox::OK_CANCEL);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	testBox = new GMessageBox("Type: QUESTION\nButtons: YES_NO_CANCEL",
	                          GMessageBox::QUESTION,
	                          GMessageBox::YES_NO_CANCEL);
	result = testBox->Run();
	cout << "You chose: " << TestGMessageBox_DecodeResult(result) << endl;
	
	cout << "Done" << endl;
}

void
TestOSAbstraction()
{
	char choice;
	
	while(1)
	{
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		cout << "OS Abstraction Layer testing program.\n"
		     << "Choose a component:\n"
		     << "(1) GFileSystem\n"
		     << "(2) GThread & GMutex\n"
		     << "(3) GTimer\n"
		     << "(4) GMessageBox\n"
		     << "(0) Quit\n"
		     << "\n"
		     << ": ";

		cin.get(choice);
		if(choice != '\n')
		{
			do
				cin.ignore(100, '\n');
			while(cin.gcount() == 100);
		}
		
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		switch(choice)
		{
		case '1':
			TestGFileSystem();
			break;
		case '2':
			TestGThread();
			break;
		case '3':
			TestGTimer();
			break;
		case '4':
			TestGMessageBox();
			break;
		case '0':
			return;
			break;
		default:
			break;
		}
	}
}

class TestStreamSource: public GRawIStream
{
public:
	TestStreamSource() {}
	virtual ~TestStreamSource() {}

	virtual int Read(void* buffer, int size)
	{
		int bytesRead;
		cin.getline((char*)buffer, size - 1);
		bytesRead = strlen((char*)buffer);
		((char*)buffer)[bytesRead] = '\n';
		((char*)buffer)[bytesRead + 1] = 0;
		return bytesRead + 1;
	}
};

class TestStreamDest: public GRawOStream
{
public:
	TestStreamDest() {}
	virtual ~TestStreamDest() {}

	virtual int Write(const void* buffer, int size)
	{
		cout.write((const char*)buffer, size);
		return size;
	}
};

void
TestStreams()
{
	pGRawIStream streamSource(new TestStreamSource);
	pGRawOStream   streamDest(new TestStreamDest);
	GIStream gcin(streamSource);
	gcin.ReadMode(GIStream::ASCII);
	GOStream gcout(streamDest);
	gcout.WriteMode(GOStream::ASCII);
	
	int i;
	float f;
	
	cout << "For this test, console input will be done through a GIStream." << endl;
	cout << "Enter an integer: ";
	gcin >> i;
	cout << "Integer read was: " << i << endl;
	cout << "Ender a float: ";
	gcin >> f;
	cout << "Fload read was: " << f << endl;
	
	cout << "Now, console output will be directed through a GOStream." << endl;
	gcout << "Your integer was: " << i << '\n';
	gcout << "Your float was:   " << f << '\n';
	gcout.Flush();
}

void
TestGDiskFile()
{
	cout << "Enter the name of a text file to read: " << endl;
	GString filename;
	cin >> filename;
	
	GDiskFile file(filename, GFile::READ);
	if(!file.IsOpen())
	{
		cout << "Could not open file" << endl;
		return;
	}
	
	pGRawIStream inSrc = file.GetRawInputStream();
	if(inSrc == NULL)
	{
		cout << "Could not read file" << endl;
		return;
	}
	
	GIStream input(inSrc, 512);
	
	char buffer[4096];
	int num;
	
	while(1)
	{
		cout << "Enter a relative position to seek to: " << endl;
		cin >> num;
		
		input.Seek(num, GIStream::CURRENT);
		
		cout << "Enter number of bytes to read ( < 4k) (0 to quit): " << endl;
		cin >> num;
		
		if(num == 0)
			break;
		
		input.Read(buffer, num);
		cout.write(buffer, num);
		
		cout << endl << endl;
	}
}

void
TestGConfigFile()
{
	pGVirtualDirectory config;
	
	cout << "For this test, a config file will be read and then re-written." << endl;
	cout << "Enter the name of a config file to read: " << endl;
	GString filename;
	cin >> filename;
	
	cout << "Enter the name of the file to which to write: " << endl;
	GString outfile;
	cin >> outfile;
	
	cout << "Opening file \"" << filename << "\"..." << endl;
	pGDiskFile file = new GDiskFile(filename, GFile::READ);
	cout << "Reading file \"" << filename << "\"..." << endl;
	config = GConfigFile::Load(file);
	cout << "Opening file \"" << outfile << "\"..." << endl;
	file = new GDiskFile(outfile, GFile::READ);
	cout << "Writing file \"" << outfile << "\"..." << endl;
	GConfigFile::Save(file, config);
	cout << "Done" << endl;
}


void
TestFiles()
{
	char choice;
	
	while(1)
	{
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		cout << "File manipulation testing program.\n"
		     << "Choose a component:\n"
		     << "(1) GDiskFile\n"
		     << "(2) GConfigFile\n"
		     << "(0) Quit\n"
		     << "\n"
		     << ": ";

		cin.get(choice);
		if(choice != '\n')
		{
			do
				cin.ignore(100, '\n');
			while(cin.gcount() == 100);
		}
		
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		switch(choice)
		{
		case '1':
			TestGDiskFile();
			break;
		case '2':
			TestGConfigFile();
			break;
		case '0':
			return;
			break;
		default:
			break;
		}
	}
}


void
PrintObjectInfo(const GObjectLoader::ObjectInfo& info)
{
	int i;
	
	if(&info == NULL)
		return;
	
	cout << "\nObject:\n"
	     << "   Classname:    " << info.mClassname << endl
	     << "   Name:         " << info.mName << endl
	     << "   Description:  " << info.mDescription << endl;
	
	cout << "   Base classes: ";
	for(i = 0; i < info.mBaseClasses.Size(); i++)
		cout << info.mBaseClasses[i] << " ";
	cout << endl;
	
	cout << "   Key words:    ";
	for(i = 0; i < info.mKeywords.Size(); i++)
		cout << info.mKeywords[i] << " ";
	cout << endl;
	
	cout << "   File types:   ";
	for(i = 0; i < info.mFileTypes.Size(); i++)
		cout << info.mFileTypes[i] << " ";
	cout << endl;
}

void
PrintPluginInfo(const GObjectLoader::PluginInfo& info)
{
	int i;
	
	if(&info == NULL)
		return;
	
	cout << "\nPlugin:\n"
	     << "   Name:         " << info.mName << endl
	     << "   Description:  " << info.mDescription << endl
	     << "   Library Name: " << info.mLibName << endl;
	
	cout << "   Objects:      ";
	for(i = 0; i < info.mObjects.Size(); i++)
		cout << info.mObjects[i].mClassname << " ";
	cout << endl;
}

void
TestObjectLoader()
{
	char choice;
	GString name, keyword, extension, baseclass;
	pGStringFile dummyFile = new GStringFile;
	
	GEngine::Init(GEngine::OBJECT_LOADER);
	
	GSmartPointer<GObjectLoader::ObjectInfo> objectInfo;
	GArray<GObjectLoader::ObjectInfo> objectInfoList;
	GSmartPointer<GObjectLoader::PluginInfo> pluginInfo;
	GArray<GObjectLoader::PluginInfo> pluginInfoList;
	int i;
	
	while(true)
	{
		cout << "Choose one:\n"
		     << "(1) Find an object\n"
		     << "(2) Find all objects with a given base class\n"
		     << "(3) Find all objects with a given keyword\n"
		     << "(4) Find all objects which can load a given file\n"
		     << "(5) List all plugins\n"
		     << "(6) Find plugin containing object\n"
		     << "(0) Quit\n"
		     << "\n"
		     << ": ";
		
		cin.get(choice);
		if(choice != '\n')
		{
			do
				cin.ignore(100, '\n');
			while(cin.gcount() == 100);
		}
		
		switch(choice)
		{
		case '1':
			cout << "Enter class name: " << endl;
			cin >> name;
			objectInfo = gMainObjectLoader->FindObject(name);
			PrintObjectInfo(*objectInfo);
			cin.get();
			break;
		case '2':
			cout << "Enter base class name: " << endl;
			cin >> baseclass;
			objectInfoList = gMainObjectLoader->FindObjectsOfType(baseclass);
			for(i = 0; i < objectInfoList.Size(); i++)
				PrintObjectInfo(objectInfoList[i]);
			cin.get();
			break;
		case '3':
			cout << "Enter base class name: " << endl;
			cin >> baseclass;
			cout << "Enter keyword: " << endl;
			cin >> keyword;
			objectInfoList = gMainObjectLoader->FindObjectsWithKeyword(keyword, baseclass);
			for(i = 0; i < objectInfoList.Size(); i++)
				PrintObjectInfo(objectInfoList[i]);
			cin.get();
			break;
		case '4':
			cout << "Enter base class name: " << endl;
			cin >> baseclass;
			cout << "Enter (fake) file name: " << endl;
			cin >> extension;
			dummyFile->Name(extension);
			objectInfoList = gMainObjectLoader->FindObjectsForFile(dummyFile, baseclass);
			for(i = 0; i < objectInfoList.Size(); i++)
				PrintObjectInfo(objectInfoList[i]);
			cin.get();
			break;
		case '5':
			pluginInfoList = gMainObjectLoader->FindAllPlugins();
			for(i = 0; i < pluginInfoList.Size(); i++)
				PrintPluginInfo(pluginInfoList[i]);
			break;
		case '6':
			cout << "Enter class name: " << endl;
			cin >> name;
			pluginInfo = gMainObjectLoader->FindPluginForObject(name);
			PrintPluginInfo(*pluginInfo);
			cin.get();
			break;
		case '0':
			GEngine::Shutdown();
			return;
		default:
			break;
		}
		cout << endl;
	}
}


void
MakeGMDL()
{
	GEngine::Init(GEngine::OBJECT_LOADER);
	
	pGObject model;
	pGFileLoader loader;
	pGDiskFile input;
	pGDiskFile output;
	GString inputname;
	GString outputname;
	
	loader = gMainObjectLoader->LoadObject("GpGMDLLoader");
	if(loader == NULL)
	{
		cout << "Error loading plugins." << endl;
		return;
	}
	
	cout << "Enter the filename of the input file:" << endl;
	cin  >> inputname;
	
	input = new GDiskFile();
	input->Open(inputname, GDiskFile::NOCREATE);
	if(!(input->IsOpen()))
	{
		cout << "Could not open file." << endl;
		return;
	}
	
	model = gMainObjectLoader->LoadObjectForFile(input, "GSuperModel");
	if(model == NULL)
	{
		cout << "No plugin found which could load the file." << endl;
		return;
	}
	
	cout << "Enter the filename of the output file:" << endl;
	cin  >> outputname;
	
	output = new GDiskFile();
	output->Open(outputname, GDiskFile::CREATE);
	if(!(output->IsOpen()))
	{
		cout << "Could not open file." << endl;
		return;
	}
	
	if(!loader->Save(output, model))
	{
		cout << "Could not write model." << endl;
		return;
	}
	
	cout << "Success!" << endl;
	
	model  = NULL;
	loader = NULL;
	input  = NULL;
	output = NULL;
	
	GEngine::Shutdown();
}


void
TestBackend()
{
	char choice;
	
	while(1)
	{
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		cout << "GAUGE back-end testing program.\n"
		     << "Choose a package:\n"
		     << "(1) 3DMath        - Basic 3D Math\n"
		     << "(2) OSAbstraction - Operating System Abstraction Layer\n"
		     << "(3) Streams       - GAUGE stream classes\n"
		     << "(4) Files         - GAUGE file classes\n"
		     << "(5) ObjectLoader  - The Object Loader\n"
		     << "(6) Make a GMDL from some other model format.\n"
		     << "(7) Cause a Fatal Error\n"
		     << "(0) Quit\n"
		     << "\n"
		     << ": ";

		cin.get(choice);
		if(choice != '\n')
		{
			do
				cin.ignore(100, '\n');
			while(cin.gcount() == 100);
		}
		
		cout << endl << "-------------------------------------------------------------------\n\n";
		
		switch(choice)
		{
		case '1':
			Test3DMath();
			break;
		case '2':
			TestOSAbstraction();
			break;
		case '3':
			TestStreams();
			break;
		case '4':
			TestFiles();
			break;
		case '5':
			TestObjectLoader();
			break;
		case '6':
			MakeGMDL();
			break;
		case '7':
			GAUGE3D_FATAL_ERROR("You caused a fatal error!")
		case '0':
			return;
			break;
		default:
			break;
		}
	}
}

} // namespace GAUGE3D

int
main(int argc, char* argv[])
{
	try
	{
		GAUGE3D::TestBackend();
	}
	catch(const GAUGE3D::GException& error)
	{
		char message[1024];
		strstream msgstream(message, sizeof(message));
		msgstream << error.Cause()
		          << "\nFile: " << error.File()
		          << "\nLine: " << error.Line() << '\0';
		GAUGE3D::GMessageBox errorBox(message, GAUGE3D::GMessageBox::ERROR, GAUGE3D::GMessageBox::OK);
		errorBox.Run();
	}
	
	return 0;
}
